#if 0
	set -ex
	cc \
		-I/usr/X11/include \
		-L/usr/X11/lib "$@" $0 -lX11
	exit
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <X11/Xlib.h>

static void
oops(char *where)
{
	fprintf(stderr, "oops: %s\n", where);
	exit(1);
}

int
main(int argc, char **argv)
{
	int
		width    =  8,
		height   = 16,
		baseline = 13,
		c;
	char
		chr,
		*font_name = NULL;

	Display *dpy;
	XFontStruct *f;
	Window pix;
	GC gc;

	extern char *optarg;
	extern int optind, opterr;
	opterr = 0;

	while ((c = getopt(argc, argv, "")) != -1) {
		switch (c) {
		default:
	usage:
			fprintf(stderr, "\
Usage: getfont font_name\n\
");
			exit(2);
		}
	}
	font_name = argv[optind++];

	if (optind != argc)
		goto usage;

	if ((dpy = XOpenDisplay(getenv("DISPLAY"))) == NULL)
		oops("open display");

	f = XLoadQueryFont(dpy, font_name);
	if (!f)
		oops("load font");

	pix = XCreatePixmap(dpy, DefaultRootWindow(dpy),
		width, height, DefaultDepth(dpy, DefaultScreen(dpy)));
	if (!pix)
		oops("create pixmap");

	gc = DefaultGC(dpy, DefaultScreen(dpy));

	XSetFont(dpy, gc, f->fid);
	XSetBackground(dpy, gc, 0);

	for (c = 0; c < 256; ++c) {
		int x, y, xel;
		XImage *image;

		XSetForeground(dpy, gc, 0);
		XFillRectangle(dpy, pix, gc, 0, 0, width, height);
		XSetForeground(dpy, gc, 1);

		chr = c;
		XDrawString(dpy, pix, gc, 0, baseline, &chr, 1);

		image = XGetImage(dpy, pix, 0, 0, width, height, 1, XYPixmap);

		for (y = 0; y < height; ++y) {

			printf("\t.byte %%");

			for (x = 0; x < width; ++x) {
				xel = XGetPixel(image, x, y);
				printf("%c", xel ? '1' : '0');
			}
			printf(" | line%x ; ", y); /* underline etc */

			for (x = 0; x < width; ++x) {
				xel = XGetPixel(image, x, y);
				printf("%c", xel ? '#' : ' ');
			}
			if (y == 0) {
				printf(" --- 0x%02X", c);
				if ((c >= 0x20 && c < 0x7E)
				 || (c >= 0xA0 && c < 0xFF))
					printf(" %c", c);
			}
			printf("\n");
		}
		/* XXX free the image */
	}
	exit(0);
}

