
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/file.h"
#include "../h/dir.h"
#include "../h/filsys.h"
#include "../h/mount.h"
#include "../h/map.h"
#include "../h/proc.h"
#include "../h/inode.h"
#include "../h/conf.h"
#include "../h/buf.h"
#include "../h/clock.h"

struct inode *rootdir;
struct proc *curproc;

int	lbolt;			/* time of day in 60th not in time */

int	mpid;			/* generic for unique process id's */
char	runrun;			/* scheduling flag */
char	updlock;		/* lock for sync */

daddr_t	rablock;		/* block to be read ahead */

/*
 * Initialization code.
 * Called from cold start routine as
 * soon as a stack and segmentation
 * have been established.
 * Functions:
 *	clear and free user core
 *	turn on clock
 *	hand craft 0th process
 *	call all initialization routines
 *	fork - process 0 to schedule
 *	     - process 2 to page out
 *	     - process 1 execute bootstrap
 */

#ifdef BCC
extern char _etext[], _edata[], _end[], _segoff[];
int aaaa = _etext;
int aaab = _edata;
int aaac = _end;
int aaad = _segoff;
#endif

int procsz = sizeof(struct proc);
int usersz = sizeof(struct user);

RMAP(coremap, 2 * NPROC + 1)

main()
{
	proc_init->p_stat = SRUN;
	proc_init->p_flag |= SLOAD|SSYS;
	proc_init->p_nice = NZERO;
	proc_init->p_parent = proc_init;
	curproc = proc_init;

	/* paragraph units */

	mfree(coremap, 0x5000, 0x1000);

	/*
	 * Initialize devices and
	 * set up 'known' i-nodes
	 */

	callout_init();
	clist_init();
	binit();
	mountroot();

	curproc->p_cdir = rootdir;
	curproc->p_cdir->i_count++;
}

mountroot()
{
	struct filsys *fp;
	struct mount *mp;
	struct buf *bp;

	for (mp = &mount[NMOUNT]; --mp > mount; )
		mp->m_dev = NODEV;

	mp->m_dev = rootdev;
	fp = &mp->m_filsys;

	(*bdevsw[major(rootdev)].d_open)(rootdev, FREAD | FWRITE);

	bp = bread(rootdev, SUPERB);
	if (errno)
		panic("iinit");

	bcopy(bp->b_un.b_addr, fp, sizeof(*fp));
	brelse(bp);

	fp->s_flock = 0;
	fp->s_ilock = 0;
	fp->s_ronly = 0;
	fp->s_lasti = 1;
	fp->s_nbehind = 0;

	rootdir = iget(mp, ROOTINO);
	rootdir->i_flag &= ~ILOCK;
	mp->m_mpoint = rootdir;
}

/*
 * This is the set of buffers proper, whose heads
 * were declared in buf.h.
 */
char	buffers[NBUF][BSIZE];

/*
 * Initialize the buffer I/O system by freeing
 * all buffers and setting all device buffer lists to empty.
 */
binit()
{
	register struct buf *bp;
	register int i;

	bfreelist.b_forw =
	bfreelist.b_back =
	bfreelist.av_forw =
	bfreelist.av_back = &bfreelist;

	for (i = 0; i < NBUF; ++i) {
		bp = &buf[i];
		bp->b_un.b_addr = buffers[i];
		bp->b_back = &bfreelist;
		bp->b_forw = bfreelist.b_forw;
		bfreelist.b_forw->b_back = bp;
		bfreelist.b_forw = bp;
		bp->b_dev = NODEV;
		bp->b_flags = B_BUSY;
		brelse(bp);
	}
	for (i = 0; i < nblkdev; ++i) {
		bp = bdevsw[i].d_tab;
		if (bp) {
			bp->b_forw = bp;
			bp->b_back = bp;
		}
	}
}

spl0() {}
splbio() {}
spltty() {}
splhigh() {}
splclock() {}
splx() {}

__mkargv() { trap(); hardclock(); syscall(); }

idle() { /* sti; halt */ }

sendsig() {}
reboot() {}

bank(b) { /* b is 0 or 1 */ }
far_bcopy() {}
bcopy() {}
bzero() {}
memcpy() {} /* struct copies */
strncmp() {}

copyin() {}
copyout() {}
subyte() {}
suword() {}
fubyte() {}
fuword() {}

