
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/dir.h"
#include "../h/proc.h"
#include "../h/map.h"
#include "../h/text.h"

static void
squeeze_out(mp, bp)
	register struct map *mp;
	register struct mapent *bp;
{
	mp->m_end--;
	bcopy(bp + 1, bp, (mp->m_end - bp) * sizeof(*bp));
}

static void
squeeze_in(mp, bp, size, a)
	register struct map *mp;
	register struct mapent *bp;
{
	bcopy(bp, bp + 1, (mp->m_end - bp) * sizeof(*bp));
	bp->m_size = size;
	bp->m_addr = a;

	if (++(mp->m_end) > mp->m_END)
		panic(mp->m_name);
}

/*
 * Allocate 'size' units from the given map.
 * Return the base of the allocated space.
 * In a map, the addresses are increasing
 * and the list is terminated by a 0 size.
 * Algorithm is first-fit.
 */
malloc(mp, size)
	register struct map *mp;
	unsigned size;
{
	register struct mapent *bp;
	unsigned a;

	for (bp = mp->m_map; bp < mp->m_end; ++bp) {
		if (bp->m_size >= size) {
			a = bp->m_addr;
			bp->m_addr += size;
			if ((bp->m_size -= size) == 0)
				squeeze_out(mp, bp);

			return (a);
		}
	}
	return (0);
}

/*
 * Free the previously allocated space a
 * of size units into the specified map.
 * Sort a into map and combine on
 * one or both ends if possible.
 */
void
mfree(mp, size, a)
	register struct map *mp;
	unsigned size, a;
{
	register struct mapent *bp;

	for (bp = mp->m_map; bp < mp->m_end && bp->m_addr <= a; ++bp)
		;

	if (bp > mp->m_map && bp[-1].m_addr + bp[-1].m_size == a) {
		bp[-1].m_size += size;
		if (bp < mp->m_end && a + size == bp->m_addr) {
			bp[-1].m_size += bp->m_size;
			squeeze_out(mp, bp);
		}
	} else if (bp < mp->m_end && a + size == bp->m_addr) {
		bp->m_size += size;
		bp->m_addr -= size;
	} else {
		squeeze_in(mp, bp, size, a);
	}
}

