
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/conf.h"
#include "../h/filsys.h"
#include "../h/mount.h"
#include "../h/errno.h"

/*
 * In case console is off,
 * panicstr contains argument to last
 * call to panic.
 */

char	*panicstr;

/*
 * Scaled down version of C Library printf.
 * Only %s %u %d (==%u) %o %x %D are recognized.
 * Used to print diagnostic information
 * directly on console tty.
 * Since it is not interrupt driven,
 * all system activities are pretty much
 * suspended.
 * Printf should not be used for chit-chat.
 */
printf(fmt, args)
	char *fmt;
{
	int c, *adx = &args;
	char *s;

	for (;;) {
		while ((c = *fmt++) != '%') {
			if (c == 0)
				return;
			putchar(c);
		}
		c = *fmt++;

		if (c == 'd' || c == 'u' || c == 'o' || c == 'x') {
			printn((long) *adx, (c == 'o') ? 8 : (c == 'x') ? 16 : 10);

		} else if (c == 's') {
			s = (char *) *adx;
			while (c = *s++)
				putchar(c);

		} else if (c == 'D' || c == 'X') {
			printn(*(long *) adx, (c == 'X') ? 16 : 10);
			adx += (sizeof(long) / sizeof(int)) - 1;
		}
		adx++;
	}
}

/*
 * Print an unsigned long in base b.
 */
printn(n, b)
	long n;
{
	register long a;

	if (n < 0) {
		n = -n;
		putchar('-');
	}
	if (a = (n / b))
		printn(a, b);

	putchar("0123456789ABCDEF"[n % b]);
}

/*
 * Panic is called on unresolvable fatal errors.
 * It syncs, prints "panic: mesg" and then loops.
 */
panic(s)
	char *s;
{
	panicstr = s;
	update();

	printf("panic: %s\n", s);

	reboot();
}

/*
 * fserr prints a warning message of the form "x/y overheating".
 */
fserr(mp, msg)
	struct mount *mp;
	char *msg;
{
	dev_t dev = mp->m_dev;

	printf("%u/%u %s\n", major(dev), minor(dev), msg);
}


