
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/filsys.h"
#include "../h/mount.h"
#include "../h/inode.h"
#include "../h/dir.h"
#include "../h/proc.h"
#include "../h/buf.h"
#include "../h/conf.h"

/*
 * Read the file corresponding to
 * the inode pointed at by the argument.
 * The actual read arguments are found
 * in the variables:
 *	u_base		core address for destination
 *	u_offset	byte offset in file
 *	u_count		number of bytes to read
 *	u_segflg	read to kernel/user/user I
 */
readi(ip, f_flag)
	struct inode *ip;
{
	struct buf *bp;
	dev_t dev;
	daddr_t lbn, bn;
	off_t diff;
	int on, type;
	unsigned n;

	if (u.u_count == 0)
		return;

	if (u.u_offset < 0) {
		errno = EINVAL;
		return;
	}
	dev  = ip->i_rdev;
	type = ip->i_mode & IFMT;

	if (type == IFCHR) {
		(*cdevsw[major(dev)].d_read)(dev, f_flag);
		return;
	}

	do {
		lbn = bn = u.u_offset >> BSHIFT;
		on = u.u_offset & BMASK;
		n = min(BSIZE - on, u.u_count);
		if (type != IFBLK) {
			diff = ip->i_size - u.u_offset;
			if (diff <= 0)
				return;
			if (diff < n)
				n = diff;
			bn = bmap(ip, bn, B_READ);
			if (errno)
				return;
			dev = ip->i_mp->m_dev;
		} else
			rablock = bn + 1;

		if ((long) bn < 0) {
			bp = geteblk();
			clrbuf(bp);
		} else if (ip->i_lastr + 1 == lbn)
			bp = breada(dev, bn, rablock);
		else
			bp = bread(dev, bn);

		ip->i_lastr = lbn;
		n = min(n, BSIZE - bp->b_resid);
		if (n)
			iomove(bp->b_un.b_addr + on, n, B_READ);
		brelse(bp);
	} while(errno == 0 && u.u_count && n);
}

/*
 * Write the file corresponding to
 * the inode pointed at by the argument.
 * The actual write arguments are found
 * in the variables:
 *	u_base		core address for source
 *	u_offset	byte offset in file
 *	u_count		number of bytes to write
 *	u_segflg	write to kernel/user/user I
 */
writei(ip, f_flag)
	struct inode *ip;
{
	struct buf *bp;
	dev_t dev;
	daddr_t bn;
	int on, type;
	unsigned n;

	if (u.u_count == 0)
		return;

	if (u.u_offset < 0) {
		errno = EINVAL;
		return;
	}
	dev = (dev_t)ip->i_rdev;
	type = ip->i_mode&IFMT;

	if (type == IFCHR) {
		(*cdevsw[major(dev)].d_write)(dev, f_flag);
		return;
	}

	do {
		bn = u.u_offset >> BSHIFT;
		on = u.u_offset & BMASK;
		n = min((unsigned)(BSIZE-on), u.u_count);

		if (type != IFBLK) {
			bn = bmap(ip, bn, B_WRITE);
			if((long)bn<0)
				return;
			dev = ip->i_mp->m_dev;
		}
		if (n == BSIZE) 
			bp = getblk(dev, bn);
		else
			bp = bread(dev, bn);
		iomove(bp->b_un.b_addr+on, n, B_WRITE);
		if (errno)
			brelse(bp);
		else
			bdwrite(bp);
		if (type != IFBLK && u.u_offset > ip->i_size)
			ip->i_size = u.u_offset;
		ip->i_flag |= IUPD|ICHG;
	} while(errno==0 && u.u_count);
}

/*
 * Move n bytes at byte location &bp->b_un.b_addr[o]
 * to/from (rw) the kernel/user/user I (u.u_segflg) area
 * starting at u.base.
 * Update all the arguments by the number of bytes moved.
 */
iomove(cp, n, rw)
	caddr_t cp;
	unsigned n;
{
	int t;

	if (n == 0)
		return;

	if (u.u_segflg) {
		if (rw == B_WRITE)
			t = copyin(u.u_base, cp, n);
		else
			t = copyout(cp, u.u_base, n);
		if (t) {
			errno = EFAULT;
			return;
		}
	} else {
		if (rw == B_WRITE)
			bcopy(u.u_base, cp, n);
		else
			bcopy(cp, u.u_base, n);
	}
	u.u_offset += n;
	u.u_base   += n;
	u.u_count  -= n;
}

