
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/dir.h"
#include "../h/proc.h"

/*
 * From tty. Send the specified signal to
 * all processes with 'pgrp' as process group.
 */
ttysignal(pgrp, sig)
	register pgrp;
{
	register struct proc *p;

	if (pgrp)
		for (p = proc; p < &proc[NPROC]; ++p)
			if (p->p_pgrp == pgrp)
				psignal(p, sig);
}

/*
 * Send the specified signal to the specified process.
 */
psignal(p, sig)
	register struct proc *p;
	register unsigned sig;
{
	int s;

	if ((sig - 1) >= NSIG - 1)
		return;

	s = splhigh();
	p->p_sig |= sigmask(sig);

	if (p->p_pri > PUSER)
		p->p_pri = PUSER;
	if (p->p_pri > PZERO && p->p_stat == SSLEEP)
		setrun(p);

	splx(s);
}

/*
 * Returns true if the current
 * process has a signal to process.
 * This is asked at least once
 * each time a process enters the
 * system.
 * A signal does not do anything
 * directly to a process; it sets
 * a flag that asks the process to
 * do something to itself.
 */
issig()
{
	register n;
	register struct proc *p;

	p = curproc;
	while (p->p_sig) {
		n = fsig(p);
		if (p->p_signal[n] != SIG_IGN)
			return (n);

		p->p_sig &= ~sigmask(n);
	}
	return (0);
}

/*
 * Perform the action specified by
 * the current signal.
 * The usual sequence is:
 *	if (issig())
 *		psig();
 */
psig()
{
	register n, handler;
	register struct proc *p;

	p = curproc;
	n = fsig(p);
	if (n == 0)
		return;

	p->p_sig &= ~sigmask(n);

	if (handler = p->p_signal[n]) {
		if (n != SIGILL
		 && n != SIGTRAP)
			p->p_signal[n] = 0;
		errno = 0;
		sendsig(handler, n);
		return;
	}
	sig_exit(n);
}

/*
 * find the signal in bit-position
 * representation in p_sig.
 */
fsig(p)
	struct proc *p;
{
	register sig = 1;
	register unsigned n = p->p_sig;

	do {
		if (n & 1)
			return (sig);
		++sig;
	} while (n >>= 1);

	return (0);
}

