
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/map.h"
#include "../h/dir.h"
#include "../h/proc.h"
#include "../h/file.h"
#include "../h/buf.h"
#include "../h/inode.h"
#include "../h/text.h"
#include "../h/exec.h"
#include "../h/reg.h"

void
newds(p, sz)
	struct proc *p;
{
	p->p_dsize = sz;
	p->p_ds = malloc(coremap, p->p_dsize);
}

void
dupds(p, parent)
	struct proc *p, *parent;
{
	p->p_dsize = parent->p_dsize;
	p->p_ds = malloc(coremap, p->p_dsize);

	far_bcopy(ptob(parent->p_ds), ptob(p->p_ds), ptob(p->p_dsize));
}

void
relmem(p)
	struct proc *p;
{
	if (p->p_textp) {
		xfree(p->p_textp);
		p->p_textp = NULL;
	}
	if (p->p_ds) {
		mfree(coremap, p->p_dsize, p->p_ds);
		wakeup(coremap);
		p->p_ds = 0;
	}
}

sys_execve(fname, argv, envp)
	char *fname, **argv, **envp;
{
	char *cp, *ucp, **v;
	int c, cc, argc, envc, twice;
	struct inode *ip;
	struct buf *bp = NULL;
	unsigned dsiz;
	struct exec ex;

	if ((ip = namei(fname, 0)) == NULL)
		return;

	if ((ip->i_mode & (IFMT|IEXEC)) != (IFREG|IEXEC))
		goto noexec;

	u.u_offset = 0;
	u.u_count = sizeof(ex);
	u.u_base = &ex;
	u.u_segflg = 0;
	readi(ip, 0);

	if (errno)
		goto out;

	if (u.u_count
	 || ex.ex_mag != 0xFEFE
	 || ex.ex_tsize > MAXTSIZ) {
noexec:
		errno = ENOEXEC;
		goto out;
	}
	if (ex.ex_tsize && ip->i_count != 1 && (ip->i_flag&ITEXT) == 0) {
		errno = ETXTBSY;
		goto out;
	}

	/* collect characters, reverse order */

	bp = geteblk();
	cp = bp->b_un.b_addr + BSIZE;
	cc = 0;

#if NCARGS > BSIZE
	needs fixing
#endif

	twice = 2;
	v = argv;
doenv:
	if (v) {
		while (ucp = fuword(v++)) {
			c = 0;
			do {
				if (++cc > NCARGS)
					errno = E2BIG;
				*--cp = c;
				if ((c = fubyte(ucp++)) < 0)
					errno = EFAULT;
				if (errno)
					goto out;
			} while (c);
		}
	}
	envc = v - envp;
	if (--twice) {
		argc = v - argv;
		v = envp;
		goto doenv;
	}

	/* (argc, argv, envp) and 2 NULLs makes 5 words */

	dsiz = ex.ex_dsize
	     + ex.ex_bsize
	     + ex.ex_ssize
	     + btop((argc + envc + 5) * sizeof(char *) + cc);

	/* committed */

	relmem(curproc);

	if (ex.ex_tsize)
		xalloc(ip, &ex);

	newds(curproc, dsiz);

	u.u_offset = sizeof(ex) + ptob(ex.ex_tsize);
	u.u_count = ex.ex_dsize;
	u.u_base = ptob(ex.ex_tsize);
	u.u_segflg = 1;
	readi(ip, 0);

	/* bss not zeroed */

	/* arguments, again reversing the order,
	 * from end of data segment downwards */

	ucp = ptob(dsiz);
	v = (char **) (ucp - cc);

	twice = 2;
doargv:
	suword(--v, NULL);
	while (envc--) {
		subyte(--ucp, 0);
		while (c = *cp++)
			subyte(--ucp, c);
		suword(--v, ucp);
	}
	argv = v;
	if (--twice) {
		envp = v;
		envc = argc;
		goto doargv;
	}
	suword(--v, envp);
	suword(--v, argv);
	suword(--v, argc);   /* ready to call main(argc, argv, envp); */

	setregs(curproc, v);

	if (errno)
		psignal(curproc, SIGKILL);
out:
	iput(ip);
	if (bp)
		brelse(bp);
}

setregs(p, v)
	struct proc *p;
	register int *v;
{
	register int i;
	int cs, ds = p->p_ds;

	cs = p->p_textp ? p->p_textp->x_cs : ds;

	suword(--v, FL_USER); /* FL      INT */
	suword(--v, cs);      /* CS */
	suword(--v, 0);       /* IP */

	suword(--v, 0); /* AX          PUSHA */
	suword(--v, 0); /* CX */
	suword(--v, 0); /* DX */
	suword(--v, 0); /* BX */
	suword(--v, 0); /* SP (ignored by POPA) */
	suword(--v, 0); /* BP */
	suword(--v, 0); /* SI */
	suword(--v, 0); /* DI */

	suword(--v, ds); /* DS       PUSH DS  PUSH ES */
	suword(--v, ds); /* ES */

	p->p_sp = v;
	p->p_ss = ds;

	for (i = 0; i < NSIG; ++i)
		if (p->p_signal[i] != SIG_IGN)
			p->p_signal[i] = SIG_DFL;

	for (i = 0; i < NOFILE; ++i) {
		if (u.u_pofile[i] & EXCLOSE) {
			closef(u.u_ofile[i]);

			u.u_ofile[i] = NULL;
			u.u_pofile[i] = 0;
		}
	}
}

sys_exit(rv)
{
	exit(rv << 8);
}

sig_exit(sig)
{
	exit(sig ? sig : SIGKILL);
}

exit(xstat)
{
	register struct proc *p = curproc;
	register int i;
	int pid;

	if (p == proc_init)
		panic("init exits");

	p->p_flag |= SWEXIT;
	p->p_clktim = 0;

	for (i = 0; i < NSIG; ++i)
		p->p_signal[i] = SIG_IGN;

	for (i = 0; i < NOFILE; ++i)
		closef(u.u_ofile[i]);

	if (p->p_ctty)
		iput(p->p_ctty);

	plock(p->p_cdir);
	iput(p->p_cdir);

	relmem(p);

	p->p_stat = SZOMB;
	p->p_xstat = xstat;

	/* gone.
	 * wakeup parent, if not init.
	 * give childs to init.
	 * wakeup init, if adopted zombies. */

	pid = p->p_pid;
	p = p->p_parent;
	i = (p == proc_init);
	if (i == 0)
		wakeup(p);

	for (p = proc; p < &proc[NPROC]; ++p) {

		if (p->p_parent == curproc) {
			p->p_parent = proc_init;
			if (p->p_stat == SZOMB)
				i = 1;
		}
	}
	if (i)
		wakeup(proc_init);

	swtch();

	/* not reached */
}

sys_wait(xloc)
	int *xloc;
{
	register struct proc *p;
	int pid;

	for (;;) {
		pid = 0;
		for (p = proc; p < &proc[NPROC]; ++p)
			if (p->p_parent == curproc) {
				pid = p->p_pid;

				if (p->p_stat == SZOMB) {
					if (xloc)
						suword(xloc, p->p_xstat);

					bzero(p, sizeof(*p));
					return (pid);
				}
			}
		if (pid == 0) {
			errno = ECHILD;
			return (-1);
		}
		if (sleep(curproc, PWAIT))
			return;
	}
}

sys_fork()
{
	register struct proc *p, *pp;
	register int i;

	p = NULL;
rescan:
	if (++mpid >= 10000)
		mpid = 100;

	for (pp = proc; pp < &proc[NPROC]; ++pp) {
		if (pp->p_stat == 0 && p == NULL)
			p = pp;
		if (pp->p_pid == mpid || pp->p_pgrp == mpid)
			goto rescan;
	}
	if (p == NULL) {
		errno = EAGAIN;
		return;
	}
	pp = curproc;
	p->p_parent = pp;

	p->p_pid  = mpid;
	p->p_stat = SIDL;
	p->p_flag = SKEEP;
	p->p_pri  = pp->p_pri;

	p->p_nice = pp->p_nice;
	p->p_pgrp = pp->p_pgrp;

	p->p_cdir = pp->p_cdir;
	p->p_cdir->i_count++;

	p->p_ctty = pp->p_ctty;
	if (p->p_ctty)
		p->p_ctty->i_count++;

	p->p_textp = pp->p_textp;
	if (p->p_textp)
		p->p_textp->x_count++;

	dupds(p, pp);

	p->p_ss = p->p_ds;

	p->p_ksp = p->p_sp = 0; /* XXX */
#if 0
	for (i = 0; i < NOFILE; ++i) {
		if ((p->p_user.u_ofile[i] = pp->p_user.u_ofile[i])) {
			p->p_user.u_ofile[i]->f_count++;
			p->p_user.u_pofile[i] = pp->p_user.u_pofile[i];
		}
	}
#endif
	p->p_flag = SLOAD;
	setrun(p);

	return (p->p_pid);
}

