
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/buf.h"
#include "../h/filsys.h"
#include "../h/mount.h"
#include "../h/dir.h"
#include "../h/proc.h"
#include "../h/inode.h"
#include "../h/file.h"
#include "../h/conf.h"
#include "../h/stat.h"

sys_fstat(fd, st)
	struct stat *st;
{
	register struct file *fp;

	fp = getf(fd);
	if (fp == NULL)
		return;

	stat(fp->f_inode, st, (fp->f_flag&FPIPE) ? fp->f_un.f_offset : (off_t) 0);
}

sys_stat(fname, st)
	char *fname;
	struct stat *st;
{
	register struct inode *ip;

	ip = namei(fname, 0);
	if (ip == NULL)
		return;

	stat(ip, st, (off_t) 0);
	iput(ip);
}

/*
 * The basic routine for fstat and stat:
 * get the inode and pass appropriate parts back.
 */
stat(ip, ub, pipeadj)
	register struct inode *ip;
	struct stat *ub;
	off_t pipeadj;
{
	struct stat st;

	st.st_dev   = ip->i_mp->m_dev;
	st.st_ino   = ip->i_number;
	st.st_mode  = ip->i_mode;
	st.st_nlink = ip->i_nlink;
	st.st_rdev  = (dev_t) ip->i_rdev;
	st.st_size  = ip->i_size - pipeadj;

	if (copyout(&st, ub, sizeof(st)))
		errno = EFAULT;
}

sys_dup2(old, new)
{
	struct file *fp;

	fp = getf(old);
	if (fp == NULL)
		return (-1);

	if (new < 0) {
		if ((new = ufalloc()) < 0)
			return (-1);
	} else {
		if (new >= NOFILE) {
			errno = EBADF;
			return (-1);
		}
	}
	if (new != old) {
		fp->f_count++;
		closef(u.u_ofile[new]);
		u.u_ofile[new] = fp;
		u.u_pofile[new] = 0;
	}
	return (new);
}

sys_mount(fspec, freg, ronly)
	char *fspec, *freg;
{
	dev_t dev;
	register struct mount *mp;
	register struct filsys *fp;
	struct inode *ip;
	struct mount *smp;
	struct buf *bp;

	dev = getmdev(fspec);
	if (errno)
		return;

	ip = namei(freg, 0);
	if (ip == NULL)
		return;

	if (ip->i_count != 1
	 || (ip->i_mode & IFMT) != IFDIR
	 || (ip->i_flag & IMOUNT))
		goto out;

	smp = NULL;
	for (mp = mount; mp < &mount[NMOUNT]; ++mp) {
		if (mp->m_dev == dev)
			goto out;

		if (mp->m_dev == NODEV && smp == NULL)
			smp = mp;
	}
	if ((mp = smp) == NULL)
		goto out;

	(*bdevsw[major(dev)].d_open)(dev, ronly ? FREAD : (FREAD | FWRITE));
	if (errno)
		goto out;
	bp = bread(dev, SUPERB);
	if (errno) {
		brelse(bp);
		goto out1;
	}
	mp->m_mpoint = ip;
	mp->m_dev = dev;
	fp = &mp->m_filsys;
	bcopy(bp->b_un.b_addr, fp, sizeof(*fp));
	brelse(bp);

	fp->s_ilock = 0;
	fp->s_flock = 0;
	fp->s_ronly = ronly ? 1 : 0;
	fp->s_lasti = 1;
	fp->s_nbehind = 0;

	ip->i_flag |= IMOUNT;
	prele(ip);
	return;

out:
	errno = EBUSY;
out1:
	iput(ip);
}

sys_umount(fspec)
	char *fspec;
{
	register struct inode *ip;
	register struct mount *mp;
	dev_t dev;

	dev = getmdev(fspec);
	if (errno)
		return;

	for (mp = mount; mp < &mount[NMOUNT]; ++mp)
		if (mp->m_dev == dev)
			goto found;

	errno = EINVAL;
	return;

found:
	for (ip = inode; ip < &inode[NINODE]; ++ip)
		if (ip->i_mp == mp) {
			errno = EBUSY;
			return;
		}

	update();

	(*bdevsw[major(dev)].d_close)(dev, 0);

	mp->m_dev = NODEV;
	ip = mp->m_mpoint;
	ip->i_flag &= ~IMOUNT;
	plock(ip);
	iput(ip);
}

/*
 * Common code for mount and umount.
 * Check that the user's argument is a reasonable
 * thing on which to mount, and return the device number if so.
 */
dev_t
getmdev(fspec)
	char *fspec;
{
	register struct inode *ip;
	dev_t dev;

	ip = namei(fspec, 0);
	if (ip == NULL)
		return (NODEV);

	if ((ip->i_mode & IFMT) != IFBLK)
		errno = ENOTBLK;

	dev = (dev_t) ip->i_rdev;
	if (major(dev) >= nblkdev)
		errno = ENXIO;

	iput(ip);

	return (dev);
}

