
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/dir.h"
#include "../h/proc.h"
#include "../h/inode.h"
#include "../h/clock.h"

sys_sync()
{
	update();
}

sys_getpid()
{
	return (curproc->p_pid);
}

sys_getppid()
{
	return (curproc->p_parent->p_pid);
}

sys_nice(niceness)
{
	register n = niceness + curproc->p_nice;

	if (n > 2*NZERO -1)
		n = 2*NZERO -1;
	if (n < 0)
		n = 0;

	curproc->p_nice = n;
}

sys_pause()
{
	sleep(&u, PSLEP);
}

sys_settimeofday(ut)
	struct timeval *ut;
{
	struct timeval tv;
	unsigned b;

	if (copyin(ut, &tv, sizeof(tv))) {
		errno = EFAULT;
		return;
	}
	b = tv.tv_usec * HZ / 1000000L;

	splclock();
	time = tv.tv_sec;
	lbolt = b;
	spl0();
}

sys_gettimeofday(ut)
	struct timeval *ut;
{
	struct timeval tv;
	unsigned b;

	splclock();
	tv.tv_sec = time;
	b = lbolt;
	spl0();

	while (b >= HZ) {
		b -= HZ;
		++tv.tv_sec;
	}
	tv.tv_usec = 1000000L / HZ * b;

	if (copyout(&tv, ut, sizeof(tv)))
		errno = EFAULT;
}

sys_alarm(sec)
{
	register struct proc *p = curproc;
	int old;

	old = p->p_clktim;
	p->p_clktim = sec * HZ;

	return (old);
}

sys_signal(sig, handler)
{
	register struct proc *p = curproc;
	int old;

	if (sig <= 0
	 || sig >= NSIG
	 || sig == SIGKILL) {
		errno = EINVAL;
		return;
	}
	old = p->p_signal[sig];

	p->p_signal[sig] = handler;
	p->p_sig &= ~sigmask(sig);

	return (old);
}

sys_kill(pid, sig)
{
	register struct proc *p;
	int err = ESRCH;

	for (p = proc; p < &proc[NPROC]; ++p) {
		if (p->p_stat) {
			if (pid == -1)
				if (p == proc_init || p->p_pgrp != curproc->p_pgrp)
					continue;
			else
				if (p->p_pid != pid)
					continue;
			if (sig)
				psignal(p, sig);
			err = 0;
		}
	}
	errno = err;
}

