
#include "../h/param.h"
#include "../h/systm.h"
#include "../h/errno.h"
#include "../h/dir.h"
#include "../h/proc.h"
#include "../h/file.h"
#include "../h/conf.h"
#include "../h/tty.h"
#include "../h/ioctl.h"
#include "../h/buf.h"  /* just iomove B_WRITE */

#define	OBUFSIZ	100
#define CANBSIZ 100

/*
 * routine called on first teletype open.
 * establishes a process group for distribution
 * of interrupts from the tty.
 */
ttyopen(tp, dev, f_flag, istty)
	register struct tty *tp;
	dev_t dev;
	int *istty;
{
	register struct proc *p = curproc;

	*istty |= 1; /* offer 'controlling terminal' */

	if (p->p_pgrp == 0) {
		if (tp->t_pgrp == 0)
			tp->t_pgrp = p->p_pid;
		p->p_pgrp = tp->t_pgrp;
	}
	tp->t_state &= ~WOPEN;
	tp->t_state |= ISOPEN;
}

/*
 * clean tp on last close
 */
ttyclose(tp)
	register struct tty *tp;
{
	tp->t_pgrp = 0;
	wflushtty(tp);
	tp->t_state = 0;
}

/*
 * set default control characters.
 */
ttychars(tp)
	register struct tty *tp;
{
	tp->t_erase = CERASE;
	tp->t_kill  = CKILL;
	tp->t_intrc = CINTR;
}

/*
 * Common code for several tty ioctl commands
 */
ttioccom(tp, cmd, arg, f_flag)
	struct tty *tp;
{
	unsigned t;
	struct ttiocb iocb;

	switch (cmd) {

	/*
	 * parameters
	 */
	case TIOCGETP:
		iocb.ioc_speed = tp->t_speed;
		iocb.ioc_flags = tp->t_flags;
		iocb.ioc_erase = tp->t_erase;
		iocb.ioc_kill  = tp->t_kill;
		if (copyout(&iocb, arg, sizeof(iocb)))
			errno = EFAULT;
		break;

	case TIOCSETP:
		wflushtty(tp);
	case TIOCSETN:
		if (copyin(arg, &iocb, sizeof(iocb))) {
			errno = EFAULT;
			break;
		}
		spltty();
		while (canon(tp) == 0) 
			;
		tp->t_speed = iocb.ioc_speed;
		tp->t_flags = iocb.ioc_flags;
		tp->t_erase = iocb.ioc_erase;
		tp->t_kill  = iocb.ioc_kill;
		spl0();
		break;

	/*
	 * prevent more opens on channel
	 */
	case TIOCEXCL:
		tp->t_state |= XCLUDE;
		break;

	case TIOCNXCL:
		tp->t_state &= ~XCLUDE;
		break;

	/*
	 * hang up line on last close
	 */
	case TIOCHPCL:
		tp->t_state |= HUPCLS;
		break;

	/*
	 * discard all queued data
	 */
	case TIOCFLUSH:
		flushtty(tp);
		break;

	default:
		return (0);
	}
	return (1);
}

/*
 * Wait for output to drain, then flush input waiting.
 */
wflushtty(tp)
	register struct tty *tp;
{

	spltty();
	while (tp->t_outq.c_cc && tp->t_state&CARR_ON) {
		(*tp->t_oproc)(tp, 1);
		tp->t_state |= ASLEEP;
		if (sleep(&tp->t_outq, TTOPRI))
			return;
	}
	flushtty(tp);
	spl0();
}

/*
 * flush all TTY queues
 */
flushtty(tp)
	register struct tty *tp;
{
	register s;

	while (getc(&tp->t_canq) >= 0)
		;
	wakeup(&tp->t_rawq);
	wakeup(&tp->t_outq);

	s = spltty();
	tp->t_state &= ~TTSTOP;
	(*tp->t_oproc)(tp, 0);

	while (getc(&tp->t_outq) >= 0)
		;
	while (getc(&tp->t_rawq) >= 0)
		;
	tp->t_delct = 0;
	splx(s);
}

/*
 * transfer raw input list to canonical list,
 * doing erase-kill processing and handling escapes.
 *
 * Wait until a full line has been typed in cooked mode,
 * or until any character has been typed in raw mode.
 */
canon(tp)
register struct tty *tp;
{
	register char *bp;
	register int c;
	int s;
	char canonb[2 + CANBSIZ];

	if (tp->t_flags&(RAW|CBREAK)) {
		if (tp->t_rawq.c_cc == 0)
			return (-1);
	} else {
		if (tp->t_delct == 0)
			return (-1);
	}
	s = spl0();
loop:
	bp = &canonb[2];
	while ((c = getc(&tp->t_rawq)) >= 0) {
		if ((tp->t_flags&(RAW|CBREAK))==0) {
			if (c == 0377) {
				tp->t_delct--;
				break;
			}
			if (bp[-1] != '\\') {
				if (c == tp->t_erase) {
					if (bp > &canonb[2])
						bp--;
					continue;
				}
				if (c == tp->t_kill)
					goto loop;
				if (c == CEOT)
					continue;
			} else {
				if (c==tp->t_erase || c==tp->t_kill)
					bp--;
			}
		}
		*bp++ = c;
		if (bp >= &canonb[CANBSIZ])
			break;
	}
	b_to_q(&canonb[2], bp - &canonb[2], &tp->t_canq);

	if (tp->t_state&TBLOCK && tp->t_rawq.c_cc < TTYHOG/5) {
		if (putc(CSTART, &tp->t_outq)==0) {
			tp->t_state &= ~TBLOCK;
			ttstart(tp);
		}
	}
	splx(s);

	return (0);
}

/*
 * Place a character on raw TTY input queue, putting in delimiters
 * and waking up top half as needed.
 * Also echo if required.
 */
ttyinput(tp, c)
	register struct tty *tp;
	register unsigned char c;
{
	if (tp->t_flags&TANDEM)
		ttyblock(tp);

	if ((tp->t_flags&RAW)==0) {
		if (tp->t_state&TTSTOP) {
			if (c == CSTOP)
				return;
			tp->t_state &= ~TTSTOP;
			ttstart(tp);
			if (c == CSTART)
				return;
		} else {
			if (c == CSTART)
				return;
			if (c == CSTOP) {
				tp->t_state |= TTSTOP;
				(*tp->t_oproc)(tp, 0);
				return;
			}
		}
		if (c == tp->t_intrc) {
			flushtty(tp);
			ttysignal(tp->t_pgrp, SIGINT);
			return;
		}
		if (c=='\r' && tp->t_flags&CRMOD)
			c = '\n';
	}
	if (tp->t_rawq.c_cc > TTYHOG) {
		flushtty(tp);
		return;
	}
	putc(c, &tp->t_rawq);

	if (tp->t_flags&ECHO) {
		ttyoutput(c, tp);
		if (c == tp->t_kill && (tp->t_flags&(RAW|CBREAK))==0)
			ttyoutput('\n', tp);
		ttstart(tp);
	}
	if ((tp->t_flags&(RAW|CBREAK)) || (c=='\n'||c==CEOT)) {
		if ((tp->t_flags&(RAW|CBREAK))== 0 && putc(0377, &tp->t_rawq)==0)
			tp->t_delct++;

		wakeup(&tp->t_rawq);
	}
}


/*
 * Send stop character on input overflow.
 */
ttyblock(tp)
	register struct tty *tp;
{
	register x = tp->t_rawq.c_cc + tp->t_canq.c_cc;

	if (tp->t_rawq.c_cc > TTYHOG) {
		flushtty(tp);
		tp->t_state &= ~TBLOCK;
	}
	if (x >= TTYHOG / 2) {
		if (putc(CSTOP, &tp->t_outq) == 0) {
			tp->t_state |= TBLOCK;
			ttstart(tp);
		}
	}
}

/*
 * put character on TTY output queue,
 * expanding tabs, and handling the CR/NL bit.
 * It is called both from the top half for output,
 * and from interrupt level for echoing.
 */
ttyoutput(c, tp)
	struct tty *tp;
{
	if ((tp->t_flags&RAW) == 0) {
		if (c < ' ') {
			switch (c) {
			case '\b':
				if (tp->t_col)
					tp->t_col--;
				break;

			case '\n':
				if (tp->t_flags&CRMOD)
					putc('\r', &tp->t_outq);
			case '\r':
				tp->t_col = 0;
				break;

			case '\t':
				if (tp->t_flags&XTABS) {
					do {
						putc(' ', &tp->t_outq);
					} while (++tp->t_col & 7);
					return;
				}
				tp->t_col = (tp->t_col | 7) + 1;
				break;
			}
		} else
			tp->t_col++;
	}
	putc(c, &tp->t_outq);
}

/*
 * Restart typewriter output following a delay
 * timeout.
 * The name of the routine is passed to the timeout
 * subroutine and it is called during a clock interrupt.
 */
ttrstrt(tp)
	register struct tty *tp;
{

	tp->t_state &= ~TIMEOUT;
	ttstart(tp);
}

/*
 * Start output on the typewriter. It is used from the top half
 * after some characters have been put on the output queue,
 * from the interrupt routine to transmit the next
 * character, and after a timeout has finished.
 */
ttstart(tp)
	register struct tty *tp;
{
	register s;

	s = spltty();
	if ((tp->t_state&(TIMEOUT|TTSTOP|BUSY)) == 0)
		(*tp->t_oproc)(tp, 1);
	splx(s);
}

/*
 * Called from device's read routine after it has
 * calculated the tty-structure given as argument.
 */
ttyread(tp)
	register struct tty *tp;
{
	int s;

	if ((tp->t_state&CARR_ON)==0)
		return (-1);

	s = spltty();
	if (tp->t_canq.c_cc == 0)
		while (canon(tp))
			if (sleep(&tp->t_rawq, TTIPRI))
				break;
	splx(s);

	while (u.u_count && tp->t_canq.c_cc) {
		if (subyte(u.u_base, getc(&tp->t_canq))) {
			errno = EFAULT;
			break;
		}
		u.u_base++;
		u.u_count--;
	}
	return (tp->t_rawq.c_cc + tp->t_canq.c_cc);
}

static void
tthiwat(tp)
	register struct tty *tp;
{
	spltty();
	while (tp->t_outq.c_cc > TTHIWAT) {
		ttstart(tp);
		tp->t_state |= ASLEEP;
		sleep(&tp->t_outq, TTOPRI);
	}
	spl0();
}

/*
 * Called from the device's write routine after it has
 * calculated the tty-structure given as argument.
 */
ttywrite(tp)
	register struct tty *tp;
{
	register char *cp;
	register int cc, i;
	char obuf[OBUFSIZ];

	if ((tp->t_state&CARR_ON)==0)
		return;

	while (cc = u.u_count) {
		if (cc > OBUFSIZ)
			cc = OBUFSIZ;
		iomove(obuf, cc, B_WRITE);
		if (errno)
			break;
		cp = obuf;
		if (tp->t_flags&RAW) {
			do {
				tthiwat(tp);
				i = b_to_q(cp, cc, &tp->t_outq);
				cp += cc-i;
				cc = i;
			} while (cc);
		} else {
			tthiwat(tp);
			do {
				ttyoutput(*cp++, tp);
			} while (--cc);
		}
	}
	ttstart(tp);
}

