
#include "head.h"
#include "pcb.h"
#include "asic.h"

#undef putc

static void
putc(c)
{
	while (inb(UART_1+USTAT) & 8)
		;
	outb(UART_1+UDATA, c);
}

static void
putb(b)
	byte b;
{
	putc(hexdig[b / 16]);
	putc(hexdig[b % 16]);
}

static void
putw(w, sep)
	word w;
{
	putb(w / 256);
	putb(w % 256);
	putc(sep);
}

static void
puts(s)
	byte *s;
{
	while (*s)
		putc(*s++);
}

static byte *names[] = {
	"divide error",
	"single step",
	"nmi",
	"breakpoint",
	"arithmetic overflow",
	"bounds check",
	"invalid opcode",
	"floating escape",
	"numerics",
	"reserved interrupt",
	"unknown interrupt",
};

void
dump(exc, ss, frame)
	word exc, ss;
	struct iregs *frame;
{
#define PEEK(r) peekw(ss, &frame->r)

	int i;
	word ww,
		bp = PEEK(bp),
		sp = PEEK(sp) + 6; /* skip ip cs flags */

#define PUTW(r, sep) ww = PEEK(r); putw(ww, sep) /* avoid a(b(c,d),e) bug */

	puts("\r\n");
	PUTW(cs, ':'); PUTW(ip, ' '); PUTW(flags, ' ');

	putw(ss, ':'); putw(sp, ' '); /* ss:sp before exception */
	putw(bp, ' ');

	PUTW(ds, ' '); PUTW(es, ' ');
	PUTW(ax, ' ');
	PUTW(bx, ' ');
	PUTW(cx, ' ');
	PUTW(dx, ' ');
	PUTW(si, ' ');
	PUTW(di, ' ');
	puts("\r\n");

	/* dozen or so words from stack, less if it wraps */

	for (i = 0; i < 14; ++i) {
		if ((sp & ~1) == 0)
			break;

		ww = peekw(ss, sp);
		putw(ww, ' ');
		sp += 2;
	}
	puts("\r\n");

	/* function trace (assuming near calls) */

	for (i = 0; i < 14; ++i) {

		ww = peekw(ss, bp + 2);
		putw(ww, ' ');
		bp = peekw(ss, bp);
		if (bp == 0)
			break;
	}
	if (exc < __SIZEOF(names)) {
		puts("\r\n");
		puts(names[exc]);
	}
}

dump_asic()
{
	int i;
	int n = 0;

	for (i = 0x00; i < 0x40; i++) {
		if (n == 0) {
			putb(i);
			putc(':');
		}
		putc(' ');
		putb(inb(i));

		if (++n >= 16) {
			n = 0;
			puts("\r\n");
		}
	}
	puts("\r\n");
}

