
#include "head.h"

static int
	program_29f(),
	program_29c(),
	program_28();

static
struct chip {
	word id, bshift;
	int (*program)();
	byte *name;            /* vanity. */
} chips[] = {
	{ 0x1FDA,  8, program_29c, "AT29C020"  }, /* default */
	{ 0x1FD5,  7, program_29c, "AT29C010A" },
	{ 0x1F5B,  9, program_29c, "AT29C040"  },
	{ 0x1FA4,  8, program_29c, "AT29C040A" },
	{ 0xDAC1,  7, program_29c,  "W29C011A" },
	{ 0xDA45,  7, program_29c,  "W29C020"  },
	{ 0xDA46,  8, program_29c,  "W29C040"  },

	{ 0x0120, 14, program_29f, "Am29F010"  },
	{ 0x01A4, 16, program_29f, "Am29F040"  },
};

#define getc getc_1

#define TEMP  0x2000
#define FINAL 0xC000

#define SAFE  (30 * HZ) /* watchdog */

static dword length;
static byte worm, cksum;
static jmp_buf garbage;

static void
store(to, buf, count)
	dword to;
	byte *buf;
	byte count;
{
	word seg, off;

	/* XXX require sequential */

	if (to != length)
		longjmp(garbage, 1);

	while (count--) {
		off = to & 15;
		seg = to >> 4;
		pokeb(TEMP + seg, off, *buf);
		to++;
		buf++;
	}
	length = to;

	putchar(" \ryum yum\r      "[++worm & 15]);
}

static byte
anybble()
{
	int c;

	while ((c = getc()) == -1)
		;

	if (_isdigit(c))
		return (c - '0');

	if (c >= 'A' && c <= 'F')
		return (c - 'A' + 10);

	if (c >= 'a' && c <= 'f')
		return (c - 'a' + 10);

	longjmp(garbage, 2);
}

static byte
abyte()
{
	byte b;

	b = anybble() << 4;
	b |= anybble();

	cksum += b;

	return (b);
}

static int
load_ihex()
{
	int i;
	byte type;
	word count, offset;
	dword base;
	byte buf[256];

	if (setjmp(garbage)) {
bad:
		return (-1);
	}
	base = 0;
	worm = 0;     /* silly */

	for (;;) {
		fido = SAFE;

		if ((i = getc()) == -1 || _isspace(i))
			continue;

		if (i != ':')
			goto bad;

		cksum = 0;

		count = abyte();
		offset = abyte() << 8;
		offset |= abyte();
		type = abyte();

		for (i = 0; i < count; ++i)
			buf[i] = abyte();

		abyte();         /* checksum byte */

		if (cksum)
			goto bad;

		switch (type) {
		default:
			goto bad;

		/* data */
		case 0x00:
			store(base + offset, buf, count);
			break;

		/* segmented/linear base */
		case 0x02:
		case 0x04:
			if (offset || count != 2)
				goto bad;
			base = buf[1] | (buf[0] << 8);
			base <<= 4;
			if (type == 0x04)
				base <<= 12;
			break;

		/* segmented/linear entrypoint */
		case 0x03:
		case 0x05:
			break;

		/* eof */
		case 0x01:
			if (offset || count != 0)
				goto bad;
			return (0);
		}
	}
}

void
regenerate()
{
	struct chip *p;
	word id, blk;

	id = identify();

	for (p = chips; p->id != id; ) {
		if (++p == &chips[__SIZEOF(chips)]) {
			printf("0x%x unknown, assuming ", id);

			p = chips; /* XXX don't assume, ASK! */
			break;
		}
	}
	printf("%s", p->name);

	blk = 1 << p->bshift;

	/* eat up junk from previous try.
	 * continue in 3 seconds. */

again:
	fido = SAFE + 3 * HZ;

	do {
		if (getc() != -1)
			goto again;

	} while (fido > SAFE);

	printf("\r\nWaiting intel hex ");

	length = 0;

	if (load_ihex()) {
		printf("\r\nHuh? Try again. ");
		goto again;
	}
	if (length == 0
	 || length > 0x40000) {
		printf("\r\nStrange length. ");
		goto again;
	}
	fido = SAFE;

	printf("\r%lu bytes. Writing ", length);

	if ((*p->program)(blk, (length + (blk - 1)) >> p->bshift, TEMP, FINAL)
	 || compare(length, TEMP, FINAL)) {

		printf("FAILED.\r\nStay calm, don't power off. Try again. ");
		goto again;
	}
	printf("OK. ");

	reboot();
}

#asm
	.text

_identify:
	push ds
	mov ax, # 0xE000
	mov ds, ax
	pushf
	cli                        ; soita paranoid
	movb [0x5555], # 0xAA
	movb [0x2AAA], # 0x55
	movb [0x5555], # 0x90      ; ident
	mov ax, [0]                ; ah from [1]
	movb [0x5555], # 0xAA
	movb [0x2AAA], # 0x55
	movb [0x5555], # 0xF0      ; idle
	popf
	pop ds
	xchg al, ah                ; manufacturer to ah, device id to al
	ret

;----------------------------------------------------------------------

_program_29f:
	push ds
	push es
	push bp                   ; 0  2  4  6  8               12  14
	mov bp, sp                ; bp es ds ra sector_count_32 src dst
	push ax                   ; -2 sector size, 0 means 0x10000

	cmp ax, # 1               ; carry if 0x10000
	rcr ax, # 1
	shr ax, # 3
	push ax                   ; -4 sector size in paragraphs

	mov si, [bp+12]           ; source segment
	mov di, [bp+14]           ; destination
fmore:
	eseg                           ; es stays at _DS
	mov [_fido], # SAFE + 10 * HZ  ; single timeout for erase + writes

	push # 0xE000             ; make offset == physical, A15 and up dont care
	pop ds
	movb [0x5555], # 0xAA     ; erase ...
	movb [0x2AAA], # 0x55
	movb [0x5555], # 0x80
	movb [0x5555], # 0xAA
	movb [0x2AAA], # 0x55

	mov ds, di                ; destination segment
	xor bx, bx                ; index

	movb [bx], # 0x30         ; erase this sector
wait_erase:
	eseg
	cmp [_fido], # SAFE
	jb failed
	cmpb [bx], # 0xFF
	jne wait_erase

	mov cx, [bp-2]            ; sector size, 0 means 0x10000
wrloop:
	push # 0xE000
	pop ds
	movb [0x5555], # 0xAA     ; program ...
	movb [0x2AAA], # 0x55
	movb [0x5555], # 0xA0

	mov ds, si                ; source segment
	mov al, [bx]
	mov ds, di                ; destination segment
	mov [bx], al              ; write

wait_write:
	eseg
	cmp [_fido], # SAFE
	jb failed
	cmp [bx], al
	jne wait_write

	inc bx
	loop wrloop               ; do one sector

	add si, [bp-4]            ; segments forward, one sector
	add di, [bp-4]
	dec [bp+8]                ; sector count (LSWord is enough XXX)
	jnz fmore

	xor ax, ax
	leave
	pop es
	pop ds
	ret

failed:
	mov al, # 1
	leave
	pop es
	pop ds
	ret

_program_29c:
	push ds
	push es
	push bp                   ; 0  2  4  6  8             12  14
	mov bp, sp                ; bp es ds ra page_count_32 src dst
	push ax                   ; -2 page size

	shr ax, # 4
	push ax                   ; -4 page size in paragraphs

	mov dx, [bp+12]           ; source segment
	mov bx, [bp+14]           ; destination
more:
	push # 0xE000             ; make offset == physical, A15 and up dont care
	pop ds

	pushf
	cli
	movb [0x5555], # 0xAA     ; program ...
	movb [0x2AAA], # 0x55
	movb [0x5555], # 0xA0

	mov ds, dx                ; source segment
	mov es, bx                ; destination
	xor si, si                ; aligned to paragraphs
	xor di, di
	mov cx, [bp-2]            ; page size
	rep
	movsb                     ; write
	popf

	xor ax, ax
byte_poll:                    ; 10 msec
	push ax
	pop ax
	dec ax                    ; XXX use _fido ?
	jz failed
	dec si
	dec di
	cmpsb
	jne byte_poll

	add dx, [bp-4]            ; segments forward, one page
	add bx, [bp-4]
	dec [bp+8]                ; page count (LSWord is enough XXX)
	jnz more

	xor ax, ax
	leave
	pop es
	pop ds
	ret

#if 0

_program_28:
	push ds
	push es
	push bp                   ; 0  2  4  6  8         12  14
	mov bp, sp                ; bp es ds ra length_32 src dst

	xor bx, bx                ; index
	jmp ebot
emore:
	mov ds, [bp+12]           ; source segment
	mov dl, [bx]

	mov ds, [bp+14]           ; destination
	mov [bx], dl              ; write

	xor ax, ax
wait_ee:
	push ax
	pop ax
	dec ax                    ; XXX use _fido ?
	jz failed
	cmp [bx], dl
	jne wait_ee

	inc bx
	jnz ebot
	add [bp+12], # 0x1000     ; hmph.
	add [bp+14], # 0x1000
ebot:
	sub [bp+8], # 1           ; --length_32
	sbb [bp+10], # 0
	jnc emore

	xor ax, ax
	leave
	pop es
	pop ds
	ret

#endif

;----------------------------------------------------------------------

_compare:             ; int compare(dword cc, word seg1, word seg2)
	push ds
	push es
	push bp
	mov bp, sp        ; bp es ds ra seg1 seg2

	mov ds, [bp+8]    ; seg1
	mov es, [bp+10]   ; seg2
	xor si, si        ; aligned to paragraphs
	xor di, di

	or bx, bx         ; cc msword
	je last           ; less than 0x10000 bytes
lots:
	mov cx, # 0x8000  ; 64kB
	rep
	cmpsw
	jne diff

	mov cx, ds
	add cx, # 0x1000  ; 64kB
	mov ds, cx
	mov cx, es
	add cx, # 0x1000
	mov es, cx
	dec bx
	jne lots          ; still more than 0x10000 bytes left
last:
	mov cx, ax        ; cc lsword
	xor ax, ax        ; also Z, in case cx=0
	rep
	cmpsb
	je same
diff:
	mov al, # 1       ; anything nonzero, if different
same:
	leave
	pop es
	pop ds
	ret

