/*
 *  DT3 (DT3000?) on D15
 *
 *  40 character columns, 240 vertical pixels
 *
 *  16257000 / 2 = 8.128500 MHz
 *  / 80 / 8 =  12.7 kHz
 *  / 240 = 52 Hz
 *
 *  40 usec for display of 40 bytes ?
 *  4 pixels go at once ?
 *
 *  fonts 8x15
 *  custom font.bin to chargen rom
 */

#include "head.h"
#include "pcb.h"
#include "crt.h"

#define LINES 16
#define COLS  40

#define GXATTR 7 /* attribute for 8x1 pixel raster */

struct file Screen = { putc_screen, getc_kbd };

void
crt_init()
{
	clear_screen();

#define EXTRA 16  /* 75 Hz, used in original sw */
#define EXTRA 40  /* 52 Hz, cannot see any difference */

	outb_crtc(HCHR_TOT, COLS + EXTRA - 1); /* Nht - 1 */
	outb_crtc(HCHR_DISPLAY, COLS);         /* Nhd */
	outb_crtc(HCHR_VIRTUAL, COLS);         /* Nir */

	outw_crtc(VPIXELS, 240 - 1);   /* Nd - 1 */
	outw_crtc(CURSOR_ADDR, 0);
	outw_crtc(START_ADDR, 0);
	outb_crtc(START_RASTER, 0);    /* Nsr */

	outb_crtc(CURSOR_START, 13 | CURSOR_ON); /* Ncs + blink mode */
	outb_crtc(CURSOR_LAST, 14);              /* Nce */
	outb_crtc(CHR_LAST, 14);                 /* Nr  */

	outb_crtc(MODE, MODE_AT | MODE_ON); /* AT is hardwired anyway */
}

#asm
	include asm.inc
	.text

;----------------------------------------------------------------------

_outb_crtc:           ; void outb_crtc(reg, value);
	ARG2 ah

	mov dx, # CRTC
	out dx, ax        ; ARG1 to address reg, ah to value register
	ret

_outw_crtc:           ; void outw_crtc(reg, word value);
	ARG2 bx           ; word value
point:
	mov ah, bh        ; al/ah, write MSByte
	mov bh, bl
	mov bl, al
	inc bl            ; bl/bh, write LSByte
	mov dx, # CRTC
	pushf             ; XXX not this way
	cli
	out dx, ax
	mov ax, bx
	out dx, ax
	popf
	ret

;----------------------------------------------------------------------
; XXX not that way

	export _update_row

	export _clear_screen
	export _update_cursor

_update_row:          ; void update_row(rownum, attribute, string);
	ARG3 bl, si
	mov cx, # COLS    ; character columns
	mul cl            ; ARG1 (max 255) * cl -> ax
	mov di, ax
	push es
	mov es, screen_seg
	mov ah, bl        ; attribute
	add di, _window
	shl di, 1         ; word index to byte offset
L1:
	lodsb             ; character
	stosw             ; with attribute
	loop L1           ; each column
	pop es
	ret

_clear_screen:
	push es
	mov es, screen_seg
	xor di, di
	mov _window, di
	mov _cursor, di
	mov ax, # 0x0020            ; space and normal attribute
	mov cx, # LINES * COLS
	rep
	stosw
	pop es

	; fallthru

_update_cursor:
	mov bx, _window
	cmp bx, window_real
	jne uc_1
	add bx, _cursor
	cmp bx, cursor_real
	jne uc_2
	ret
uc_1:
	mov window_real, bx
	mov al, # START_ADDR
	call point
	mov bx, _window
	add bx, _cursor
uc_2:
	mov cursor_real, bx
	mov al, # CURSOR_ADDR
	jmp point

;----------------------------------------------------------------------
; graphics mode
;----------------------------------------------------------------------
	export _graphics_on
	export _graphics_off
	export _draw_pixel
	export _draw_line
	export _draw_box
	export _draw_pic

_draw_pic:                  ; void draw_pic(src, x,y, width,height);
	push es
	push bp
	mov bp, sp
	mov si, ax              ; wqx_bits
	mov cx, [bp+6]          ; x
	shr cx, 3               ; XXX
	imul di, [bp+8], # COLS ; y
	add di, cx
	add di, _window
	mov bx, [bp+10]         ; width
	add bx, # 7             ; XXX
	shr bx, 3
	mov bp, [bp+12]         ; and height.

	mov dx, # COLS
	sub dx, bx              ; stride
	shl di, 1
	shl dx, 1

	mov ah, # GXATTR
	mov es, screen_seg
copy:
	mov cx, bx
copy_1:
	lodsb                ; XXX unroll a bit
	stosw
	loop copy_1
	add di, dx
	dec bp
	jne copy
	pop bp
	pop es
	ret

_draw_box:                  ; void draw_box(pattern, x,y, width,height);
	push es
	push bp
	mov bp, sp
	mov si, ax              ; pattern
	mov cx, [bp+6]          ; x
	shr cx, 3               ; XXX
	imul di, [bp+8], # COLS ; y
	add di, cx
	add di, _window
	mov bx, [bp+10]         ; width
	shr bx, 3               ; XXX
	mov bp, [bp+12]         ; and height.

	mov dx, # COLS
	sub dx, bx              ; stride
	shl di, 1
	shl dx, 1

	mov ax, si              ; pattern in al
	shr si, 8               ; msbyte is pattern rotation
	mov ah, # GXATTR
	mov es, screen_seg
fill:
	mov cx, bx
	rep
	stosw
	add di, dx
	mov cx, si          ; pattern rotation
	rol al, cl          ; 000, 0FF, 155, 233, 211 and 401 look nice
	dec bp
	jne fill
	pop bp
	pop es
	ret

_draw_pixel:                 ; draw_pixel(op, x,y)
	push ds
	push bp
	mov bp, sp               ; bp ds ra
	mov dx, ax               ; op
	mov cx, [bp+6]           ; x
	mov bx, cx
	and bx, # 7              ; x % 8
	shr cx, 3
	imul di, [bp+8], # COLS  ; y to beginning of line
	add di, _window
	add di, cx               ; x / 8
	mov al, _pixmask[bx]
	mov ds, screen_seg
	call dx                  ; _pixel_xxx()
	pop bp
	pop ds
	ret

_draw_line:            ; draw_line(op, x0,y0, x1,y1);
	push ds
	push bp
	mov bp, sp         ; bp ds ra
	push ax            ; _pixel_xxx()

	mov ax, [bp+6]     ; x0 line from ax,di
	mov di, [bp+8]     ; y0
	mov dx, [bp+10]    ; x1 to dx,bx
	mov bx, [bp+12]    ; y1

	mov cx, bx         ; y1
	sub cx, di         ; y1 - y0 is delta y
	jns dl_1           ; y1 > y0
	neg cx             ; abs delta y (major ?)
	mov di, bx         ; smaller y to di, y moves always down
	xchg ax, dx        ; swap x
dl_1:
	mov si, # dl_tab   ; x moves right

	sub dx, ax         ; x1 - x0 is delta x
	jns dl_2
	neg dx             ; abs delta x
	add si, # 2        ; x moves left, not right
dl_2:
	mov bx, ax         ; x0
	and bx, # 7        ; x0 % 8
	shr ax, 3          ; x0 / 8

	imul di, di, # COLS
	add di, _window
	add di, ax           ; first byte here
	mov al, _pixmask[bx] ; this bit

	cmp cx, dx         ; delta y - delta x
	je diagonal
	ja y_wise

	xchg dx, cx        ; delta x in cx (major), delta y in dx
	add si, # 4        ; going x-wise
y_wise:
	add si, # 4        ; going y-wise
	push ax            ; save the bitmask in al
	xor ax, ax         ; smaller delta * 0x10000
	div cx
	mov dx, ax         ; delta ratio in 0.16 format
	mov bx, # 0x8000   ; error accu
	pop ax             ; bitmask back
diagonal:
	mov si, [si]       ; from dl_tab
	mov ds, screen_seg
	pop bp             ; _pixel_xxx()
	call bp            ; plot first pixel

	jcxz dl_out        ; major count 0. an expensive put_pixel.
	call si            ; plot rest
dl_out:
	pop bp
	pop ds
	ret

;------------------------------------------

	.data
dl_tab:
	.word diag_right
	.word diag_left
	.word y_right
	.word y_left
	.word x_right
	.word x_left

	.text

; XXX can do better

diag_right:
	ror al, 1          ; one right
	adc di, # COLS     ; one down, byte right if carry
	call bp            ; f_op
	loop diag_right
	ret

diag_left:
	rol al, 1          ; one left
	sbb di, # -COLS    ; one down, byte left if carry
	call bp            ; f_op
	loop diag_left
	ret

vertical:
	add di, # COLS
	call bp
	loop vertical
	ret

y_right:
	add bx, dx         ; accumulate error
	jnc dl1
	ror al, 1          ; one right
dl1:
	adc di, # COLS     ; one down, byte right if carry from ror
	call bp            ; f_op
	loop y_right
	ret

y_left:
	add bx, dx         ; accumulate error
	jnc dl2
	rol al, 1          ; one left
dl2:
	sbb di, # -COLS    ; one down, byte left if carry from rol
	call bp            ; f_op
	loop y_left
	ret

	; mov ah, [di] XXX cache pixels
	; mov [di], ah
x_right:
	add bx, dx         ; accumulate error
	jnc dl3
	add di, # COLS     ; one down
dl3:
	ror al, 1          ; one right
	adc di, # 0        ; byte right if carry
	call bp            ; f_op
	loop x_right
	ret

x_left:
	add bx, dx         ; accumulate error
	jnc dl4
	add di, # COLS     ; one down
dl4:
	rol al, 1          ; one left
	sbb di, # 0        ; byte left if carry
	call bp            ; f_op
	loop x_left
	ret

;------------------------------------------
; only via draw_xxx(pixel_op, ...)

	export _pixel_set
	export _pixel_clr
	export _pixel_flip
	export _pixel_dots_01
	export _pixel_dots_10

_pixel_flip:
	shl di, 1         ; XXX
	xor [di], al
	shr di, 1
	ret

_pixel_dots_10:
	mov bp, # _pixel_dots_01
_pixel_set:
	shl di, 1
	or [di], al
	shr di, 1
	ret

_pixel_dots_01:
	mov bp, # _pixel_dots_10
_pixel_clr:
	not al           ; XXX
	shl di, 1
	and [di], al
	shr di, 1
	not al
	ret

;----------------------------------------------------------------------
; text mode
;----------------------------------------------------------------------
	export _putc_screen

_putc_screen:
	call doit

recheck:
	mov ax, _cursor
	cmp ax, region_start
	jl backward              ; signed
	cmp ax, region_end
	jge forward              ; signed
	ret

backward:
	push # recheck           ; slow, if cursor is far out

	mov cx, # COLS
	add ax, cx
	mov _cursor, ax
	jmp [wind+0]             ; screen_up or region_up, cx characters

forward:
	push # recheck

	mov cx, # COLS
	sub ax, cx
	mov _cursor, ax
	jmp [wind+2]             ; screen_down or region_down, cx characters

;----------------------------------------------------------------------

screen_up:
	sub _window, cx          ; move backward cx characters
	xor di, di               ; clear above old window
	jmp erase

screen_down:
	add _window, cx          ; move forward cx characters
	mov di, # LINES * COLS   ; clear below old window
	sub di, cx
erase:
	add di, _window
	shl di, 1
	push es
	mov es, screen_seg
	mov ax, _blank           ; ah _attribute
	rep
	stosw                    ; newly open line(s) cleared
	pop es
	ret

;----------------------------------------------------------------------

region_up:
	mov di, region_end
	mov si, di
	sub si, cx
	mov bx, cx               ; characters to clear
	mov cx, si
	sub cx, region_start     ; characters to move

	dec si                   ; *--d = *--s
	dec di
	std
	jmp move

region_down:
	mov di, region_start
	mov si, di
	add si, cx
	mov bx, cx
	mov cx, region_end
	sub cx, si
move:                        ; DF set/clr, as needed
	add si, _window
	shl si, 1
	add di, _window
	shl di, 1
	push es
	push ds
	mov es, screen_seg
	mov ds, screen_seg
	rep
	movsw                    ; cx, characters to move
	pop ds

	mov cx, bx               ; bx, characters to clear
	mov ax, _blank           ; ah _attribute
	rep
	stosw
	pop es
	cld
	ret

;----------------------------------------------------------------------
; al arg0
; ah arg1
; dl arg0, 0 replaced by 1
; dh arg1, 0 replaced by 1
;----------------------------------------------------------------------
; cursor sticks to the glass

scroll_down:
	mov al, # COLS
	mul dl
	mov cx, ax              ; cx, characters to clear
	jmp [wind+0]            ; screen_up or region_up

scroll_up:
	mov al, # COLS
	mul dl
	mov cx, ax
	jmp [wind+2]            ; screen_down or region_down

;----------------------------------------------------------------------

set_region:
	cmp dl, dh
	jae reset_region            ; CSI r = "default" or height < 2
	cmp dx, # 0x100 * LINES + 1
	je reset_region             ; region == screen
	cmp dh, # LINES
	ja reset_region             ; ends below screen

	dec dl                      ; start of first line
	mov al, # COLS
	mul dl
	mov _cursor, ax             ; region home
	mov region_start, ax        ; top left of region

	mov al, # COLS              ; no DEC, end of last line
	mul dh
	mov region_end, ax

	mov [wind+0], # region_up
	mov [wind+2], # region_down
	ret

reset_region:
	mov _cursor, # 0            ; screen home

	mov region_start, # 0
	mov region_end, # LINES * COLS

	mov [wind+0], # screen_up
	mov [wind+2], # screen_down
	ret

;----------------------------------------------------------------------
; _window and _window+_cursor are word indexes, not byte offsets.
; both wrap at 0x10000 (not 0x8000), like the registers in 64645.
; video start <= video start + cursor < video start + 640, modulo 0x10000
; there are 32768 words of videoram.
;----------------------------------------------------------------------

doit:
	cmp al, # 0x20         ; control character ?
	jb control

	jmp [p_put]            ; put or escape_x or ...
put:
	mov ah, _attribute     ; printing characters here.
	push es
	les di, _cursor
	inc _cursor            ; XXX brain needed.
	add di, _window
	shl di, 1
	stosw
	pop es
	ret

control:                   ; 0 ... 0x1F always come here.
	mov bl, al
	mov bh, # 0
	shl bx, 1
	jmp control_table[bx]

;----------------------------------------------------------------------
; return
;   di  _cursor
;   cx  COLS
;   ax  line    signed, 0 for top of window
;   dx  column  0 ... COLS-1

rowcol:
	mov cx, # COLS
	mov ax, _cursor
	mov di, ax
	cwd
	idiv cx
	test dx, dx
	jns column_ok         ; 0 ... 39
	add dx, cx            ; flip -39 ... -1 to 1 ... 39
	dec ax                ; line ... -1
column_ok:
	ret                   ; XXX argh

;----------------------------------------------------------------------
; handle a few basic ANSI sequences
; characters below 0x20 do not arrive here (except ESC via p_put)

escape:
	mov [p_put], # escape_1
	ret

escape_1:
	cmp al, # 0x5B        ; [
	jne not_a_CSI

	xor ax, ax            ; ESC [ here
	mov [argc], ax
	mov [args+0], ax
	mov [args+2], ax
	mov [args+4], ax
ARGC equ 6

	mov [p_put], # escape_2
	ret

escape_2:
	cmp al, # 0x40        ; @
	jae escape_end        ;   command byte
	cmp al, # 0x30        ; 0
	jb discard            ;   "intermediate", not accepted
	cmp al, # 0x39        ; 9
	jbe arg               ;   more digits
	cmp al, # 0x3B        ; ;
	jne discard           ;   non-digit "parameter", not accepted

	incb [argc]           ;   'argc' counts semicolons, off by one at end.
	cmpb [argc], # ARGC
	jae discard           ;   too many arguments
	ret

arg:
	sub al, # 0x30        ; minimal atoi()
	mov bx, [argc]
	mov ah, args[bx]
	aad 10                ; what strange opcodes
	mov args[bx], al
	ret

escape_end:
	mov [p_put], # put    ; done after this returns

	cmp al, # last_escape
	ja outta              ;   no such command byte. done
	sub al, # 0x40        ; @
	mov bl, al
	mov bh, # 0           ; index to escape_table[]

	mov ax, [args]        ; arg0 and 1
	mov dx, ax
	sub dl, # 1
	adc dl, # 1           ; change 0 to 1
	sub dh, # 1
	adc dh, # 1           ; same for arg1

	shl bx, 1
	jmp escape_table[bx]

discard:
	mov [p_put], # upto_command_byte
	ret

upto_command_byte:
	cmp al, # 0x40        ; @
	jb outta              ;   keep discarding
	mov [p_put], # put    ;   finally.
outta:
	ret

;----------------------------------------------------------------------

not_a_CSI:
	mov [p_put], # put    ; assume will be done

	cmp al, # 0x44        ; ESC D
	je index
	cmp al, # 0x4D        ; ESC M
	je rindex
	cmp al, # 0x45        ; ESC E
	je newline
	cmp al, # 0x37        ; ESC 7
	je save
	cmp al, # 0x38        ; ESC 8
	je restore
	cmp al, # 0x63        ; ESC c
	je reset
	cmp al, # 0x31        ; ESC 1
	je _graphics_on
	cmp al, # 0x32        ; ESC 2
	je _graphics_off

	cmp al, # 0x30
	jb discard            ; ESC ( A or such, wait for the A
	ret                   ; ESC = or ESC A or such. done.

reset:
	call _graphics_off
	call reset_region
	ret                   ; XXX maybe capslock off ?

_graphics_on:
	movb _attribute, # GXATTR

	mov dx, # CRTC
	mov ax, # CURSOR_START + 0x100 * (0 | CURSOR_OFF) ; XXX
	out dx, ax
	mov ax, # CURSOR_LAST + 0x100 * 0
	out dx, ax
	mov ax, # CHR_LAST + 0x100 * 0
	out dx, ax
	ret

_graphics_off:
	movb _attribute, # 0

	mov dx, # CRTC
	mov ax, # CURSOR_START + 0x100 * (13 | CURSOR_ON)
	out dx, ax
	mov ax, # CURSOR_LAST + 0x100 * 14
	out dx, ax
	mov ax, # CHR_LAST + 0x100 * 14
	out dx, ax
	ret

save:
	mov al, _attribute
	mov save_attr, al
	mov ax, _cursor
	mov save_cursor, ax
	ret

restore:
	mov al, save_attr
	mov _attribute, al
	mov ax, save_cursor
	mov _cursor, ax
	ret

rindex:
	sub _cursor, # COLS
	ret

newline:
	push # carriage_return
index:
linefeed:                ; LF control character
	add _cursor, # COLS
	ret

carriage_return:
	call rowcol       ; cx = COLS, ax = line, dx = column, di = cursor
	sub di, dx
	mov _cursor, di
	ret

backspace:
	dec _cursor
	ret

bell:
noop:
	ret

;----------------------------------------------------------------------

tab:                  ; TAB control character
	mov dl, # 1

tab_forward:
	mov dh, # 0
	shl dx, 3         ; times 8 XXX

	mov ax, _cursor
	and ax, #~7       ; snip
	add ax, dx        ; and tab right
	mov _cursor, ax
	ret

tab_backward:
	mov dh, # 0
	shl dx, 3         ; times 8 XXX

	mov ax, _cursor
	add ax, # 7       ; one to the right, unless exactly at tabstop
	and ax, #~7       ; snip
	sub ax, dx        ; and tab left
	mov _cursor, ax
	ret

cursor_up_col0:
	push # carriage_return

cursor_up:
	mov al, # COLS
	mul dl
	sub _cursor, ax
	ret

cursor_down_col0:
	push # carriage_return

cursor_down:
	mov al, # COLS
	mul dl
	add _cursor, ax
	ret

cursor_forward:
	mov dh, # 0
	add _cursor, dx
	ret

cursor_backward:
	mov dh, # 0
	sub _cursor, dx
	ret

cursor_position:
	sub dx, # 0x0101     ; top left is 0,0
	mov al, # COLS
	mul dl
	add al, dh
	adc ah, # 0
	cmp ax, # LINES * COLS      ; for the benefit of 'resize' command
	jb within_screen
	mov ax, # LINES * COLS - 1  ; bottom right
within_screen:
	mov _cursor, ax
	ret

cursor_column:
	dec dl               ; left is 0
	mov dh, # 0
	mov bx, dx           ; new column

	call rowcol          ; cx = COLS, ax = line, dx = column, di = cursor
	sub bx, dx
	add _cursor, bx
	ret

cursor_line:
	dec dl              ; top is 0
	mov al, # COLS
	mul dl
	mov bx, ax          ; bol

	call rowcol         ; cx = COLS, ax = line, dx = column, di = cursor
	add bx, dx
	mov _cursor, bx
	ret

;----------------------------------------------------------------------

erase_character:
	mov dh, # 0
	mov cx, dx
	mov di, _cursor
	jmp near erase

erase_display:          ; 0 down 1 up 2 whole screen
	mov bl, al
	mov cx, # LINES * COLS
	mov di, _cursor
	mov dx, di
	jmp erase_

erase_line:             ; 0 eol 1 bol 2 whole line
	mov bl, al

	call rowcol         ; cx = COLS, ax = line, dx = column, di = cursor
erase_:
	cmp bl, # 1
	ja erase_all        ; 2 or more
	je erase_backward   ; 1

	sub cx, dx          ; cols - column or screen size - cursor
	jmp near erase

erase_backward:
	mov cx, dx          ; column or cursor
erase_all:
	sub di, dx          ; bol or 0
	jmp near erase

;----------------------------------------------------------------------

delete_character:
	mov dh, # 0
	mov bx, dx           ; characters to delete

	call rowcol          ; cx = COLS, ax = line, dx = column, di = cursor
	sub cx, dx           ; characters on the right
	jmp delete

delete_line:
	mov al, # COLS
	mul dl
	mov bx, ax           ; characters to delete

	call rowcol
	sub di, dx           ; current line
	mov cx, region_end
	sub cx, di           ; characters to end of window
delete:
	mov si, di
	add si, bx           ; here are
	sub cx, bx           ; characters to keep
	jmp near move        ; moves, then erases opened area

;----------------------------------------------------------------------

insert_character:
	mov dh, # 0
	mov bx, dx           ; characters to insert

	call rowcol          ; cx = COLS, ax = line, dx = column, di = cursor
	sub cx, dx           ; characters on the right
	jmp insert

insert_line:
	mov al, # COLS
	mul dl
	mov bx, ax           ; characters to insert

	call rowcol
	sub di, dx           ; current line
	mov _cursor, di      ; cursor moves to beginning of first opened line
	mov cx, region_end
	sub cx, di           ; characters to end of window
insert:
	mov si, di
	add di, bx           ; move to here
	sub cx, bx           ; characters to keep

	add si, cx           ; end
	add di, cx
	std                  ; *--d = *--s
	dec si
	dec di
	jmp near move        ; moves, then erases opened area

;----------------------------------------------------------------------

attributes:
	mov al, _attribute
	mov si, # args
attrlop:
	mov bl, [si]
	cmp bl, # last_attribute
	ja nosuch
	mov bh, # 0
	shl bx, 1
	call attribute_table[bx] ; al in/out
nosuch:
	inc si
	dec [argc]             ; ESC[m argc 0, ESC[;m argc 1
	jns attrlop

	mov bl, al
	and bl, # 7            ; 7th font slot (GXATTR) is used by graphics
	add bl, # -7
	sbb al, # 0            ; cannot have underscored alternate standout
	mov _attribute, al
	ret

;----------------------------------------------------------------------
; chargen address bits
;  1  standout
;  2  alternate font
;  4  underscore
;
; CRTC attribute bits
;  8  reverse
; 20h blink
;
; al contains _attribute
; si must be kept

attribute_0:        ; cancel all attributes
	mov al, # 0
	ret

attribute_1:        ; standout
	or al, # 0x01
	ret
attribute_22:       ; cancel standout and dim
	and al, #~0x01
	ret

attribute_11:       ; alternate font
	or al, # 0x02
	ret
attribute_10:       ; primary font
	and al, #~0x02
	ret

attribute_4:        ; underscore
	or al, # 0x04
	ret
attribute_24:       ; cancel underscore
	and al, #~0x04
	ret

attribute_7:        ; reverse video
	or al, # 0x08
	ret
attribute_27:       ; cancel reverse video
	and al, #~0x08
	ret

attribute_5:        ; slow blink
attribute_6:        ; fast blink
	or al, # 0x20
	ret
attribute_25:       ; cancel blink
	and al, #~0x20
	ret

attribute_2:        ; dim
attribute_3:        ; italic
attribute_8:        ; concealed
attribute_9:        ; reserved
attribute_12:
attribute_13:
attribute_14:
attribute_15:
attribute_16:
attribute_17:
attribute_18:
attribute_19:
attribute_20:
attribute_21:
attribute_23:
attribute_26:
	ret

;----------------------------------------------------------------------

enquiry:
	ret

device_attributes:
	push # termid
	mov ax, [_filee]
	call _fprintf
	add sp, # 2
	ret

device_status:
	cmp al, # 6
	je cursor_position_report
	ret

cursor_position_report:
	call rowcol
	inc ax
	inc dx

	push dx          ; column + 1
	push ax          ; row + 1
	push # cpr
	mov ax, [_filee]
	call _fprintf
	add sp, # 6
	ret

;----------------------------------------------------------------------
; 4  Set insert mode, new characters shove existing ones to the right
; 4  Reset insert mode, new characters replace existing ones

set_modes:
reset_modes:
	ret             ; XXX insert/replace mode is missing

tab_control:
tab_clear:
	ret             ; XXX fixed tab stops at every 8th column

;----------------------------------------------------------------------
; double 8 pixels in AL to 16 in BX

stretch:
	mov bx, # 0x8000  ; carry out when filled
s_1:
	ror al, 1         ; get bit
	rcr bx, 1         ; insert once
	sar bx, 1         ;  twice
	ror al, 1         ; sama ruotsiks
	rcr bx, 1
	sar bx, 1
	jnc s_1           ; for all 8 input bits
	ret

;----------------------------------------------------------------------
	.data

_window:      .word 0            ; absolute word index, byte address / 2
_cursor:      .word 0            ; relative to _window
screen_seg:   .word SCREEN       ; just below _cursor

window_real:  .word 0            ; for delayed hw update
cursor_real:  .word 0

_blank:       .byte 0x20
_attribute:   .byte 0x00         ; just below _blank

region_start: .word 0            ; relative to _window
region_end:   .word LINES * COLS

wind:         .word screen_up, screen_down

p_put:        .word put          ; normal output of printing characters

termid:
	.byte 033
	.asciz "[?;99c"         ; reply to Device Attributes query
cpr:
	.byte 033
	.asciz "[%d;%dR"

	.lcomm save_cursor, 2   ; word
	.lcomm save_attr,   1   ; byte

	.lcomm argc, 2          ; word
	.lcomm args, ARGC       ; bytes

control_table:
	.word noop              ; 0x00 NUL
	.word noop              ; 0x01 SOH
	.word noop              ; 0x02 STX
	.word noop              ; 0x03 ETX
	.word noop              ; 0x04 EOT
	.word enquiry           ; 0x05 ENQ
	.word noop              ; 0x06 ACK
	.word bell              ; 0x07 BEL
	.word backspace         ; 0x08 BS
	.word tab               ; 0x09 TAB
	.word linefeed          ; 0x0A LF
	.word linefeed          ; 0x0B VT    XXX same as LF
	.word linefeed          ; 0x0C FF    XXX same as LF
	.word carriage_return   ; 0x0D CR
	.word noop              ; 0x0E SO
	.word noop              ; 0x0F SI
	.word noop              ; 0x10 DLE
	.word noop              ; 0x11 XON
	.word noop              ; 0x12 DC2
	.word noop              ; 0x13 XOFF
	.word noop              ; 0x14 DC4
	.word noop              ; 0x15 NAK
	.word noop              ; 0x16 SYN
	.word noop              ; 0x17 ETB
	.word noop              ; 0x18 CAN
	.word noop              ; 0x19 EM
	.word noop              ; 0x1A SUB
	.word escape            ; 0x1B ESC
	.word noop              ; 0x1C FS
	.word noop              ; 0x1D GS
	.word noop              ; 0x1E RS
	.word noop              ; 0x1F US

escape_table:
	.word insert_character  ; 0x40 @
	.word cursor_up         ; 0x41 A
	.word cursor_down       ; 0x42 B
	.word cursor_forward    ; 0x43 C
	.word cursor_backward   ; 0x44 D
	.word cursor_down_col0  ; 0x45 E
	.word cursor_up_col0    ; 0x46 F
	.word cursor_column     ; 0x47 G
	.word cursor_position   ; 0x48 H
	.word tab_forward       ; 0x49 I
	.word erase_display     ; 0x4A J
	.word erase_line        ; 0x4B K
	.word insert_line       ; 0x4C L
	.word delete_line       ; 0x4D M
	.word noop              ; 0x4E N erase in field
	.word noop              ; 0x4F O erase in area
	.word delete_character  ; 0x50 P
	.word noop              ; 0x51 Q set editing extent
	.word noop              ; 0x52 R active pos report
	.word scroll_up         ; 0x53 S
	.word scroll_down       ; 0x54 T
	.word noop              ; 0x55 U next page
	.word noop              ; 0x56 V prev page
	.word tab_control       ; 0x57 W
	.word erase_character   ; 0x58 X
	.word noop              ; 0x59 Y vertical tab forward
	.word tab_backward      ; 0x5A Z
	.word noop              ; 0x5B [
	.word noop              ; 0x5C \ gotcha
	.word noop              ; 0x5D ]
	.word noop              ; 0x5E ^
	.word noop              ; 0x5F _
	.word cursor_column     ; 0x60 `
	.word cursor_forward    ; 0x61 a
	.word noop              ; 0x62 b repeat previous printable char
	.word device_attributes ; 0x63 c
	.word cursor_line       ; 0x64 d
	.word cursor_down       ; 0x65 e
	.word cursor_position   ; 0x66 f
	.word tab_clear         ; 0x67 g
	.word set_modes         ; 0x68 h
	.word noop              ; 0x69 i media copy
	.word cursor_backward   ; 0x6A j
	.word noop              ; 0x6B k line pos backw
	.word reset_modes       ; 0x6C l
	.word attributes        ; 0x6D m
	.word device_status     ; 0x6E n
	.word noop              ; 0x6F o qualify area
	.word noop              ; 0x70 p
	.word noop              ; 0x71 q
	.word set_region        ; 0x72 r
last_escape equ 0x72

attribute_table:
	.word attribute_0
	.word attribute_1
	.word attribute_2
	.word attribute_3
	.word attribute_4
	.word attribute_5
	.word attribute_6
	.word attribute_7
	.word attribute_8
	.word attribute_9
	.word attribute_10
	.word attribute_11
	.word attribute_12
	.word attribute_13
	.word attribute_14
	.word attribute_15
	.word attribute_16
	.word attribute_17
	.word attribute_18
	.word attribute_19
	.word attribute_20
	.word attribute_21
	.word attribute_22
	.word attribute_23
	.word attribute_24
	.word attribute_25
	.word attribute_26
	.word attribute_27
last_attribute equ 27

;----------------------------------------------------------------------
