
/*
 *  0       MBUS aka "MASC"
 *  1 2,3   GPS
 *  2 12,13 printer
 *  3 16,17 barcode reader (not the one in keypad)
 *  4 ?,?   "TL2"
 *
 *  XXX after writing to MBUS, read 0x18 to avoid extra net active interrupt.
 */

#include "head.h"
#include "pcb.h"
#include "asic.h"

struct file
	Uart_1 = { putc_1, getc_1 },
	Uart_2 = { putc_2, getc_2 },
	Uart_3 = { putc_3, getc_3 },
	Uart_4 = { putc_4, getc_4 },
	Mbus   = { putc_0, getc_0 };

void
uart_config(u, cfg)
{
	/* 0x20?  cs7 if set?
	 * 0x18?  parity, 0 no, 3 even?
	 * 0xC0?  stopbits?
	 * 0x07   baud:
	 * 0x06    19200bd
	 * 0x00      300bd */

	outb(u+UCTRL, cfg);

	/* 0x01 rxintena
	 * 0x02 txintena
	 * 0x04 ?
	 * 0x08 ? */

	outb(u+UIENB, 0x03);
}

/* MBUS (and ?) idle monitor
 * 0x08 if change
 * 0x10 if line idle */

void
netfree(b)
	byte b;
{
#if 0
	printf(" nf%x ", b);
#endif
}

#asm
	include asm.inc
	.text

	RINT equ 0x01
	TINT equ 0x02
	RXBF equ 0x04
	TXBF equ 0x08
	ERR  equ 0x70
	EPAR equ 0x10 ; parity ?
	EOVR equ 0x20 ; overrun ?
	EBRK equ 0x30 ; framing/break ?


;----------------------------------------------------------------------
; wasteful, but

	macro uart           ; name reg_base buffer_size

	.bss
	.comm tx_buf_?1, ?3      ; buffer_size
	.comm rx_buf_?1, ?3      ; buffer_size

	.data
rx_ins_?1: .word rx_buf_?1
rx_rem_?1: .word rx_buf_?1

tx_ins_?1: .word tx_buf_?1
tx_rem_?1: .word tx_buf_?1
_tx_act_?1: .byte 0

	export _tx_act_?1

	.text
	export uart_?1_int

uart_?1_int:                    ; from boot.c:int_0
	in al, ?2 + USTAT
uart_?1_again:
	mov ah, al                  ; status in AH
	test al, # RINT
	jz no_rx_?1

	in al, ?2 + UDATA           ; rx data in AL
	test ah, # ERR
	jz good_?1
	or al, # 0x80               ; XXX flag errors
good_?1:
	mov bx, [rx_ins_?1]
	mov rx_buf_?1[bx], al
	inc bx
	and bx, # ?3 - 1            ; buffer_size - 1
	mov [rx_ins_?1], bx
no_rx_?1:
	test ah, # TINT
	jz no_tx_?1

	mov bx, [tx_rem_?1]
	cmp bx, [tx_ins_?1]
	mov al, # 0                 ; if done
	jz tx_done_?1

	mov al, tx_buf_?1[bx]
	out ?2 + UDATA, al
	inc bx
	and bx, # ?3 - 1            ; buffer_size - 1
	mov [tx_rem_?1], bx
	mov al, # 1                 ; not done
tx_done_?1:
	mov [_tx_act_?1], al
no_tx_?1:
	in al, ?2 + USTAT
	test al, # RINT | TINT
	jnz uart_?1_again           ; XXX seems to be necessary ??
	ret


	export _getc_?1
	export _putc_?1

_getc_?1:
	xor ax, ax
	mov bx, [rx_rem_?1]
	cmp bx, [rx_ins_?1]
	jnz gotc_?1
	dec ax                   ; -1
	ret
gotc_?1:
	mov al, rx_buf_?1[bx]    ; 0...255
	inc bx
	and bx, # ?3 - 1         ; buffer_size - 1
	mov [rx_rem_?1], bx
	ret

_putc_?1:
	mov bx, # _tx_act_?1
	pushf
	cli
	cmpb [bx], # 0
	jz start_?1
	mov bx, [tx_ins_?1]
	mov tx_buf_?1[bx], al
	inc bx
	and bx, # ?3 - 1         ; buffer_size - 1
	mov [tx_ins_?1], bx
	popf
	ret
start_?1:
	movb [bx], # 1
	out ?2 + UDATA, al
	popf
	ret

	mend

	uart 0,UART_0,1024
	uart 1,UART_1,1024
	uart 2,UART_2,1024
	uart 3,UART_3,1024
	uart 4,UART_4,1024

	export _uart_unwedge

_uart_unwedge:                ; XXX
	pushf
	cli
	call uart_0_int
	call uart_1_int
	call uart_2_int
	call uart_3_int
	call uart_4_int
	popf
	ret

;----------------------------------------------------------------------
	.data

;----------------------------------------------------------------------
	.bss

