
; don't know if the glitch in P/V flag after LD A, I can happen
; in TMPZ84C00AT-8. doing disable/restore nesting by hand anyway.
; many other cpus in all those different logics, too.
;
; Sign Zero - HalfCarry - ParityEven/oVerflow add/sub Carry

	.module boot

	.include "iomap.inc"
	.include "cfg.inc"

	.area _MEMORY (ABS) ; ordering of segments for the linker
	.area _CODE
	.area _GSINIT
	.area _GSFINAL
	.area _DATA
	.area _DATA2
	.area _DATA3
	.area _BSS

;----------------------------------------------------------------------
; above C000 is either
;  - 16 kB work ram
; or
;  -  8 kB bbram, visible in C000 and E000. lower 4kB used
;  -  2 kB EEPROM in L8M
;
; P8N and L8M are the troublesome logics, P8E and L8TM have single bbram chip

	.area _MEMORY (ABS)

	.org 0x8000     ; banked rom area
_rom_bank::

	.org 0xC000     ; 4 kB reserved for bbram, cfg etc
_ram::
	.ds 0x10        ; 16 reserved bytes
_bbram::            ; START OF NONVOLATILE AREA -----------------------

	.org 0xC100     ; 240 unassigned bytes
_cfg_START::
_cfg::
	.IF L8M ;==========================================================

	.org 0xC400     ; 768 bytes for cfg
_cfg_END::
_memories::         ; 768 bytes for 40 memories

	.org 0xC700
_rfc_table::        ; 256 bytes for RFC, then 2kB hole until D000 XXX

	.org 0xC800
_bbend::            ; END OF NONVOLATILE AREA -------------------------

	.org 0xBFFF     ; ROM stack for nmi while eeprom active
bombstack:          ; (L8M rom1 cannot be writeable eeprom)

	.ELSE ; P8x or L8TM ===============================================

	.org 0xCA00     ; 2304 bytes for cfg
_cfg_END::
_memories::         ; 1280 bytes for 100 memories

	.org 0xCF00
_rfc_table::        ; 256 bytes for RFC

	.org 0xD000
_bbend::            ; END OF NONVOLATILE AREA -------------------------

	.org 0xC010     ; minimal initial stack for nmi while bbram active
bombstack:          ; (either rom could be writeable eeprom)

	.ENDIF ;===========================================================

	.org 0xD000
_buffers::          ; afsk_rx buffers, [256] times 8 XXX cheap tricks

	.org 0xD800     ; sio buffers [256] each
_sio_a_rx_buf:

	.org 0xD900
_sio_a_tx_buf:

	.org 0xDA00
_sio_b_rx_buf::

	.org 0xDB00
_sio_b_rx_bufend::
_sio_b_tx_buf::

	.org 0xDC00
_sio_b_tx_bufend::
_afsk_sine::        ; 3 sines [256] (different amplitudes/centers)

	.org 0xDD00
_dtmf_sine::

	.org 0xDE00
_ctcss_sine::

	.org 0xDF00
_nmea_buf::            ; sentence buffer 128 bytes
	.ds 0x80
nmea_bufend:
_nmea::                ; sentence structure 64 words

	.org 0xE000        ; 175 bytes for stack pointers
	.ds 2              ; TASK_DUMMY, saved when idle, but never loaded
_TASK_0::
	.ds 2
_TASK_1:
	.ds 2
_TASK_2:
	.ds 2              ; list ends in zero word

	.org 0xE200 - 337  ; 0xE0AF ... 0xE1FF, bit 1 of MSByte clear
_dcs_wave::            ; end in 0xE200, bit 1 of MSByte set XXX cheap tricks
	DCS_WAVEBIT = 1    ; wrap from 1 to 2

	.org 0xE200     ; rest of 16 kB ram here (_data) XXX ./mak must match

;----------------------------------------------------------------------
; SP should be kept well above 0xF000,
; to protect nvram from NMI spike.
;
; Zilog appnote
; "The Z80 Family Program Interrupt Structure", May 1978, says:
;
; "If an /NMI occurs during a reset true condition, the CPU will
; execute one mainline program step after reset is taken away
; and then process the /NMI."
;
; reset sets SP=FFFF, it is said, so no problem anyway.

	.org 0          ; RST 0   32 kB non-banked rom area
_rom::
reset:              ; PC=0 IFF1=IFF2=0 IM=0 I=R=0 (undoc: SP=AF=FFFF)
_boot::
_exit::             ; there is no exit
	ld sp, # 0      ; TASK_0 stack quickly. always executes before nmi
	di
	jp boot_continue
	.byte 0         ; cksum by "mkbin"

	.org 0x08       ; RST 08h not used, filled by FF (SDCC runtime (?))

	.org 0x10       ; RST 10h JP (HL)
	jp (hl)
_nop::
	ret

	.org 0x14
	pop hl          ; where_am_i in fixed location 0x0014
	jp (hl)

	.org 0x18       ; RST 18h JP (IX)
	jp (ix)

	.org 0x20       ; RST 20h JP (IY)
	jp (iy)

	.org 0x28       ; RST 28h RETI
	reti

	.org 0x30       ; RST 30h DISABLE
_disable::
	di
	push hl
	ld hl, # DInest
	inc (hl)
	pop hl
	ret p           ; else oops, too many (128) nested disables

	.org 0x38       ; RST 38h (opcode 0xFF)
	call _crash     ; oops

JPHL    = 0x10      ; cannot be labels (?!?)
JPIX    = 0x18
JPIY    = 0x20
RETI    = 0x28
DISABLE = 0x30

;----------------------------------------------------------------------
; not restarts. usually jumped to, not called.

	.org 0x40

_restore::          ; inverse of rst DISABLE aka disable()
	push hl
	ld hl, # DInest
	dec (hl)
	pop hl
	ret nz
	ei
	ret

di_undo::           ; inverse of plain di (in quickies)
	push af
	ld a, (DInest)
	or a
	jr nz, 1$
	ei
1$:
	pop af
	ret

di_undo_trash_af::  ; should be a macro
	ld a, (DInest)
	or a
	ret nz
	ei
	ret

;----------------------------------------------------------------------

	.org 0x66          ; NMI
nmi:

	push af

	ld a, i            ; sets Z and P/V. IFF1=0 prevents the P/V glitch
	jr z, _boot        ; I still 0, not fully booted, try again

	push af

	cp #>_iv
	jr nz, _boot       ; strange I, try again

	push hl            ; HL IFF2 AF PC on stack

	ld hl, # 0         ; where is the stack ?
	add hl, sp
	ld a, h

	;------------------

	.IF L8TM

	cp #> _end         ; should always be above bss
	jr c, _boot

	ld a, (_bank)
	out (BANK), a      ; normal rom banking

	.ELSE              ; P8x or L8M

	cp #> bombstack    ; carry, zero, or neither if normal stack
	jr c, _boot        ; below ram, never happened

	.IF P8x

	; might have been reading "DTMF-prom" (ROM1)
	; or second rom bank (top 16 kB of ROM0);
	; force rom-space back to contiguous 48 kB.

	ld a, (_out_2)     ; assume normal ram banking
	jr nz, 1$          ; if so
	ld a, d            ; D keeps out_2 while in save_bbram
	ld sp, iy          ; IY keeps stack
1$:
	out (OUT_2), a     ; now normal rom/ram banking

	.ENDIF             ; of dance with P8x

	.IF L8M

	ld a, (_pio_b)
	jr nz, 1$
	ld a, d            ; D keeps pio_b while in save_bbram
	ld sp, iy
1$:
	out (PIO_B_DATA), a

	.ENDIF             ; of dance with L8M

	push af            ; Z STILL SET if save_bbram was interrupted

	.ENDIF             ; of not L8TM

	;------------------

	push bc            ; not the same stack frame as with interrupts
	push de
	push iy
	push ix
	ex af, af'         ; beware just one exchange
	exx
	push af
	push hl
	push bc
	push de

	rst DISABLE        ; make things look regular (IFF1=IFF2=0, DInest > 0)

	call which_nmi     ; watchdog, power loss or #on/off ?

	ld hl, # DInest
	dec (hl)           ; restore DInest back to what it was
	call m, _crash     ; out of sync, bug somewhere

	pop de
	pop bc
	pop hl
	pop af
	ex af, af'         ; beware just one exchange
	exx
	pop ix
	pop iy
	pop de
	pop bc

	.IF P8x + L8M

	pop af             ; Z set if save_bbram was interrupted
	jr nz, 2$

	ld sp, # bombstack
	jp retry_save      ; AF HL have been lost, no EI
2$:
	.ENDIF

	pop hl
	pop af             ; old IFF2
	jp pe, 3$          ; was 1, do EI

	pop af             ; was 0, no EI
	ret
3$:
	pop af
	ei
	ret

which_nmi:
	in a, (PIO_A_DATA)
	bit 2, a               ; PA_WDR
	call nz, _crash        ; watchdog reset

	bit 3, a               ; PA_PWR, switch 'on' is '0'
	jp z, _power_loss      ; loss of supply, if so

	; #on/off to off-position

	ld a, (_cfg + stay_on)
	or a
	ret nz                 ; ignore #on/off

	jp _power_off          ; might still return ?

;----------------------------------------------------------------------
	.area _CODE      ; org 0x0100

	; does not belong here,
	; but where is .align ?

	.include "tables.inc"

;----------------------------------------------------------------------

boot_continue:
	xor a
	ld i, a           ; flag for nmi (needed if a soft reset)
	ld r, a           ; R.7 could be another flag someday
	im 1              ; rst 0x38 "oops" if never happens

	out (DA_TPC), a   ; avoid disaster
	out (WDOG), a

	;------------------------------------
	; fix things for contiguous 48 kB ROM and 16 kB work RAM

	.IF P8x

	ld a, # O2_SMEM | O2_ROM0
	out (OUT_2),a              ; P8N: work-ram. both: ROM0

	ld a, # 1
	out (CSMEM), a             ; P8E: work-ram

	.ENDIF                     ; P8x banking setup

	.IF L8M

	ld a, # PB_SMEM
	out (PIO_B_DATA), a   ; SMEM=1 work-ram, ROM S/L=0

	ld a, # 0x00          ; flush, in case it was waiting for something
	out (PIO_B_CTRL), a
	out (PIO_B_CTRL), a

	ld a, # 0xCF          ; ddr-set, ones are inputs
	out (PIO_B_CTRL), a
	ld a, # PB_DCU | PB_EXIN1 | PB_EXIN2
	out (PIO_B_CTRL), a

	.ENDIF                ; L8M banking setup

	.IF L8TM

	ld a, # BNK_RXCA_RXC
	out (BANK), a         ; L8TM: 32+16 kB rom in two pcs of 27512

	ld a, # PB_SMEM       ; XXX no need to rush here, there is pullup
	out (PIO_B_DATA), a   ; work-ram

	ld a, # 0x00          ; flush, in case it was waiting for something
	out (PIO_B_CTRL), a
	out (PIO_B_CTRL), a

	ld a, # 0xCF          ; ddr-set, ones are inputs
	out (PIO_B_CTRL), a
	ld a, # PB_DCU | PB_EXIN1 | PB_EXIN2 | PB_AVAIL
	out (PIO_B_CTRL), a

	.ENDIF                ; L8TM banking setup

	;------------------------------------
	; PIO A to mode 3, all input (for now)

	ld a, # 0x00          ; flush
	out (PIO_A_CTRL), a
	out (PIO_A_CTRL), a

	ld a, # 0xCF          ; ddr-set, ones are inputs
	out (PIO_A_CTRL), a
	ld a, # 0xFF
	out (PIO_A_CTRL), a

	; short settling time.
	; on L8TM, waits DS1210 to wake up
0$:
	xor a
1$:
	ld hl, # _bbend   ; scratch location in ram
	ld (hl), a
	cp (hl)
	jr nz, 0$         ; did not stick. restart loop
	inc a
	jr nz, 1$         ; test all 256 bit patterns (on this single byte)
2$:
	ld (hl), a        ; zero everything after nonvolatile ram area
	cp (hl)
	jr nz, 0$         ; did not stick. restart loop
	inc hl
	cp h
	jr nz, 2$         ; zero _bbend ... FFFF

	ld de, # _data    ; copy initialized data
	ld hl, # _etext   ; from rom image, built by mkbin.c
	ld c, (hl)        ; 1st word is length
	inc hl
	ld b, (hl)
	inc hl            ; and data follows
	ldir

	; stack usable now (hopefully).
	; reset interrupt logic. /RESET goes to SIO, but not to PIO

	ld b, # 0         ; 256 times (two would do, for pio a and b)
3$:
	rst RETI
	djnz 3$

	call gsinit       ; argh
	call logic_probe  ; P8E or P8N ?

	ld ix, # 0        ; C frame pointer

	call _initialise  ; lotsa OUT etc

	ld a, #>_iv
	im 2              ; ready now
	ld i, a           ; now nmi doesn't just boot

	;------------------------------------

task_0:

	ld a, # 60        ; power off in 60 seconds, if low battery persists
	ld (batt_timo), a


	ld hl, # 0xFC00      ; top 1kB for task 0
	ld de, # _TASK_1

	ld bc, # task_1
	call make_task

	.globl _TASK_2
	ld bc, # _task_2
	call make_task

	ld a, # 0x83
	ld (multitask), a    ; 3 runnable tasks, switch allowed

	xor a
	ld (DInest), a       ; XXX the bug is still there ? XXX
	ei

mainloop:                ; forever.
	call _hare

	call _sleep          ; woken every tick XXX PTT can slip and delay 4ms

	ld hl, # 0
	ld (_fido), hl       ; 10 second watchdog in tick()

	call _hare

	ld hl, # slugf
	srl (hl)
	call c, tortoise

	jp mainloop

;----------------------------------------------------------------------
; 256 Hzy

tortoise:
	ld hl, (_leds_XXX)   ; H = leds
	ld a, h
	cp l
	call nz, _update_leds

	ld hl, # _dpy_XXX
	dec (hl)
	call z, _display

	ld hl, (keypad_XXX)
	ld a, h
	or l                 ; clears carry
	jr z, 1$
	ld bc, (timef)
	sbc hl, bc
	call m, keypad
1$:
	ld hl, (_key_timo)
	ld a, h
	or l                 ; clears carry
	jr z, 2$
	ld bc, (timef)
	sbc hl, bc
	call m, _key_timeout
2$:
	ld hl, # slugf + 2
	srl (hl)
	call c, _slug

	in a, (PIO_B_DATA)       ; changes on PB ?
	ld hl, # _pio_b_in
	ld c, (hl)
	ld (hl), a
	xor c                    ; changes
	and # PB_EXIN2           ; interesting changes
	call nz, pio_b_changed

	; finally

	call deq_incoming
	ret z

	push hl
	call _handle_incoming
	pop af
	ret

;----------------------------------------------------------------------

pio_b_changed:
	ld hl, # _icon + 7      ; car_u toggled
	ld a, (hl)
	sub # 1
	sbc a, a
	ld (hl), a
	call _display_soon
	ret

;----------------------------------------------------------------------

task_1:
2$:
	call _sleep             ; woken by sio_a_rca

	.IF P8x                 ; async sio a only on P8x
1$:
	ld hl, (_sio_a_rx_rem)
	di
	ld a, (_sio_a_rx_ins)
	cp l
	jr z, 2$                ; empty, back to sleep
	ei

	ld a, (hl)
	inc l
	ld (_sio_a_rx_rem), hl

	ld hl, (sio_a_input)
	rst JPHL                ; byte in A
	jp 1$                   ; look for more

	.ELSE

	jp 2$

	.ENDIF

;----------------------------------------------------------------------
; 256 Hz zerolevel interrupt

tick:
	out (WDOG), a        ; hold RESET for another second

	ld hl, (_fido)
	inc hl
	ld (_fido), hl       ; valid range for fido is 0 ... 0x9FF
	ld a, h              ; testing it every time here
	cp # 10              ; 10 seconds since last zero ?
	call nc, _crash      ; mainline is stuck

	ld a, (_cfg + stay_on)
	or a
	call nz, _toggle_MON   ; #ON/OFF override

	;------------------------------------------------------

	ld a, (_af_tx_timer)   ; ccir, dtmf, morse etc
	sub # 1
	jp c, 1$               ; 0 is inactive
	call z, do_af_tx       ; do something and give new count
	ld (_af_tx_timer), a   ; or just decrement
1$:
	ld a, (_alert)
	or a
	call nz, do_alert

	;------------------------------------------------------

	call _read_ccir        ; return in A and L
	ld hl, # ccir_old
	cp (hl)
	jp z, 2$               ; no change
	ld (hl), a
	call do_ccir           ; new tone in A
2$:

	ld a, (_cfg + dtmf_decoder)     ; 0 means none
	or a
	call nz, do_dtmf

	ld a, (_audio_flags)   ; 0x80 disables squelch
	or a
	call p, do_squelch

	;------------------------------------------------------
	; after squelch has had a chance to use "oldest rssi"

	ld a, (_ad_rssi)

	ld hl, (_ad_rssi_peak) ; downcount in H
	cp l
	jr nc, 3$
	dec h                  ; hold the old peak ?
	jr nz, 4$
3$:
	ld l, a                ; new peak
	ld h, # 0              ; hold for one second
4$:
	ld (_ad_rssi_peak), hl

	ld hl, (donald_rssi)
	ld (hl), a             ; fresh rssi overwrites oldest
	inc hl                 ; oldest is now here
	ld a, # rssi_log_end
	cp l
	jr nz, 5$
	ld hl, # rssi_log      ; wrap
5$:
	ld (donald_rssi), hl   ; rssi 31 msec ago, found here

	;------------------------------------------------------

	call _tick

	;------------------------------------------------------

	ld hl, # _TASK_0       ; woken up every 4 msec
	call wake

	ld hl, # slugf
	ld (hl), # 1           ; tortoise
	inc hl
	inc (hl)
	jr z, tick_1Hz
	ret po                 ; done, 254 times out of 256

; 0x80, bottom of second

	jp check_battery

; 0x00, top of second

tick_1Hz:
	inc hl
	ld (hl), # 1           ; slug

	; finally, set ON-led and reset lights

	ld hl, (light_timo)    ; while positive, counts down
	bit 7, h
	jr nz, 1$              ; do nothing, if already negative
	ld a, h
	or l                   ; was zero?
	dec hl
	ld (light_timo), hl
1$:
	ld hl, # _leds
	set LED_ON, (hl)       ; ON stays on, or blinks at 1 Hz

	ret nz
	res LED_LCDLIGHT, (hl) ; was zero = lights_off
	res LED_KEYLIGHT, (hl)
	ret

;----------------------------------------------------------------------
; b0 illegal tx
; b1 low battery

do_alert:
	di                 ; irritating atonal buzz, not very loud

	ld a, (_out_0)
	xor # O0_AUDIOC    ; XXX O0_MTC for fainter sound
	out (OUT_0), a
	xor # O0_AUDIOC
	out (OUT_0), a

	jp di_undo_trash_af

;----------------------------------------------------------------------
; 82C54 CLK_1  4.032 MHz ~248 nsec
; P8N   clock  4.032 MHz
; P8E   clock  8.064 MHz ~124 nsec (1 waitstate per each opcode fetch)

logic_probe:

	.IF P8x           ; not for L8M or L8TM

	ld c, # TMR_CTRL
	ld b, # TMR_SEL_MT | TMR_LATCH
	ld a, # TMR_SEL_MT | TMR_LSB | TMR_INTTC
	out (c), a

	ld a, # 73        ; near midpoint
	out (TMR_MT), a   ; count 73

	.byte 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0 ; 20 nops, 100T or 80T

	out (c), b        ; latch count
	in a, (TMR_MT)    ; P8E count 18, P8N count -18
	rlca
	and # 1
	ld (_logic), a    ; P8E 0, P8N 1

	.ENDIF            ; P8x

	ret

;----------------------------------------------------------------------
; priority order is
; pa > pb > sa rx > sa tx > sa esc > sb rx > sb tx > sb esc
;----------------------------------------------------------------------
; 1968.75 Hz into PA.0
; slow things down to 256 Hz tick
;----------------------------------------------------------------------
; one of pio_a_int (_XXX) is active

pio_a_int_dcs::
	ex af, af'
	exx

	ld hl, (dcs_ptr)    ; 337 bytes makes it about 134.37 Hz
	ld a, (hl)
	out (DA_RFC), a
	inc hl
	bit DCS_WAVEBIT, h  ; buffer ends at 0xE200 XXX
	jr z, 1$            ; valid area 0xE0AF ... 0xE1FF
	ld hl, # _dcs_wave  ; 0xE200 - 337
1$:
	ld (dcs_ptr), hl

	ld hl, (adc_reader)
	jp (hl)             ; jumps back to adc_done:

pio_a_int_ctcss:
	ex af, af'
	exx

	ld bc, (ctcss_phinc)
	ld hl, (ctcss_phacc)
	add hl, bc
	ld (ctcss_phacc), hl

	ld l, h
	ld h, #>_ctcss_sine ; centered on RFC setting
	ld a, (hl)
	out (DA_RFC), a

	ld hl, (adc_reader)
	jp (hl)             ; jumps back to adc_done:

;----------------------------------------------------------------------

pio_a_int:              ; 1968.75 Hz
	ex af, af'
	exx

	ld hl, (adc_reader) ; one A/D finished, start another
	jp (hl)             ; jumps back to adc_done:

adc_done:               ; common part of pio_a_int (_XXX)
	ld (adc_reader), hl ; remember position in adc_loop

	ld bc, # 0x214A     ; 1.0 / 1968.75 * 2^24 ~ 1.0000 Hz (0x2149C2)
	ld hl, (time_lsb)
	add hl, bc
	ld (time_lsb), hl   ; increment 48.24 time
	jr c, clock
fast_out:
	ex af, af'
	exx
	ei
	reti

clock:
	ld hl, # timef - 1  ; 256 Hz. increment 48.8 time
1$:
	inc l
	inc (hl)            ; will crash in 8919594 years
	jr z, 1$

	ld hl, # in_int     ; initially 0
	ld a, (hl)          ; nz if nested
	ld (dotick), a      ; nz if left pending, 0 if doing it now
	or a
	jr nz, fast_out
	ld (hl), h          ; nz, "in int" now

	ex af, af'
	exx
	push af             ; pusha
	push hl
	push bc
	push de
	push iy
	push ix
again:
	ei                  ; PIO (or SIO) needs the first RETI
	rst RETI            ; latter are superfluous, but harmless
	call tick

	ld hl, # multitask  ; XXX use bits in single byte
	set 5, (hl)         ; 0x20, time to change task

leave_zero:
	ld hl, # dotick
	xor a
	di
	cp (hl)             ; check the flag
	ld (hl), a          ; zero it
	jr nz, again

	inc l               ; &in_int
	ld (hl), a          ; zero it too. no EIs any more, until gone

	inc l               ; &multitask
	ld a, (hl)          ; 0x40 can be used to fool the check
	cp # 0xA2           ; allowed, time to switch, multiple runnable tasks ?
	jp c, leave
	and # ~0x60         ; force-a and time-to flags cleared
	ld (hl), a

	ld hl, # 0          ; old stack stays runnable, saved sp.15 set
	add hl, sp
	call next_stack     ; HL in/out
	ld sp, hl           ; new (maybe) XXX rom banking ?
leave:
leave_nested:
	pop ix              ; popa
	pop iy
	pop de
	pop bc
	pop hl
	pop af
	ei                  ; superfluous, if leave_nested
	reti                ; often superfluous, but harmless

;----------------------------------------------------------------------
; HL in/out
; save current stack (as runnable or suspended)
; and find next runnable stack.
; if none, return old

next_stack:
	ex de, hl           ; old stack in DE
	ld hl, (sp_ptr)
	ld (hl), e
	inc l
	ld (hl), d          ; old stack saved (maybe to TASK_DUMMY)

	xor a               ; compare 0 - sp_msbyte
1$:
	inc l               ; &lsbyte
	inc l               ; &msbyte
	cp (hl)             ; 0 - 00 ends, 0 - 70 skips, 0 - F0 accepts
	jp m, 1$            ; 70, suspended, try next
	jr nz, 3$           ; F0, valid, take it

	ld l, # 1           ; TASK_0 - 1
2$:
	inc l               ; scan again, now all of them
	inc l
	cp (hl)
	jp m, 2$
	jr z, 4$            ; 00, no runnable stacks
3$:
	ld d, (hl)
	dec l               ; nz, as TASK_DUMMY is at ??00
	ld e, (hl)          ; new stack in DE
	ld (sp_ptr), hl
	ex de, hl           ; HL and nz
	ret

4$:
	ld (sp_ptr), a      ; TASK_DUMMY
	ex de, hl           ; old HL and z
	ret

;----------------------------------------------------------------------
; make task HL runnable
; use only AF HL

_wake::
	pop af
	pop hl
	push hl
	push af
wake:
	ld a, # 0x80      ; sp.15
	inc l             ; sp msbyte

	di

	xor (hl)
	jp p, di_undo_trash_af ; runnable already
	ld (hl), a             ; now runnable

	ld hl, # multitask
	inc (hl)              ; one more runnable task

;	set 5, (hl)       ; 0x20 time to switch XXX maybe not to the correct task
; not worth it XXX ?

	jp di_undo_trash_af

;----------------------------------------------------------------------
; hog, sleep and yield keep all regs and flags

_hog::
	push hl
	ld hl, # multitask
	res 7, (hl)         ; 0x80, block switch in leave_zero
	pop hl
	ret

;----------------------------------------------------------------------
; sleep/yield is also an implicit enable + unhog

_sleep::                ; suspend until wake()
	push af
	push hl             ; keep scratch registers too, just for fun
	ld hl, # 0x8000     ; suspended, saved sp.15 clear
	jr yield_1

_yield::                ; give way, but stay runnable
	push af             ; pusha
	push hl
	ld hl, # 0          ; runnable, saved sp.15 set
yield_1:
	push bc
	push de
	push iy
	push ix             ; std frame: pc af hl bc de iy ix

	xor a               ; A 0 and NC
	add hl, sp          ; if sleep, CY and b15 flips to 0

	ld de, # multitask
	di
	ld (DInest), a      ; undo any number of disables
	ld a, (de)
	sbc # 0             ; if sleeping, one less runnable task
	or # 0x80           ; undo a possible hog()
	ld (de), a
1$:
	call next_stack     ; HL in/out
	jr z, 2$            ; nobody wants to run

	ld sp, hl           ; new (or not) XXX rom banking ?
	ei
	pop ix              ; popa
	pop iy
	pop de
	pop bc
	pop hl
	pop af
	ret

; leave_zero does not switch, unless 2+ tasks wake up.
; nmi and "fast interrupts" do not switch at all.
; idle, and rescan here, using old stack.

2$:
	ei
	halt                ; somebody maybe takes over
	di                  ; did not
	ld a, (multitask)
	and # 0x1F
	jr z, 2$
	jp 1$               ; value in HL does not matter anymore

;----------------------------------------------------------------------
; initial pc in BC.
; 'sp' kept in HL, stacks slot kept in DE

make_task:
	dec hl
	ld (hl), #>_crash
	dec hl
	ld (hl), #<_crash    ; task returns to _crash() XXX exit() ?

	dec hl
	ld (hl), b
	dec hl
	ld (hl), c           ; initial pc

	ld bc, # -12         ; std frame: pc af hl bc de iy ix
	add hl, bc           ; all but pc zero

	ex de, hl
	ld (hl), e           ; TASK_n in stacks array
	inc l
	ld (hl), d           ; initially "runnable", saved sp.15 set
	inc l
	ex de, hl

	ld bc, # 16 - 1024   ; 1kB stack XXX separate intstack
	add hl, bc
	ret

;----------------------------------------------------------------------

	.IF 0

pio_b_int:
	.include "pusha_int"
	jp _pio_b_int

	.ENDIF

;----------------------------------------------------------------------

; parity, framing or overrun. just discard XXX

sio_a_src:
	push af
	in a, (SIO_A_DATA)
	ld a, # WR0_ERROR_RESET
	out (SIO_A_CSR), a
	pop af
	ei
	reti

	.IF P8x       ; async SIO A on P8x

sio_a_rca:
	push af
	push hl
	ld hl, (_sio_a_rx_ins)
1$:
	in a, (SIO_A_DATA)
	ld (hl), a
	inc l
	in a, (SIO_A_CSR)        ; XXX most of the time not useful
	rra                      ; RR0_RCA to carry
	jr c, 1$
	ld (_sio_a_rx_ins), hl

	ld hl, # _TASK_1
	call wake                ; XXX wake only if newline ?

	pop hl
	pop af
	ei
	reti

sio_a_tbe:
	push af
	push hl
	ld hl, (_sio_a_tx_rem)
	ld a, (_sio_a_tx_ins)
	sub l                    ; get free zero
	jr z, 1$                 ; if no more
	ld a, (hl)               ; this probably happens more often
	out (SIO_A_DATA), a
	inc l
	ld (_sio_a_tx_rem), hl
	pop hl
	pop af
	ei
	reti
1$:
	ld (_sio_a_tx_active), a ; not anymore
	ld a, # WR0_RESET_TXINT
	out (SIO_A_CSR), a
	pop hl
	pop af
	ei
	reti

	.ENDIF        ; async SIO A on P8x

	.IF L8x       ; sync SIO A on L8x

sio_a_rca:
	.include "pusha_int"
	jp _sio_a_rca

sio_a_tbe:
	.include "pusha_int"
	jp _sio_a_tbe

	.ENDIF        ; sync SIO A on L8x

sio_a_esc:
	.include "pusha_int"

sio_a_esc_pickup::         ; pick up changes after busywaiting SA_DA XXX

	.IF P8x                ; reset HOOK edge detect and TMR_2_OUT latch

	in a, (TMR_CTRL)

	.ENDIF

	in a, (SIO_A_CSR)      ; new RR0
	ld hl, # _sio_a_rr0
	ld c, (hl)             ; old
	ld (hl), a
	xor c                  ; changes
	push af
	inc sp                 ; sio_a_esc(byte changes)

	and # SA_DA            ; SA_INT
	jr z, 1$
	ld bc, # 40 * 256 / 1000      ; keypad() after 40 msec
	ld hl, (timef)
	add hl, bc
	set 0, l
	ld (keypad_XXX), hl
1$:
	call _sio_a_esc
	inc sp

	ld a, # WR0_RESET_ESCINT
	out (SIO_A_CSR), a
	ret

;----------------------------------------------------------------------

sio_b_src:
	.include "pusha_int"
	jp _sio_b_src

sio_b_rca_UNUSED:
	.include "pusha_int"
	jp _sio_b_rca

sio_b_rca:
	push af
	push hl
	ld hl, (_sio_b_rx_ins)
1$:
	in a, (SIO_B_DATA)
	ld (hl), a
	inc l
	in a, (SIO_B_CSR)
	rra                       ; RR0_RCA to carry
	jr c, 1$
	ld (_sio_b_rx_ins), hl
	pop hl
	pop af
	ei
	reti

sio_b_tbe:
	.include "pusha_int"
	jp _sio_b_tbe

sio_b_esc:
	.include "pusha_int"

	in a, (SIO_B_CSR)      ; new RR0
	ld hl, # _sio_b_rr0
	ld c, (hl)             ; old
	ld (hl), a
	xor c                  ; changes
	push af
	inc sp                 ; sio_b_esc(byte changes)
	call _sio_b_esc
	inc sp

	ld a, # WR0_RESET_ESCINT
	out (SIO_B_CSR), a
	ret

;----------------------------------------------------------------------
; ADC0809
; ALE+START wired together
;
; sometimes, wiggling d/c power cable,
; ad_batt sticks at FF (ought to be around F0, 14+ volts)
;
; ordered, just for fun, on P8x, to change only one mux bit
; before RSSI, SQL, BATT measurements.
; RSSI and SQL done double rate, always new values to tick.
;
; one straw still to try, try doing IN7 before BATT XXX
;
; use only AF HL (BC ? check)

adc_first:
	out (AD_RSSI), a   ; leading edge ALE, channel in address bus lsbits
	nop                ; 2.5 usec from 1st ALE to 2nd START
	nop
	nop
	out (AD_RSSI), a   ; leading edge SAR reset (ALE again), trailing START
	.include "adc_ret" ; resume later (ld hl, . + 6; jp adc_done)

adc_loop:
	.include "adc_ret" ; might have been long disable(), every another time
	in a, (AD_RSSI)    ; OE (address bus don't care)
	out (AD_SQL), a
	ld (_ad_rssi), a   ; delay between OUTs, gratis
	out (AD_SQL), a
	.include "adc_ret"
	.include "adc_ret"

	in a, (AD_SQL)
	out (AD_IN7), a
	ld (_ad_sql), a
	out (AD_IN7), a
	.include "adc_ret"
	.include "adc_ret"

	in a, (AD_IN7)
	out (AD_TPC), a
	ld (_ad_in7), a
	out (AD_TPC), a
	.include "adc_ret"
	.include "adc_ret"

	in a, (AD_TPC)
	out (AD_BATT), a
	ld (_ad_tpc), a
	out (AD_BATT), a
	.include "adc_ret"
	.include "adc_ret"

	in a, (AD_BATT)
	out (AD_RSSI), a
	ld (_ad_batt), a
	out (AD_RSSI), a
	.include "adc_ret"
	.include "adc_ret"

	in a, (AD_RSSI)
	out (AD_SQL), a
	ld (_ad_rssi), a
	out (AD_SQL), a
	.include "adc_ret"
	.include "adc_ret"

	in a, (AD_SQL)
	out (AD_TP4), a
	ld (_ad_sql), a
	out (AD_TP4), a
	.include "adc_ret"
	.include "adc_ret"

	in a, (AD_TP4)
	out (AD_RPM), a
	ld (_ad_tp4), a
	out (AD_RPM), a
	.include "adc_ret"
	.include "adc_ret"

	in a, (AD_RPM)
	out (AD_FPM), a
	ld (_ad_rpm), a
	out (AD_FPM), a
	.include "adc_ret"
	.include "adc_ret"

	in a, (AD_FPM)
	out (AD_RSSI), a
	ld (_ad_fpm), a
	out (AD_RSSI), a

	ld hl, # adc_loop      ; rewind to top
	jp adc_done

;----------------------------------------------------------------------

	.IF L8M            ; no free chipselects on 74138
	.ELSE

_read_8888::

	di                 ; get instantaneous state

	in a, (MT8888_RS)  ; read status
	cpl                ; DELAYED STEERING# in b3 is
	rla
	and # 0x10         ;  inverse of 8870 StD
	ld l, a

	in a, (MT8888_RD)  ; read data
	and # 0xF          ;  is valid before and after StD
	or l
	ld l, a            ; in A and L

	jp di_undo         ; "StD b3 b2 b1 b0"

	.ENDIF

; if ROM1 is missing, bus echoes the last opcode (AND (HL))

_read_8870::           ; a byte from ROM1 (2nd one), the silly DTMF decoder

	ld hl, # _rom_bank ; into ROM1, any address, StD D3...0 ? ? afsqw

	di

	.IF P8x

	ld a, (_out_2)
	ld b, a          ; keep here for a while
	and # ~O2_ROM0   ; select ROM1. ROMA14 ... ROMA16 not touched
	out (OUT_2), a

	.ENDIF

	ld a, (hl)       ; if nmi comes and goes, might read ROM0[0x8000]
	and # 0xF8
	ld c, a          ; first sampling

	ld a, # 0xF8
	and (hl)         ; spread the three (hl) accesses evenlyish
	cp c
	jp z, 1$         ; i'll take it
	ld a, # 0xF8
	and (hl)        ; must be good now XXX
1$:
	rrca
	rrca
	rrca             ; StD D3...0
	ld l, a          ; just the 8870 bits, rightjust, in A and L

	.IF P8x

	ld c, # OUT_2
	out (c), b       ; back to regular state

	.ENDIF

	jp di_undo

_read_ccir::
	in a, (PIO_A_DATA)
	and # PA_CCIR          ; 0xF0
1$:
	ld l, a
	in a, (PIO_A_DATA)
	and # PA_CCIR
	cp l
	jr nz, 1$        ; re-read while changing

	rrca
	rrca
	rrca
	rrca
	ld l, a          ; rightjust in A and L
	ret

;----------------------------------------------------------------------

	.IF L8TM

_load_bbram::          ; L8TM, nothing to do
_save_bbram::
	ret

	.ELSE              ; "else" is a long story

_load_bbram::

	ld a, (_logic)
	or a
	ret z              ; P8E, nothing to do

	ld a, i            ; should be 0, called only once from initialise
	call nz, _crash    ; naughty

	.IF L8M             ; P8N or L8M then

	ld e, # ~PB_SMEM
	ld c, # PIO_B_DATA
	ld a, (_pio_b)

	.ENDIF

	.IF P8x

	ld e, # ~O2_SMEM
	ld c, # OUT_2
	ld a, (_out_2)

	.ENDIF

	ld d, a             ; D selects work ram
	and e
	ld e, a             ; E selects bbram/eeprom

	ld iy, # 0
	add iy, sp          ; keep SP in IY

	ld hl, # _bbram
	ld sp, # bombstack  ; safe area, in case nmi strikes (and reboots)

	.IF L8M      ; wait for any pending EEPROM write to finish

	out (c), e   ; eeprom visible
0$:
	ld a, (hl)
	cp (hl)      ; toggle bit is steady ?
	jr nz, 0$    ; wait until it is

	.ENDIF       ; L8M dancing
1$:
	out (c), e          ; bbram/eeprom visible
	ld a, (hl)
	out (c), d          ; work ram back
	ld (hl), a
	inc l
	jp nz, 1$           ; do one page, _bbend is aligned [256]
	inc h

	.IF L8M

	bit 2, h    ; A10 ?
	jr z, 7$
	set 0, e    ; PB_EEA10
7$:
	.ENDIF

	ld a, #>_bbend
	cp h
	jp nz, 1$           ; _bbram ... _bbend

	ld sp, iy           ; back to ram stack
	ret

_save_bbram::           ; similar to above, but more contorted

	ld a, (_logic)
	or a                ; clears carry
	ret z               ; P8E, nothing to do

	ld bc, # MAGIC       ; P8N or L8M then
	ld hl, (_cfg + 0)
	sbc hl, bc
	ret nz

	ld bc, # MAGIC >> 16
	ld hl, (_cfg + 2)
	sbc hl, bc
	ret nz              ; junk, no sale

	ld iy, # 0
	add iy, sp          ; keep SP in IY

	rst DISABLE

	; when restarted by nmi,
	; SMEM is active (work ram)
	; bbram contents might have changed (hmm?)
	; SP = bombstack
	; AF HL are lost
	; other context is kept

retry_save:

	.IF L8M

	ld e, # ~PB_SMEM
	ld c, # PIO_B_DATA
	ld a, (_pio_b)

	.ENDIF

	.IF P8x

	ld e, # ~O2_SMEM
	ld c, # OUT_2
	ld a, (_out_2)

	.ENDIF

	ld d, a                ; D normal
	and e
	ld e, a                ; E select bbram/eeprom

	ld hl, # _bbram
	ld sp, # bombstack     ; safe area, and flag for nmi

	.IF L8M      ; wait for any pending EEPROM write to finish

	out (c), e
0$:
	ld a, (hl)
	cp (hl)      ; toggle bit is steady ?
	jr nz, 0$

	out (c), d

	.ENDIF       ; L8M dancing

1$:
	ld a, (hl)          ; loop top
	out (c), e          ; bbram

	.IF P8x

	ld (hl), a    ; RAM. just store every byte

	.ENDIF

	.IF L8M       ; NMC9817A can update 32 bytes at once

	cp (hl)       ; avoid EEPROM update, if possible
	jp z, 5$      ; same, skip.
	jr 3$         ; load first byte
2$:
	out (c), d    ; XXX is there a way to compare, after 1st store ?
	ld a, (hl)
	out (c), e
3$:
	ld (hl), a    ; store.
	inc l
	ld a, l
	and # 31      ; end of eepage ?
	jp nz, 2$     ; go back for more

	dec l         ; back at last written byte
4$:
	ld a, (hl)    ; toggle bit steady ?
	cp (hl)
	jr nz, 4$     ; wait until it is
5$:
	.ENDIF        ; L8M dancing

	out (c), d          ; work ram
	inc l
	jp nz, 1$
	inc h

	.IF L8M

	bit 2, h    ; A10 ?
	jr z, 7$
	set 0, e    ; PB_EEA10
7$:
	.ENDIF

	ld a, #>_bbend
	cp h
	jp nz, 1$           ; _bbram ... _bbend

	ld sp, iy           ; back to ram stack
	jp _restore

	.ENDIF              ; of P8x and L8M xxx_bbram()

;----------------------------------------------------------------------
; Async SIO A only on P8x

	.IF P8x

_sio_a_putc::         ; void sio_a_putc(word);
	pop hl
	pop bc            ; word arg
	push bc
	push hl
putc:
	di

	ld a, (_sio_a_tx_active)   ; already active ?
	or a
	jr z, 1$
	ld hl, (_sio_a_tx_ins)
	ld (hl), c
	inc l                      ; aligned 256
	ld (_sio_a_tx_ins), hl
	jp di_undo_trash_af
1$:
	inc a                      ; a now 1
	ld (_sio_a_tx_active), a   ; is active
	ld a, c
	out (SIO_A_DATA), a
	jp di_undo_trash_af        ; RET

_sio_a_puts::
	pop bc
	pop de
	push de
	push bc

	jr 2$
1$:
	ld c, a
	call putc
	inc de
2$:
	ld a, (de)
	or a
	jr nz, 1$
	ret

	.ENDIF       ; Async SIO A only on P8x

;----------------------------------------------------------------------
; CTCSS is centered on RFC if full duplex, to not foul RX,
; center is 0x80 otherwise, to get full adjustment range.

_ctcss::                      ; void ctcss(word dHz)

	ld a, (_cfg + full_duplex)
	or a
	ld l, # 0x80              ; center if not fdx
	call nz, _RFC_get         ; center if fdx
	ld d, l                   ; new and
	ld a, (_ctcss_sine + 0)   ;  old center
	cp d
	ld a, (_cfg + ctcss_gain) ; new and
	ld hl, # ctcss_ogain      ;  old gain

	jr nz, 1$                 ; center has changed
	cp (hl)
	jr z, 2$                  ; neither has changed
1$:
	ld (hl), a
	ld e, a
	ld iy, # _ctcss_sine
	call make_sine            ; IY tab, D center, E gain
2$:
	pop af                    ; arguments:
	pop de                    ; tone frequency in dHz
	push de
	push af

	; Hz / 1968.75 * 2^16
	; dHz * 2^16 / 19688

	ld hl, # 19688 / 2        ; DEHL = dHz << 16 and rounding
	ld bc, # 19688
	call div_dehl_bc          ; HL = phase increment

	ld (ctcss_phinc), hl

	ld hl, # pio_a_int_ctcss
	ld (_iv_pio_a), hl
	ret

_dcs_go::
	ld hl, # _dcs_wave
	ld (dcs_ptr), hl

	ld hl, # pio_a_int_dcs
	ld (_iv_pio_a), hl
	ret

_encoder_turnoff::

	ld a, (_iv_pio_a)            ; compare of lsbyte is enough
	cp # pio_a_int_dcs
	jr z, dcs_turnoff
	cp # pio_a_int_ctcss
	ret nz                       ; huh?

	ld hl, # ctcss_phacc+1       ; CTCSS phase msbyte
	ld a, # 0x80                 ; msbit

	di

	xor (hl)
	ld (hl), a                   ; invert phase
	jp di_undo_trash_af

dcs_turnoff:
	ld hl, # pio_a_int           ; XXX should emit burst of 01010101010101
	ld (_iv_pio_a), hl           ; XXX at double speed
	ret

_encoder_end::

	ld hl, # pio_a_int           ; encoders off
	ld (_iv_pio_a), hl

	ld hl, # 0
	ld (ctcss_phacc), hl         ; next time, start from zero

	ret

;----------------------------------------------------------------------
; 256 values, page aligned table,
; 512 kHz per each
; use bits 0x1FE00 for index

p_rfc:
	ld hl, (_rx_freq + 1)  ; drop 8 bits
	rr h
	rr l                   ; one more
	ld h, #>_rfc_table     ; page aligned
	ret

_RFC_get::                 ; byte RFC_get();  (of rx_freq)
	call p_rfc
	xor a                  ; find nonzero value
	inc l
1$:
	dec l                  ;  4 scan down
	jr z, 2$               ;  7 stop at index 0
	cp (hl)                ;  7 zero value ?
	jp z, 1$               ; 10 keep looking, while so
2$:
	ld l, (hl)
	ret                    ; value in L

_RFC_set::                 ; void RFC_set(byte); (of rx_freq)
	pop hl
	pop de                 ; in e
	push de
	push hl

	call p_rfc
	ld (hl), e
	ld a, e
	out (DA_RFC), a
	ret

_RFC_update::
	call _RFC_get
	ld a, l
	out (DA_RFC), a
	ret

;----------------------------------------------------------------------

_CCIRC_on::
	ld de, # O0_CCIRC * 0x101
	jp out_0_change

_CCIRC_off::
	ld de, # O0_CCIRC
	jp out_0_change

_out_0_change::       ; void out_0_change(byte clear, byte set);
	pop hl
	pop de            ; e clear, d set
	push de
	push hl
out_0_change::        ; args in DE

	ld a, e
	cpl               ; bits to keep
	ld hl, # _out_0

	di

	and (hl)          ; out_0 = (out_0 & ~clear) | set;
	or d
	ld (hl), a
	out (OUT_0),a
	jp di_undo_trash_af

_out_1_change::       ; void out_1_change(byte clear, byte set);
	pop hl
	pop de            ; e clear, d set
	push de
	push hl

	ld a, e
	cpl               ; bits to keep
	ld hl, # _out_1

	di

	and (hl)          ; out_1 = (out_1 & ~clear) | set;
	or d
	ld (hl), a
	out (OUT_1),a
	jp di_undo_trash_af

_out_2_change::       ; void out_2_change(byte clear, byte set);
	pop hl
	pop de            ; e clear, d set
	push de
	push hl
out_2_change::        ; arg in DE

	ld a, e
	cpl               ; bits to keep
	ld hl, # _out_2

	di

	and (hl)          ; out_2 = (out_2 & ~clear) | set;
	or d
	ld (hl), a
	out (OUT_2),a
	jp di_undo_trash_af

_out_2_toggle::       ; void out_2_toggle(byte bits);
	pop hl
	pop de
	push de
	push hl

	ld a, e           ; bits
	ld hl, # _out_2

	di

	xor (hl)          ; out_2 ^= bits;
	ld (hl), a
	out (OUT_2), a
	jp di_undo_trash_af

;----------------------------------------------------------------------

_pio_a_change::       ; void pio_a_change(byte clear, byte set);
	pop hl
	pop de
	push de
	push hl

	ld a, e
	cpl               ; bits to keep
	ld hl, # _pio_a

	di

	and (hl)          ; pio_a = (pio_a & ~clear) | set;
	or d
	ld (hl), a
	out (PIO_A_DATA), a
	jp di_undo_trash_af

_pio_b_change::       ; void pio_b_change(byte clear, byte set);
	pop hl
	pop de
	push de
	push hl

	ld a, e
	cpl               ; bits to keep
	ld hl, # _pio_b

	di

	and (hl)          ; pio_b = (pio_b & ~clear) | set;
	or d
	ld (hl), a
	out (PIO_B_DATA), a
	jp di_undo_trash_af

_pio_a_change_ddr::   ; void pio_a_change_ddr(byte clear, byte set);
	pop hl
	pop de
	push de
	push hl

	ld a, e
	cpl               ; bits to keep
	ld hl, # _pio_a_ddr

	di

	and (hl)          ; pio_b = (pio_b & ~clear) | set;
	or d
	ld (hl), a

	ld a, # 0xCF        ; ddr-set-command
	out (PIO_A_CTRL), a
	ld a, (hl)
	out (PIO_A_CTRL), a
	jp di_undo_trash_af

_pio_b_change_ddr::   ; void pio_b_change_ddr(byte clear, byte set);
	pop hl
	pop de
	push de
	push hl
pio_b_change_ddr::

	ld a, e
	cpl               ; bits to keep
	ld hl, # _pio_b_ddr

	di

	and (hl)          ; pio_b = (pio_b & ~clear) | set;
	or d
	ld (hl), a

	ld a, # 0xCF        ; ddr-set-command
	out (PIO_B_CTRL), a
	ld a, (hl)
	out (PIO_B_CTRL), a
	jp di_undo_trash_af

;----------------------------------------------------------------------

_sio_a_change::       ; void sio_a_change(byte clear, byte set);
	pop hl
	pop de
	push de
	push hl

	ld a, e
	cpl               ; bits to keep
	ld hl, # _sio_a_wr5

	di

	and (hl)          ; sio_a_wr5 = (sio_a_wr5 & ~clear) | set;
	or d
	ld (hl), a

	ld a, # 5
	out (SIO_A_CSR), a
	ld a, (hl)
	out (SIO_A_CSR), a
	jp di_undo_trash_af

_sio_b_change::       ; void sio_b_change(byte clear, byte set);
	pop hl
	pop de
	push de
	push hl

	ld a, e
	cpl               ; bits to keep
	ld hl, # _sio_b_wr5

	di

	and (hl)          ; sio_b_wr5 = (sio_b_wr5 & ~clear) | set;
	or d
	ld (hl), a

	ld a, # 5
	out (SIO_B_CSR), a
	ld a, (hl)
	out (SIO_B_CSR), a
	jp di_undo_trash_af

_toggle_MON:
	ld a, # SB_MON
	ld hl, # _sio_b_wr5

	di

	xor (hl)          ; sio_b_wr5 ^= SB_MON;
	ld (hl), a

	ld a, # 5
	out (SIO_B_CSR), a
	ld a, (hl)
	out (SIO_B_CSR), a
	jp di_undo_trash_af

;----------------------------------------------------------------------
; MB87006 or MC145156 or ...
;
; interrupts are disabled throughout pll_load()

_pll_strobe::        ; void pll_strobe(byte mask);
	pop hl
	pop de
	push de          ; ra in hl

	ld a, (_out_1)
	xor e            ; pulse to active state
	out (OUT_1), a
	xor e            ; back
	out (OUT_1), a

	jp (hl)          ; ret


_pll_bits::          ; void pll_bits(byte mask, byte bits);
	pop hl
	pop de           ; d bits, e mask
	push de
	push hl

	ld hl, # _out_1

	ld a, #~(O1_SD | O1_CLK)     ; XXX O1_SRE, O1_STE, O1_SCE not touched
	and (hl)
	ld (hl), a
	ld h, a          ; SD and CLK zeroed here

2$:
	ld a, d
	and e            ; bits & mask
	jr z, 1$
	ld a, # O1_SD
1$:
	or h             ; TXON/OFF, TPS, RAS and SxE
	out (OUT_1), a
	or # O1_CLK
	out (OUT_1), a
	and #~O1_CLK
	out (OUT_1), a   ; 1 usec setup/hold/clock/anything

	srl e            ; mask >>= 1
	jp nz, 2$

	ret

;----------------------------------------------------------------------
; PCF2111T or 4035 shifters

cbit:                   ; AF HL trashed
	jr nc, cbit_0
cbit_1:
	ld a, # ~(O2_CLK | O2_DP)
	ld hl, # _out_2

	di

	and (hl)            ; CLK zeroed
	or # O2_DP          ; DP set
	out (OUT_2), a
	ld (hl), a          ; data setup/hold min 3 usec

	or # O2_CLK
	out (OUT_2), a

	and # ~O2_CLK       ; clock pulse min 1 usec
	out (OUT_2), a

	jp di_undo_trash_af ; CLK zeroed, DP left 1

cbit_0:
	ld a, # ~(O2_CLK | O2_DP)
	ld hl, # _out_2

	di

	and (hl)            ; CLK zeroed, DP too
	out (OUT_2), a
	ld (hl), a          ; data setup/hold min 3 usec

	or # O2_CLK
	out (OUT_2), a

	and # ~O2_CLK       ; clock pulse min 1 usec
	out (OUT_2), a

	jp di_undo_trash_af ; CLK zeroed, DP left 0

;----------------------------------------------------------------------
;  74C923, PCF2111T 2pcs and some shifters etc
;
;       display_1[6]
;       display_2[10]
;       scattered icons
;  indicator and lighting leds
;
; car clock --- ---  key  h *
; car phone 88:88:88 book s mast
;         88:8:8888888
;
; each character has 4 segments in "segment group 0"
; and 3 segments in "segment group 1".
;
; bits in each byte of cu53_font[] are ordered 456_3210
;
;            B
;
;     CL  1 2 3
; +   STO 4 5 6
; -   RCL 7 8 9
;     ENT * 0 #
;
;  74C923 gives 0...19 or 0x1F. 1 extra button (marked ?) wireable

_cu53_keymap:
	.ascii "321C654S987R#0*EB?-+"
	.byte 0,0,0,0,0,0,0,0,0,0,0,0

_cu53_dtmf_keys:
	.asciz "0123456789CSRE*#B"

_cu53_alphakeys:
	.ascii "1ABC12DEF23GHI3"
	.ascii "4JKL45MNO56PQR6"
	.ascii "7STU78VWX89 YZ9"
	.asciz "0-/,.*#?0"

;----------------------------------------------------------------------

SELECT = O2_CS1 | O2_CS2
LCD2   = O2_CS1 | O2_CS2
LATCH  = O2_CS1
LCD1   = O2_CS2
KEYPAD = 0

_cu53_keypad::
	in a, (SIO_A_CSR)
	and # SA_DA
	ld l, # 0               ; 0, no key down
	ret nz                  ; if DA low (inversion in sio)

	; tiptoeing, avoid glitching LATCH

cu53_get:
	ld de, # SELECT + LCD1 * 0x100     ; LCD2 to LCD1
	call out_2_change

	ld d, # KEYPAD          ; then KEYPAD
	call out_2_change

	call cbit_0             ; latch the keypad data to shifters

	ld d, # LCD1            ; then LCD1
	call out_2_change

	ld d, e                 ; and finally LCD2
	call out_2_change

;	in a, (PIO_B_DATA)      ; photoresistor state
;	and # PB_DCU
;	ld (darkness), a

	ld de, # 0x0008         ; CY after RL E 5 times. D 0
1$:
	call cbit_0

	in a, (PIO_B_DATA)
	and # PB_DCU
	add # 0xFF
	rl e
	jr nc, 1$

	ld hl, # _cu53_keymap
	add hl, de
	ld l, (hl)              ; ascii key
	ret

;----------------------------------------------------------------------

_cu53_model::               ; the hardwired msbits from another 4035
	call cu53_get           ; get and ignore data from 74923

	ld e, # 0x40            ; CY after RL E 2 times
1$:
	call cbit_0             ; shake out 2 more bits

	in a, (PIO_B_DATA)
	and # PB_DCU
	add # 0xFF
	rl e
	jr nc, 1$

	ld l, e
	ret

;----------------------------------------------------------------------
; just for fun, the no-connects are sent too

_cu53_leds::               ; leds_XXX in A
	rla
	ld d, a
	call cbit              ; available, no connect
	rl d
	call cbit              ; call
	rl d
	call cbit              ; roam
	rl d
	call cbit              ; keylight
	rl d
	call cbit              ; serv
	rl d
	call cbit              ; on
	rl d
	call cbit              ; lcdlight
	rl d
	call cbit              ; bright

	ld de, # SELECT + LATCH * 0x100    ; LCD2 to LATCH
	call out_2_change

	ld d, e                ; still CS1 | CS2, back to LCD2
	jp out_2_change

;----------------------------------------------------------------------
; cumbersome segment layout.
;
; each startbit is the load pulse for the previous group.
; at end, one extra.

_cu53_update::

	ld a, (_keycnt)         ; bottom row ?
	or a
	ld hl, # _display_2
	jr z, 1$
	ld hl, # _buffer        ; show buffer on second row
1$:
	push hl                 ; bottom
	push hl                 ; needed 4 times
	push hl
	push hl

	ld d, # LCD1
	call lcd_start

	pop iy                  ; bottom + 2
	inc iy
	inc iy
	ld b, # 8
	call cu53_bits_0123

	call cbit_0             ; LCD1 group 0

	ld d, # LCD2
	call lcd_start

	ld iy, # _display_1
	ld b, # 6
	call cu53_bits_0123

	pop iy                  ; bottom
	ld b, # 2
	call cu53_bits_0123

	call cbit_0             ; LCD2 group 0

	ld d, # LCD1
	call lcd_start

	ld de, # _icon + 8      ; icon.v_u

	pop iy                  ; bottom + 2
	inc iy
	inc iy
	ld b, # 8
	call cu53_bits_4567

	call cbit_1             ; LCD1 group 1

	ld d, # LCD2
	call lcd_start

	ld de, # _icon + 0      ; icon.colon_ul

	ld iy, # _display_1
	ld b, # 6
	call cu53_bits_4567     ; leaves de = _icon + 6 = &icon.car_d

	pop iy                  ; bottom
	ld b, # 2
	call cu53_bits_4567

	call cbit_1             ; LCD2 group 1

	ld d, # LCD1            ; extra, for LCD1 load pulse
	call lcd_start

	; XXX might as well read the photoresistor here ?

	ld de, # SELECT + LCD2 * 0x100    ; end with idle condition
	jp out_2_change                   ; done.


lcd_start:
	ld e, # SELECT + O2_CLK  ; mask. make sure CLK is low
	call out_2_change        ; select LCD1/LCD2

	jp cbit_0                ; send startbit and RET


cu53_bits_0123:        ; B characters from IY
1$:
	ld l, (iy)         ; *dp
	ld h, #>_cu53_font ; &font[*dp] (page-aligned table)
	ld c, (hl)         ; ____3210
	rr c
	call cbit          ; 0 trashes AF HL
	rr c
	call cbit          ; 1
	rr c
	call cbit          ; 2
	rr c
	call cbit          ; 3

	inc iy
	djnz 1$            ; while (--cc)
	ret

cu53_bits_4567:        ; B characters from IY, B icons from DE
1$:
	ld l, (iy)         ; *dp
	ld h, #>_cu53_font ; &font[*dp] (page-aligned table)
	ld c, (hl)         ; 456_____
	rl c
	call cbit          ; 4 trashes AF HL
	rl c
	call cbit          ; 5
	rl c
	call cbit          ; 6

	ld a, (de)
	add # 0xFF
	call cbit          ; 7th is icon

	inc de
	inc iy
	djnz 1$            ; while (--cc)
	ret

;----------------------------------------------------------------------

_cu53_dtmf::               ; void cu53_dtmf(byte c);
	pop af
	pop bc
	push bc
	push af

	ld a, c
	or a                   ; key break ?
	ret z                  ; nothing to do

	ld hl, # _cu53_dtmf_keys
	ld bc, # 17            ; 17th is 1750 Hz
	cpir                   ; do { A - (HL++); } while (--BC && NZ);
	ret nz                 ; +, - or something strange

	ld a, # 17 - 1
	sub c
	push af                ; arg for dtmf_timer_mt_a()
	inc sp

	call _timer_mt_off
	ld de, # (O0_CCIRC | O0_MTC | O0_MICM) * 0x101  ; all on
	call out_0_change

	call _dtmf_timer_mt_a  ; busywaits DA
	inc sp                 ; purge arg

	ld de, # (O0_CCIRC | O0_MTC | O0_MICM)          ; all off
	call out_0_change
	jp _timer_mt_off       ; done

;----------------------------------------------------------------------
; For more anal probe,
; CU53 routes DP back to DCU via the shifters, delayed by 8 CLKs.

_cu53_probe::
	call _cu53_model  ; 0, 1, 2 or 3
	ld a, l
	or a
	ret z             ; FALSE. handset missing, just a pulldown on DCU XXX

	; 1 ?
	; 2 ?
	; 3 CU53AN (FI keypad ?)

	add #'A
	ld (_display_2 + 8), a

	ld a, # 6
	ld (_display_1_len), a
	ld a, # 10
	ld (_display_2_len), a

	ld hl, # _cu53_alphakeys
	ld (_alphakeys), hl

	ld hl, # _cu53_keypad
	ld (_cu_keypad), hl
	ld hl, # _cu53_update
	ld (_cu_update), hl
	ld hl, # _cu53_leds
	ld (_cu_leds), hl
	ld hl, # _cu53_dtmf
	ld (_cu_dtmf), hl

	ld l, # 1               ; TRUE
	ret

;----------------------------------------------------------------------

_cu58af_leds::             ; leds_XXX in A
	ld bc, # 0x44          ; C = LEDS, B = 0

	rla                    ; available
	jr nc, 7$
	set 7, b               ; no connect in CU58AF
7$:
	rla                    ; call
	jr nc, 6$
	set 1, b
6$:
	rla                    ; roam
	jr nc, 5$
	set 3, b
5$:
	rla                    ; keylight
	jr nc, 4$
	set 5, b
4$:
	rla                    ; serv
	jr nc, 3$
	set 0, b
3$:
	rla                    ; on
	jr nc, 2$
	set 2, b
2$:
	rla                    ; lcdlight
	jr nc, 1$
	set 6, b
1$:
	rla                    ; bright
	jr nc, 0$
	set 4, b               ; no connect in CU58AF
0$:
	jp i2c_put             ; C addr B data

;----------------------------------------------------------------------

_cu_probe::
	call _cu58af_probe       ; try this first, less likely to false
	inc l
	dec l
	call z, _cu53_probe      ; if 0

	ld a, (_display_1_len)
	ld hl, # _display_1
	call add_hl_a
	ld (_display_1_end), hl

	ld a, (_display_2_len)
	ld hl, # _display_2
	call add_hl_a
	ld (_display_2_end), hl
	ret

;----------------------------------------------------------------------

_update_leds::
	ld a, (_leds)       ; desired state, asynchronous
	ld (_leds_XXX), a   ; synchronous state

	ld hl, (_cu_leds)
	jp (hl)             ; leds in A

_display_soon::
	ld a, # 10            ; 50 msec or so
	ld (_dpy_XXX), a      ; XXX snappy S-meter (sigh)
	ret

;----------------------------------------------------------------------

_null_keypad::
	ld l, # 0xFF          ; "cannot tell"
	ret

keypad:
	ld hl, # 0
	ld (keypad_XXX), hl   ; clear timeout

	ld hl, (_cu_keypad)
	rst JPHL

	ld a, l               ; FF or new key (0 if break)
	cp # 0xFF
	ret z                 ; cannot tell
	ld hl, (_key)         ; current active key
	cp l
	ret z                 ; false alarm, same as old key

	or a                  ; chorded ?
	jr z, 1$              ; this is a break
	inc l
	dec l
	jr z, 1$              ; no old key

	ld h, a               ; save new key
	push hl               ; break(old key);
	call _key_break
	pop af                ; new key back
1$:
	ld (_key), a
	or a
	jr z, 2$              ; break

	ld h, a               ; keep a copy
	ld l, a               ; arg
	push hl
	call _key_make        ; make(key);
	pop af
	ld (_last_key), a

	ld bc, # 0x00C0       ; repeat in 3/4 seconds
	ld hl, (timef)
	add hl, bc
	set 0, l              ; must be nz
	ld (_key_timo), hl

	call _lights_on       ; XXX long synch action ?
	jp _display_soon
2$:
	ld a, (_last_key)     ; break(key)
	push af
	inc sp
	call _key_break
	inc sp
	ret

;----------------------------------------------------------------------
; from tick

do_squelch:
	ld a, (_squelch_open)
	neg                        ; carry, if was open
	ld a, (_cfg + sql_method)
	dec a  
	ld a, (_ad_sql)            ; 0
	jp m, 1$
	ld a, (_ad_rssi)           ; 1
	jr z, 1$
	ld a, (_ad_in7)            ; 2
1$:
	ld hl, # _cfg + sql_level
	jr c, 2$                   ; if was open

	cp (hl)                    ; ad - open_level
	jp c, rewind_squelch       ; below open level, stay and rewind

	ld bc, # 4                 ; XXX msec, approx XXX -4 and ADD ADD
	ld hl, (squelch_timo)
	add hl, bc                 ; clears carry unless strange timeout
	ld (squelch_timo), hl
	ld bc, (_cfg + sql_open)   ; msec
	sbc hl, bc
	ret c                      ; waiting to open

	ld a, (_cfg + sql_light)
	or a
	call nz, _lights_on

	call _display_soon    ; XXX snappy S-meter (sigh)

	jp _open_squelch

2$:
	sub (hl)                   ; ad - open_level
	jp nc, rewind_squelch      ; not below, stay and rewind
	ld hl, # _cfg + sql_hyst
	add (hl)                   ; exceeding hystereris ?
	jp c, rewind_squelch       ; not, stay and rewind

	ld a, (_cfg + sql_bi_rssi)
	ld hl, (donald_rssi)       ; rssi 31 msec ago in rssi_log[]
	cp (hl)
	jp c, _close_squelch       ; abrupt close

	ld bc, # 4                 ; XXX msec, approx
	ld hl, (squelch_timo)
	add hl, bc
	ld (squelch_timo), hl
	ld bc, (_cfg + sql_close)  ; msec
	sbc hl, bc
	ret c                      ; waiting to close
	jp _close_squelch

rewind_squelch::
	ld hl, # 0
	ld (squelch_timo), hl
	ret

_snap_squelch::                ; force quick reaction
	ld hl, # 0xFFFF - 4        ; compensation for += 4 in do_squelch
	ld (squelch_timo), hl
	ret

;----------------------------------------------------------------------
; from tick

do_ccir:                   ; new tone in A
	ld hl, (ccir_ptr)

	cp # 0xE
	jr c, 3$               ; normal character,
	jr z, 2$               ; repeat character or else notone

	ld a, l                ; small buf, under 256 bytes. LSbyte is enough
	cp #<ccir_bufe
	jr z, 1$               ; truncated message, discard
	sub #<ccir_buf         ; length
	ret z                  ; nothing in buffer (should not happen)

	ld hl, (timef)         ; range 127 seconds
	ld bc, (ccir_timo_1)
	and a
	sbc hl, bc               ; duration
	ld bc, # -(256 * 3 / 10) ; 300 msec ?
	add hl, bc
	jp nc, 1$              ; too quick, the normal case

	ld hl, (time)          ; range 32767 seconds
	ld bc, (ccir_timo_2)   ; coarse starting time
	and a
	sbc hl, bc             ; duration
	ld bc, # -5            ; 5 seconds ?
	add hl, bc
	jp c, 1$               ; too long

	ld l, a                ; len
	ld h, # 1              ; CCIR
	push hl
	ld hl, # ccir_buf
	push hl
	call _incoming_b       ; (buf, len, CCIR)
	pop af
	pop af
1$:
	ld hl, # ccir_buf      ; rewind
	ld (ccir_ptr), hl
	ret                    ; done
2$:
	dec hl
	ld a, (hl)             ; previous character
	inc hl
	or a
	jr nz, 4$              ; does exist
	ld a, # 0xE            ; funny, starts with an E
3$:
	cp # 10                ; make ascii
	sbc # 0x69
	daa
4$:
	ld (hl), a
	ld a, l
	cp #<ccir_bufe         ; buffer full ?
	ret z                  ; yes
	inc hl
	ld (ccir_ptr), hl
	cp #<ccir_buf          ; first character ?
	ret nz                 ; timers already set. done

	ld hl, (timef)
	ld (ccir_timo_1), hl   ; fine and

	ld hl, (time)
	ld (ccir_timo_2), hl   ; coarse starting time
	ret

;----------------------------------------------------------------------
; from tick

do_dtmf:                    ; cfg.dtmf_decoder in A
	call read_dtmf          ; StD b3...b0 to A

	ld hl, # dtmf_old
	cp (hl)
	jr nz, 1$               ; something changed

	ld hl, (_dtmf_timo)     ; no change, timeout ?
	bit 0, l
	ret z                   ; not set
	ld bc, (timef)
	sbc hl, bc
	ret p                   ; not yet
	jp _dtmf_timeout
1$:
	ld (hl), a              ; new StD status and/or character bits
	push af
	inc sp
	call _dtmf_change
	inc sp
	ret

read_dtmf:
	dec a
	jp z, _read_8870   ; 1

	.IF L8M            ; no free chipselects on 74138

	xor a              ; 2 "nothing"
	ret

	.ELSE

	jp _read_8888      ; 2

	.ENDIF

;----------------------------------------------------------------------

_buf_alloc::
	di

	ld hl, (_buf_free)     ; p ?
	ld a, h
	or l
	jp z, di_undo          ; NULL.

	ld e, (hl)             ; next
	ld (hl), # 0
	inc hl
	ld d, (hl)
	ld (hl), # 0           ; p->next = NULL
	dec hl
	ld (_buf_free), de

	jp di_undo             ; still NZ

_buf_relse::
	pop de
	pop hl             ; buf
	push hl
	push de
buf_relse::

	di

	ld de, (_buf_free)
	ld (_buf_free), hl

	ld (hl), e
	inc hl
	ld (hl), d

	jp di_undo

;----------------------------------------------------------------------

_ad_batt_millivolt::   ; 61 * ad_batt is approx millivolts
	ld hl, (_ad_batt)
	ld h, # 0
	ld e, l            ; ad_batt in E
	ld d, h
	add hl, hl         ; 2
	add hl, de         ; 3
	add hl, hl         ; 6
	add hl, de         ; 7
	add hl, hl         ; 14
	add hl, de         ; 15
	add hl, hl         ; 30
	add hl, hl         ; 60
	add hl, de         ; 61  millivolt in HL
	ret

; from tick_1Hz

check_battery:
	call _ad_batt_millivolt     ; happens to clear carry

	ld bc, (_cfg + batt_cutoff) ; XXX silly millivolts
	sbc hl, bc                  ; not that 1mV would much matter

	ld hl, # batt_timo
	jr c, 1$               ; low battery

	ld (hl), # 0           ; battery ok. rewind to 256 sec (4+ minutes)

	ld hl, # _alert        ; clear buzzer
	res 1, (hl)            ; alert.1 low battery

	ld a, # 253            ; 15+ volts
	cp e                   ; ad_batt
	ret nc                 ; no overvoltage

	jr 2$                  ; no buzzer or powerdown, only blink
1$:
	dec (hl)
	call z, _power_off     ; low battery for too long time

	ld hl, # _alert        ; set buzzer
	set 1, (hl)            ; alert.1 low battery
2$:
	ld hl, # _leds
	res LED_ON, (hl)       ; ON blinks at 1Hz
	ret

;----------------------------------------------------------------------

nmea_input::
	ld hl, (nmea_ptr)

	cp # 0x20
	jr c, nmea_line        ; CR or LF
	cp # '$
	jr z, 1$               ; start of sentence

	ld (hl), a
	inc l
	ret m                  ; discard mode or truncated at 128 bytes
	ld (nmea_ptr), hl
	ret
1$:
	ld hl, # _nmea_buf     ; in collect mode
	ld (nmea_ptr), hl
	ret

nmea_line:
	cp # 0x0D              ; CR ?
	jr z, 2$
	cp # 0x0A              ; LF ?
	jr nz, 1$
2$:
	ld a, l                ; length 10 ... 110 ?
	sub # 10
	cp # 100
	jr nc, 1$

	ld (hl), # 0           ; \0
	ld hl, # _nmea_buf

	ld a, (hl)             ; talker AA ... ZZ ?
	sub # 'A
	cp # 'Z-'A+1
	jr nc, 1$
	ld a, (_nmea_buf+1)
	sub # 'A
	cp # 'Z-'A+1
	jr nc, 1$

	call nmea_split        ; from HL, field pointers into 'nmea'
	ld a, h
	or l
	call z, _process_nmea
1$:
	ld hl, # nmea_bufend   ; in discard mode
	ld (nmea_ptr), hl
	ret

;----------------------------------------------------------------------

	.IF 0

; idle state of out_2.CS1 determines output polarity:
; if 1, normal asynch TTL
; if 0, "sort of RS232" with 0/5 volts
;
; fixed to 38400 baud, 105T or 210T

_cs1_putc::             ; XXX P8x could use O2_DP
	pop de              ; L8x does not have dedicated "DP", just one "SD"
	pop bc              ; byte in C
	push bc
	push de

	ld b, # 10          ; start, 8 data, stop
	sla c               ; startbit

	di                  ; XXX
1$:
	ld hl, # _out_2     ; 10/11 here, just for timing

	rr c                ;  8/10
	ccf                 ;  4/5  zerobits will change the pin to active state
	sbc a, a            ;  4/5
	and # O2_CS1        ;  7/8
	xor (hl)            ;  7/8
	out (OUT_2), a      ; 11/12 only the pin is updated

	ld a, # 0           ;  7/8  dummy, 105T on P8N
	ld a, (_logic)      ; 13/14
	cp # 0              ;  7/8  P8E ?
	call z, 2$          ; 10/18 burn some more time

	scf                 ;  4/5  stopbits march in
	djnz 1$             ; 13/14

	jp di_undo_trash_af ; done XXX it took 260 usec

2$:                     ; P8E must burn 210T in total
	ex (sp), ix         ; /25
	ex (sp), ix         ; /25
	inc ix              ; /12
	dec ix              ; /12
	nop                 ; /5
	nop                 ; /5
	ret                 ; /11

; assume coming from pio_b_int
; XXX hopeless

_dcu_getc::

	; XXX busywait until at the middle of bit 0

	ld hl, # PB_DCU * 256 + 0x80   ; mask and shifter
1$:
	ld a, (_logic)
	cp # 0              ; P8E ?
	call z, 2$          ; burn some more

	in a, (PIO_B_DATA)  ; sample PB_DCU
	and h               ; isolate and
	sub h               ;  invert to carry. RS232 polarity
	rr l
	jp nc, 1$

	; ignore stopbit

	ld a, # 0x97        ; pio_b_icw. all falling XXX edges of
	out (PIO_B_CTRL), a ; the character have triggered interrupt.
	ld a, # ~PB_DCU     ; flush it.
	out (PIO_B_CTRL), a ; XXX works this way ?
	ret


2$:
	ret

;----------------------------------------------------------------------

_cointoss::
	ld a, r
	and # 1
	ld l, a
	ret

	.ENDIF

;----------------------------------------------------------------------

	.area _GSINIT
gsinit:              ; SDCC collects initialisations here

;----------------------------------------------------------------------
	.area _DATA      ; org 0xE200
_data::

_iv:                 ; aligned
	.word 0,0,0,0
	.word 0,0,0,0
_iv_sio::
	.word sio_b_tbe
	.word sio_b_esc
	.word sio_b_rca
	.word sio_b_src
	.word sio_a_tbe
	.word sio_a_esc
	.word sio_a_rca
	.word sio_a_src
_iv_pio_a::
	.word pio_a_int  ; in ram, easy to change
_iv_pio_b::
	.word pio_b_int

	.word 0          ; safety first,
	.word 0          ; reboot if never happens.
	.word 0
	.word 0
	.word 0
	.word 0
	.word 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
	.word 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
	.word 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
	.word 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
	.word 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
	.word 0,0,0,0, 0,0,0,0, 0,0,0,0, 0,0,0,0
	.word 0,0,0,0, 0,0,0,0
	.word 0          ; 129 words, last at 0xE300

time_lsb:            ; these together, within a page
	.byte 0
time_frac::
	.byte 0
timef::
	.byte 0
timex::
time::
	.word 0
	.word 0          ; 32.24 time, seconds
	.word 0          ; overflow area, unbounded increment in clock:

DInest:
	.byte 1          ; reset implies DI. inc/dec by disable/restore

dotick:              ; these 3 together in this order
	.byte 0          ; nz if leave_zero must call tick() XXX bit would do
in_int:
	.byte 0          ; nz if in interrupt
multitask:
	.byte 1          ; number of runnable. msbits are flags

sp_ptr:
	.word _TASK_0    ; pointer to array of saved SPs

adc_reader:
	.word adc_first

donald_rssi:
	.word rssi_log   ; rssi 31 msec ago, for bi-level squelch

break:
	.word _end

_sio_a_rx_ins: .word _sio_a_rx_buf
_sio_a_rx_rem: .word _sio_a_rx_buf

_sio_a_tx_ins: .word _sio_a_tx_buf
_sio_a_tx_rem: .word _sio_a_tx_buf
_sio_a_tx_active: .byte 0

_sio_b_rx_ins:: .word _sio_b_rx_buf
_sio_b_rx_rem:: .word _sio_b_rx_buf

_sio_b_tx_ins:: .word _sio_b_tx_buf
_sio_b_tx_rem:: .word _sio_b_tx_buf
_sio_b_tx_active:: .byte 0

sio_a_input: .word nmea_input  ; sink of sio a rx bytes

nmea_ptr:    .word nmea_bufend ; in discard mode
ccir_ptr:    .word ccir_buf


_cu_keypad:: .word _null_keypad
_cu_update:: .word _nop
_cu_leds::   .word _nop
_cu_dtmf::   .word _nop
_cu_volume:: .word 0              ; only internal volume mux


_out_0::     .byte O0_INH | O0_MICM
_pio_a::     .byte 0
_pio_a_ddr:: .byte 0xFF

_pio_b_in::  .byte PB_EXIN2

; 3 dummy zerobytes below.
; trying to keep variables at same addresses, just for fun.

	.IF P8x
_logic::     .byte 1    ; P8N, but probed at boot

_out_1::     .byte O1_TPS | O1_TXOFF
_out_2::     .byte O2_CS1 | O2_CS2 | O2_SMEM | O2_ROM0
_pio_b::     .byte 0
_pio_b_ddr:: .byte PB_DCU | PB_EXIN1 | PB_EXIN2 | PB_TMR0 | PB_RXON
_fx429_cmd:: .byte 0
	.ENDIF

	.IF L8M
_logic::     .byte 2    ; L8M
             .byte 0

_out_1::     ; same as out_2
_out_2::     .byte O1_TPS | O1_TXOFF | O2_CS1 | O2_CS2 | O1_STE
_pio_b::     .byte PB_SMEM
_pio_b_ddr:: .byte PB_DCU | PB_EXIN1 | PB_EXIN2
             .byte 0
	.ENDIF

	.IF L8TM
_logic::     .byte 3    ; L8TM
             .byte 0

_out_1::     ; same as out_2
_out_2::     .byte O1_TPS | O1_TXOFF | O2_CS1 | O2_CS2
_pio_b::     .byte PB_SMEM
_pio_b_ddr:: .byte PB_DCU | PB_EXIN1 | PB_EXIN2 | PB_AVAIL
_bank::      .byte BNK_RXCA_RXC
	.ENDIF

_display_1_len:: .byte 16
_display_2_len:: .byte 16
_display_1_end:: .word _display_1 + 16
_display_2_end:: .word _display_2 + 16

;----------------------------------------------------------------------
	.area _BSS
bss::

_ad_rssi::      .ds 1  ; keep these eight together
_ad_sql::       .ds 1  ; ad_xxx in channel order (just for fun)
_ad_batt::      .ds 1  ; (wrong order with L8x)
_ad_tpc::       .ds 1
_ad_fpm::       .ds 1
_ad_rpm::       .ds 1
_ad_tp4::       .ds 1  ; ad_rfc AD_RFC on L8TM (L8x?)
_ad_in7::       .ds 1

_ad_rssi_peak::      .ds 1  ; peak value and downcount
_ad_rssi_peak_timo:: .ds 1

rssi_log:       .ds 8  ; *donald_rssi is 31 msec ago
rssi_log_end:

ctcss_phacc:    .ds 2
ctcss_phinc:    .ds 2
ctcss_ogain:    .ds 1

dcs_ptr:        .ds 2  ; into dcs_wave[337]

dtmf_old:       .ds 1
ccir_old:       .ds 1
                .ds 1  ; previous is zero, if buffer is empty
ccir_buf:       .ds 60
ccir_bufe:      .ds 1
ccir_timo_1:    .ds 2
ccir_timo_2:    .ds 2

batt_timo:      .ds 1  ; seconds downcount
light_timo::    .ds 2
squelch_timo::  .ds 2

keypad_XXX:     .ds 2

_dpy_XXX::      .ds 1
_leds_XXX:      .ds 1
_leds::         .ds 1  ; follows leds_XXX
_alert::        .ds 1

slugf:  .ds 1       ; 1 every 4ms
        .ds 1       ; counts up
        .ds 1       ; 1 every second

_fido::
	.ds 2           ; watchdog upcount

             .ds 11  ; overflow areas for crazy numbers
_display_1:: .ds 16
             .ds 11
_display_2:: .ds 16
             .ds 11

_icon::      .ds 16  ; byte per icon

_buffer::    .ds 16
             .ds 1

_keycnt::    .ds 1
_key::       .ds 1
_last_key::  .ds 1

_alpha::     .ds 1
_alphamode:: .ds 1

;----------------------------------------------------------------------
