
	.module i2c

	.include "iomap.inc"

PB_SDA = PB_DCU
O2_SCL = O2_CLK

	.area _CODE

bit:
	ld a, # 0xCF         ; ddr-set-command

	ld hl, # _pio_b_ddr

	di

	out (PIO_B_CTRL), a
	ld a, (hl)           ; here, 1 means input (lucky this time)
	and # ~PB_SDA
	or e                 ; output bit in E, 0 or PB_SDA
	ld (hl), a           ; change is permanent
	out (PIO_B_CTRL), a
rbit:                    ; rbit(), PB_SDA must be in input mode
	di                   ; superfluous, if via bit:

	ld a, (_out_2)
	ld h, a
	or # O2_SCL
	out (OUT_2), a       ; SCL high

	in a, (PIO_B_DATA)   ; XXX maybe cheat and read before CLK pulse ?
	ld l, a              ; return bit in L, not masked yet

	ld a, h
	out (OUT_2), a       ; SCL low

	jp di_undo_trash_af

scl_low:
	ld de, # O2_SCL
	jp out_2_change      ; keeps BC (DE)

scl_high:
	ld de, # O2_SCL * 0x101
	jp out_2_change

sda_pull:
	ld de, # PB_SDA
	jp pio_b_change_ddr  ; keeps BC (DE)

sda_release:
	ld de, # PB_SDA * 0x101
	jp pio_b_change_ddr


start_r:                 ; C addr
	set 0, c             ; READ address
start_w:                 ; WRITE address is default
	call scl_high        ; keeps BC
	call sda_release     ; keeps BC

	ld d, # 10           ; retries
1$:
	in a, (PIO_B_DATA)
	and # PB_SDA
	jp nz, 2$            ; start can be created, continue

	dec d
	ret m                ; my give up, flags (S and) NZ for error

	push de
	call scl_low         ; wriggle out from stuck condition
	call scl_high
	pop de
	jr 1$                ; recheck
2$:
	call sda_pull        ; start.
	call scl_low         ; then fall to write, for address byte

write:                   ; byte from C
	ld d, # PB_SDA       ; constant for the loop below
	scf
	rl c                 ; 1 in lsbit, first bit in CY
1$:
	sbc a, a             ; 0 or FF
	and d                ; 0 or PB_SDA
	ld e, a
	call bit             ; keeps BC DE
	ld a, l
	and d
	cp e                 ; in == out ?
	ret nz               ; contested bit, fail NZ
	sla c
	jp nz, 1$            ; loop for 8 databits

	ld e, d              ; release SDA for ack
	call bit             ; keeps BC DE
	ld a, l
	and d                ; NZ if no ack
	ret

_i2c_get::               ; status i2c_get(byte addr);
	pop hl
	pop bc               ; address in C
	push bc
	push hl

	call start_r         ; ack of address byte leaves PB_SDA as input
	jr nz, bad

	ld de, # 0xFF00 + PB_SDA ; constants for the loop below
	ld c, # 1            ; collect bits
1$:
	call rbit            ; keeps BC DE
	ld a, l
	and e                ; isolate and
	add d                ;  push PB_SDA to CY
	rl c
	jp nc, 1$            ; loop for 8 databits

	call rbit            ; nak, SDA still input. keeps BC DE
	ld a, l
	and e                ; in == out ? now NZ if OK
	jp nz, stop          ; return C
	jr bad

_i2c_put::               ; status i2c_put(byte addr, byte data)
	pop hl
	pop bc               ; C addr B databyte
	push bc
	push hl
i2c_put::                ; C addr B databyte

	call start_w         ; keeps B
	ld c, b
	call z, write
	jp z, good           ; return 0
bad:
	ld c, # 0xFF         ; "not 0" for writes, "inactive 8574" for read
	jr stop

_i2c_write::             ; status i2c_write(byte addr, byte n, byte *p)
	pop de
	pop bc               ; C addr B n
	pop iy               ; IY p
	push iy
	push bc
	push de

	call start_w         ; keeps B
	jr nz, bad
1$:
	ld c, (iy)
	call write           ; keeps B
	jr nz, bad
	inc iy               ; p++
	djnz 1$              ; do while --n
good:
	ld c, # 0            ; good return
stop:                    ; return value in C
	call sda_pull        ; keeps BC
	call scl_high
	call sda_release     ; stop. XXX check if stuck

	.IF L8x              ; has only one serial clock,
	call scl_low         ; and other parts require it to idle low
	.ENDIF

	ld l, c              ; return value in L
	ret

