
#ifdef SDCC
#define ior(ioa)       (ioa)
#define iow(ioa, byte) (ioa) = (byte)

#define IO_ADDR(lab, io) __sfr __at io lab;
#else
#define IO_ADDR(lab, io) enum { lab = io; };
#endif


#ifdef P8x
#include "iomap_P8x.h"
#endif

#ifdef L8TM
#define L8x 1
#include "iomap_L8TM.h"
#endif

#ifdef L8M
#define L8x 1
#include "iomap_L8M.h"
#endif

#define out_0_set(what)    out_0_change((what), (what))
#define out_0_clr(what)    out_0_change((what), 0)

#define out_1_set(what)    out_1_change((what), (what))
#define out_1_clr(what)    out_1_change((what), 0)

#define out_2_set(what)    out_2_change((what), (what))
#define out_2_clr(what)    out_2_change((what), 0)

#define pio_a_set(what)    pio_a_change((what), (what))
#define pio_a_clr(what)    pio_a_change((what), 0)

#define pio_a_input(what)  pio_a_change_ddr((what), (what))
#define pio_a_output(what) pio_a_change_ddr((what), 0)

#define pio_b_set(what)    pio_b_change((what), (what))
#define pio_b_clr(what)    pio_b_change((what), 0)

#define pio_b_input(what)  pio_b_change_ddr((what), (what))
#define pio_b_output(what) pio_b_change_ddr((what), 0)

extern void
	out_0_change(byte clear, byte set),
	out_1_change(byte clear, byte set),
	out_2_change(byte clear, byte set),
	out_2_toggle(byte bits),

	pio_a_change(byte clear, byte set),
	pio_b_change(byte clear, byte set),
	pio_a_change_ddr(byte clear, byte set),
	pio_b_change_ddr(byte clear, byte set),

	sio_a_change(byte clear, byte set),
	sio_b_change(byte clear, byte set),
	sio_b_toggle(byte bits),

	LPF(word hz),
	timer_mt(word hz),
	timer_mt_off();

#define TMR_INTTC    0x00	/* mode 0, intr on terminal cnt */
#define TMR_ONESHOT  0x02	/* mode 1, one shot */
#define TMR_RATEGEN  0x04	/* mode 2, rate generator */
#define TMR_SQWAVE   0x06	/* mode 3, square wave */
#define TMR_SWTRIGG  0x08	/* mode 4, s/w triggered strobe */
#define TMR_HWTRIGG  0x0A	/* mode 5, h/w triggered strobe */
#define TMR_LATCH    0x00
#define TMR_LSB      0x10	/* r/w counter LSB */
#define TMR_MSB      0x20	/* r/w counter MSB */
#define TMR_BOTH     0x30	/* r/w counter 16 bits, LSB first */

#define WR0_SEND_ABORT			0x08
#define WR0_RESET_ESCINT		0x10
#define WR0_CHANNEL_RESET		0x18
#define WR0_REARM_1STC			0x20
#define WR0_RESET_TXINT			0x28
#define WR0_ERROR_RESET			0x30
#define WR0_RETINT				0x38
#define WR0_RESET_RXCRC			0x40
#define WR0_RESET_TXCRC			0x80
#define WR0_RESET_UNDERRUN		0xc0

#define WR1_ESCINT_ENB			0x01
#define WR1_TXINT_ENB			0x02
#define WR1_STATUS_VECTOR		0x04
#define WR1_RXINT_DISABLE		0x00
#define WR1_INT_1STC			0x08
#define WR1_INT_ALLC			0x10
#define WR1_INT_ALLC_PNS		0x18
#define WR1_WAITREADY_RT		0x20
#define WR1_WAITREADY_FN		0x40
#define WR1_WAITREADY_ENB		0x80

#define WR3_RX_ENB				0x01
#define WR3_SYNCLOAD_INHIBIT	0x02
#define WR3_ADDRESS_SEARCH_MODE	0x04
#define WR3_RXCRC_ENB			0x08
#define WR3_ENTER_HUNT			0x10
#define WR3_AUTO_ENABLES		0x20
#define WR3_5BIT_RX				0x00
#define WR3_6BIT_RX				0x40
#define WR3_7BIT_RX				0x80
#define WR3_8BIT_RX				0xc0

#define WR4_PARITY_ENB			0x01
#define WR4_PARITY_ODD			0x00
#define WR4_PARITY_EVEN			0x02
#define WR4_SYNC_MODES			0x00
#define WR4_1STOPBIT			0x04
#define WR4_15STOPBIT			0x08
#define WR4_2STOPBIT			0x0c
#define WR4_8BITSYNC			0x00
#define WR4_16BITSYNC			0x10
#define WR4_SDLC_MODE			0x20
#define WR4_EXTSYNC				0x30
#define WR4_X1_CLK				0x00	/* 153600 bd, but... */
#define WR4_X16_CLK				0x40	/*   9600 bd */
#define WR4_X32_CLK				0x80	/*   4800 bd */
#define WR4_X64_CLK				0xc0	/*   2400 bd */

#define WR5_TXCRC_ENB			0x01
#define WR5_RTS					0x02
#define WR5_CCITT_CRC			0x00
#define WR5_CRC16_CRC			0x04
#define WR5_TX_ENB				0x08
#define WR5_SEND_BREAK			0x10
#define WR5_5BIT_TX				0x00
#define WR5_7BIT_TX				0x20
#define WR5_6BIT_TX				0x40
#define WR5_8BIT_TX				0x60
#define WR5_DTR					0x80

#define RR0_RCA					0x01
#define RR0_INT_PENDING			0x02
#define RR0_TBE					0x04
#define RR0_DCD					0x08
#define RR0_SYNC				0x10
#define RR0_CTS					0x20
#define RR0_UNDERRUN			0x40
#define RR0_BREAK_ABORT			0x80

#define RR1_ALL_SENT			0x01

#define RR1_PARER				0x10
#define RR1_ROVER				0x20
#define RR1_FRMERR				0x40
#define RR1_CRCERR				0x40
#define RR1_ENDFRAME			0x80

