
	.module morse

	.include "iomap.inc"
	.include "cfg.inc"

	.area _BSS

_af_tx_timer:: .ds 1   ; 0 means idle, else tick count to next slot
af_tx_pc::     .ds 2   ; state

morse_bits:    .ds 1   ; character pattern
morse_slot:    .ds 1   ; timing unit in ticks
morse_tone:    .ds 2   ; buzzer timecount
morse_ptr:     .ds 2   ; asciz string

	.area _CODE

do_af_tx::                 ; from tick.
	ld hl, (af_tx_pc)
	jp (hl)                ; returns new count, 0 if ended

af_tx_sent::
	call _af_tx_stop
	ld a, (_af_tx_timer)   ; 0 unless new message was started
	ret

_af_tx_stop::
	xor a
	ld (_af_tx_timer), a    ; block background
	ld (_icon + 14), a      ; crossed_handset

	ld hl, # _crash
	ld (af_tx_pc), hl       ; soita paranoid

	call _CCIRC_off
	call _timer_mt_off

	ld hl, (_cfg + lpf_tx)
	jp LPF                  ; done

af_tx_start::
	ld (af_tx_pc), hl

	ld a, # 1
	ld (_af_tx_timer), a    ; start in next tick()

	ld (_icon + 14), a      ; crossed_handset
	ret

;----------------------------------------------------------------------
; from left, 1 is dit, 0 is dash,
; and pattern ends at rightmost 1

morse_tab:
	.byte 0b00000000 ; ! XXX  control characters and space not in table
	.byte 0b10110110 ; "
	.byte 0b11101010 ; # encodes VA aka SK
	.byte 0b10111100 ; $ encodes AS
	.byte 0b00000000 ; % XXX
	.byte 0b00000000 ; & XXX
	.byte 0b10000110 ; '

	.byte 0b01001010 ; (
	.byte 0b01001010 ; )
	.byte 0b00000000 ; * XXX
	.byte 0b00000000 ; + XXX
	.byte 0b00110010 ; ,
	.byte 0b01111010 ; -
	.byte 0b10101010 ; .
	.byte 0b01101100 ; /

	.byte 0b00000100 ; 0
	.byte 0b10000100 ; 1
	.byte 0b11000100 ; 2
	.byte 0b11100100 ; 3
	.byte 0b11110100 ; 4
	.byte 0b11111100 ; 5
	.byte 0b01111100 ; 6
	.byte 0b00111100 ; 7

	.byte 0b00011100 ; 8
	.byte 0b00001100 ; 9
	.byte 0b00011110 ; :
	.byte 0b00000000 ; ; XXX
	.byte 0b00000000 ; < XXX
	.byte 0b01110100 ; =
	.byte 0b00000000 ; > XXX
	.byte 0b11001110 ; ?

	.byte 0b00000000 ; @ ` XXX upper and lowercase folded
	.byte 0b10100000 ; A a
	.byte 0b01111000 ; B b
	.byte 0b01011000 ; C c
	.byte 0b01110000 ; D d
	.byte 0b11000000 ; E e
	.byte 0b11011000 ; F f
	.byte 0b00110000 ; G g

	.byte 0b11111000 ; H h
	.byte 0b11100000 ; I i
	.byte 0b10001000 ; J j
	.byte 0b01010000 ; K k
	.byte 0b10111000 ; L l
	.byte 0b00100000 ; M m
	.byte 0b01100000 ; N n
	.byte 0b00010000 ; O o

	.byte 0b10011000 ; P p
	.byte 0b00101000 ; Q q
	.byte 0b10110000 ; R r
	.byte 0b11110000 ; S s
	.byte 0b01000000 ; T t
	.byte 0b11010000 ; U u
	.byte 0b11101000 ; V v
	.byte 0b10010000 ; W w

	.byte 0b01101000 ; X x
	.byte 0b01001000 ; Y y
	.byte 0b00111000 ; Z z
	.byte 0b00000000 ; [ { XXX
	.byte 0b00000000 ; \ | XXX
	.byte 0b00000000 ; ] } XXX
	.byte 0b00000000 ; ^ ~ XXX
	.byte 0b00000000 ; _ DEL XXX

morse_nationals:           ; in ascending order from 0x80 to 0xFF
	.byte ',   0b10101000 ; iti
	.byte ',   0b10100110 ; ke
	.byte ',   0b00011000 ; ljy
	.byte ',   0b10101000
	.byte ',   0b10100110
	.byte ',   0b00011000

	.byte 0xFF, 0b00000000 ; end


lookup:
	or a
	jp p, usascii     ; 7-bit character

	ld hl, # morse_nationals
1$:
	cp (hl)           ; character
	inc hl            ; pattern
	jr z, found
	inc hl            ; next character
	jr nc, 1$         ; list is sorted

blank:
	xor a             ; emits intercharacter pause
	ret

usascii:
	cp # 0x60
	jr c, 1$
	and #~0x20        ; lower to upper case
1$:
	sub # 0x21        ; control characters and space not in table
	jr c, blank

	ld hl, # morse_tab
	call add_hl_a
found:
	ld a, (hl)
	ret

;----------------------------------------------------------------------
; XXX word space 5 slots. double blank, if 7 truely needed.

pause:
	pop hl
	ld (af_tx_pc), hl
	ld a, (morse_slot)      ; new downcount.
	ret

begin:                      ; from tick()
	ld hl, (morse_ptr)
	ld a, (hl)
	or a
	jp z, af_tx_sent
	inc hl
	ld (morse_ptr), hl

	call lookup             ; character to pattern
	jr 3$                   ; start from bottom
2$:
	ld (morse_bits), a      ; save rest of pattern. current bit in CY

	ld hl, (morse_tone)     ; 1 kHz or such
	call set_tone           ; keeps CY
	jr c, 1$                ; dit ?
	call pause
	call pause
1$:
	call pause              ; wait 1 or 3 slots

	ld hl, # 4              ; ultrasonic, does not pass LPF, keeps DC offset
	call set_tone

	call pause              ; 1 slot from end of tone

	ld a, (morse_bits)
3$:
	add a, a                ; shift left
	jr nz, 2$               ; while pattern not spent

	call pause              ; delay between characters (and words)
	call pause
	jr begin                ; 1+2 (or 1+2+2) slots from end of tone

;----------------------------------------------------------------------

set_tone:                   ; keeps CY
	di

	ld a, l
	out (TMR_MT), a
	ld a, h
	out (TMR_MT), a

	jp di_undo

;----------------------------------------------------------------------

_morse_send::               ; void morse_send(byte *s);
	ld hl, (_cfg + morse_Hz)
	ld a, h
	or l
	ret z                   ; morse Hz 0 means off

	ld a, (_cfg + morse_cpm)
	or a
	ret z                   ; morse cpm 0 means off

	cp # 20                 ; slot (in ticks) must fit in a byte
	jr nc, 1$
	ld a, # 20
1$:
	ld c, a
	ld b, # 0
	ld hl, # 6 * TICK_HZ    ; morse timing, 6000 / cpm [msec]
	call div_hl_bc
	ld a, l                 ; slot length
	push af

	ld bc, (_cfg + morse_Hz)
	ld de, # TMR_CLK / 65536
	ld hl, # TMR_CLK % 65536
	call div_dehl_bc        ; timeconstant in hl

	xor a
	ld (_af_tx_timer), a    ; block background

	pop af                  ; slot length

	pop de                  ; arguments:
	pop bc                  ; asciz string
	push bc
	push de

	ld (morse_ptr), bc
	ld (morse_slot), a
	ld (morse_tone), hl

	ld a, # TMR_SEL_MT | TMR_BOTH | TMR_SQWAVE
	out (TMR_CTRL), a

	ld hl, (_cfg + morse_Hz)
	call LPF_ccir
	call _CCIRC_on

	ld hl, # begin
	jp af_tx_start

