
#define STUFF /* hide extern */

#include "head.h"
#include "cfg.h"
#include "cu.h"
#include "iomap.h"
                                /* YYMMDD#rev */

const byte version[] = "@(#)R58bis 110318#57"
#ifdef P8x
	" P8E/P8N"
#endif
#ifdef L8M
	" L8M"
#endif
#ifdef L8TM
	" L8TM"
#endif
	;

const word cfg_size = sizeof(cfg);

#ifdef P8x

#define debug_putc sio_a_putc
#define debug_puts sio_a_puts

#endif
#ifdef L8x

#define debug_putc sio_b_putc
#define debug_puts sio_b_puts

#endif

#define TCXO    12800000U
#define TMR_CLK  4032000U

#define MSEC   * 256L / 1000 /* scale arg of timef_mark() */

byte
	sio_a_rr0,
	sio_a_wr5 = WR5_8BIT_TX | WR5_TX_ENB, /* no SA_TXE or SA_nRXE */

	sio_b_rr0,
	sio_b_wr5 = WR5_8BIT_TX | WR5_TX_ENB,

	squelch_open,
	audio_flags,       /* audio muted or opened somehow */
	tx_vco_on,
	ptt_works,
	last_memory = -1,
	transmitting,
	transmit_state,    /* 0 no, 1 no power, 0x80 power on */
	transmit_tail,
	afc, afc_zero;

dword
	rx_freq,
	tx_freq;

signed char
	doppler;

extern word
	iv_sio[],
	iv_pio_a,
	iv_pio_b;

extern byte
	sio_b_rx_buf[],
	sio_b_rx_bufend[],
	sio_b_tx_buf[],
	sio_b_tx_bufend[],

	*sio_b_rx_ins,
	*sio_b_rx_rem,

	*sio_b_tx_ins,
	*sio_b_tx_rem,
	sio_b_tx_active;

extern void
	key_setup(byte),
	quick_setup(byte);

word
	tot,
	key_timo;

/*
 * these are for
 *  CCIR rx
 *  DTMF rx
 *  FX419/FX429 rx and tx
 */

struct buf
	*fx_tx_head, **fx_tx_tail = &fx_tx_head,
	*buf_free,
	buf_space[32];

/* XXX */

void ax_send(byte *pkt, word len); /* synchronous (very) */

byte fx_check(byte *p, word datalen);
void fx_send(byte *pkt, word datalen); /* asynchronous */

void close_audio(), open_audio();
void close_squelch(), open_squelch();
void volume(int add);

/* used by dtmf.s */
 
#define tone(hz) ((0x10000 * (hz) + 7875) / 15750)

const word
dtmf_mt_tab[] = {
	tone(941), tone(1336), /* 0 */
	tone(697), tone(1209), /* 1 */
	tone(697), tone(1336), /* 2 */
	tone(697), tone(1477), /* 3 */
	tone(770), tone(1209), /* 4 */
	tone(770), tone(1336), /* 5 */
	tone(770), tone(1477), /* 6 */
	tone(852), tone(1209), /* 7 */
	tone(852), tone(1336), /* 8 */
	tone(852), tone(1477), /* 9 */
	tone(697), tone(1633), /* A */
	tone(770), tone(1633), /* B */
	tone(852), tone(1633), /* C */
	tone(941), tone(1633), /* D */
	tone(941), tone(1209), /* * */
	tone(941), tone(1477), /* # */

	tone(1750), tone(1750), /* twice, to get full amplitude */
};
#undef tone

#if 1
#define progress(c) /* nothing */

#else

/* printf debugger */

static void
progress(byte c)
{
	extern void cu53_update();

	display_2[0] = c;

	if (cu_update != nop)
		(*cu_update)();
	else
		cu53_update();
}

#endif

extern byte cu58af_buttons;

byte
ptt_pressed()
{
	return ((ior(SIO_B_CSR) & SB_PTT) | (cu58af_buttons & 0x40));
}

byte
off_hook()
{
	return (
#ifdef P8x
	(ior(PIO_A_DATA) & PA_OFFHOOK)
#endif
#ifdef L8x
	(~ior(SIO_A_CSR) & SA_ONHOOK)
#endif
		| (cu58af_buttons & 0x80));
}

static byte
fx_transmitting()
{
#ifdef P8x
	return (fx429_cmd & FX429_TXENB);
#endif
#ifdef L8x
	return (sio_a_wr5 & SA_TXE);
#endif
}

struct pll {
	byte strobe;
	byte A;
	word N;
	word R;
};

/*
 * S8M, the old synthesizer used with L8M
 *
 * RA210    R   channel with 12.8 MHz TCXO
 *   0      8   1600.0
 *   1     64    200.0
 *   2    128    100.0
 *   3    256     50.0
 *   4    640     20.0
 *   5   1000     12.8
 *   6   1024     12.5   default
 *   7   2048      6.25
 *
 * synth=S8M_10 expects modification which halves
 * TCXO and rewires RA to /= 640. step 10 kHz
 */

static void
load_MC145156(struct pll *p)
{
	/* pll_bits() and pll_strobe() do not protect themselves */

	disable();

	pll_bits(0x02, 3);      /* SW1 0x2 and SW2 0x1 */

	pll_bits(0x02, p->N >> 8); /* 10 bits */
	pll_bits(0x80, p->N);

	pll_bits(0x40, p->A);      /* 7 bits */

	if (p->strobe == O1_STE)
		out_1_clr(O1_STE);  /* latch and enable */
	else
		pll_strobe(p->strobe);

	restore();
}

/* all others have */

static void
load_MB87006(struct pll *p)
{
	/* protected throughout, so that vco
	 * does not have so much time to
	 * wander off with new NA and old R.
	 * same reason for the order;
	 * 15 bits between strobes, not 18. */

	disable();

	pll_bits(0x02, p->N >> 8); /* 10 bits */
	pll_bits(0x80, p->N);

	pll_bits(0x40, p->A);      /* 7 bits */

	pll_bits(0x01, 0);      /* control bit '0' */
	pll_strobe(p->strobe);

	pll_bits(0x20, p->R >> 8); /* 14 bits */
	pll_bits(0x80, p->R);

	pll_bits(0x01, 1);      /* control bit '1' */
	pll_strobe(p->strobe);

	restore();
}

#ifdef L8TM
#define PA_SLOCKBE PA_SCE  /* same pin PA.1 on L8M and L8TM */
#endif

#define SC_RX_LOW_BAND 0x01 /* 1 low band, 0 high */
#define SC_TX_LOW_BAND 0x02
#define SC_TX_BITS_1   0x04 /* prediv and pll, 0 on XXX keep supply here ? */
#define SC_TX_BITS_2   0x08 /* vco and amps, 1 on */

#define VBUFF 0x04          /* differs from S8C/D */

static void
synth_setup()
{
	switch (cfg.synthesizer) {

	case SYNTH_S8M_12:
	case SYNTH_S8M_10:
		out_1_set(O1_STE);      /* nSTE */
#ifdef PA_SLOCKBE
		pio_a_set(PA_SLOCKBE);
		pio_a_output(PA_SLOCKBE); /* buffer disable */
#else
		out_1_set(O1_SCE);      /* PA_SLOCKBE */
#endif
		break;

	case SYNTH_COMPUTEC:
#ifdef PA_SLOCKBE
		pio_a_clr(PA_SLOCKBE);      /* PA_SCE */
		pio_a_output(PA_SLOCKBE);
#endif
		break;
	}
}

static void
vbuff_off()
{
	switch (cfg.synthesizer) {

	case SYNTH_S8M_12:
	case SYNTH_S8M_10:
#ifdef PA_SLOCKBE
		pio_a_output(PA_SLOCKBE);
#else
		out_1_set(O1_SCE);
#endif
		break;
	}
}

void
load_synth_ctrl()
{
	byte x;

	switch (cfg.synthesizer) {

	case SYNTH_S8M_12: /* control 4094 does not exist in S8M */
	case SYNTH_S8M_10:

		if (tx_vco_on) {
#ifdef PA_SLOCKBE
			pio_a_input(PA_SLOCKBE);  /* pulled down by /SLOCK */
#else
			out_1_clr(O1_SCE);    /* P8x cannot read /SLOCK info */
#endif
		} else {
#ifdef PA_SLOCKBE
			pio_a_output(PA_SLOCKBE);
#else
			out_1_set(O1_SCE);
#endif
			out_1_set(O1_STE);    /* nSTE */
		}
	default:
		return; /* nothing more */


	case SYNTH_COMPUTEC:
		x = tx_vco_on ? (SC_TX_BITS_2 | VBUFF) : 0;
		break;

	case SYNTH_S8C:
	case SYNTH_S8D:
		x = tx_vco_on ? SC_TX_BITS_2 : SC_TX_BITS_1;
	}

	if (rx_freq < cfg.rx_band_center)
		x |= SC_RX_LOW_BAND;

	if (tx_freq < cfg.tx_band_center)
		x |= SC_TX_LOW_BAND;

	disable();

	pll_bits(0x08, revnybble[cfg.deviation & 15]); /* LSbit goes first */
	pll_bits(0x08, x);

#ifdef PA_SLOCKBE
	pio_a_set(PA_SLOCKBE); /* aka PA_SCE */
	pio_a_clr(PA_SLOCKBE);
#else
	pll_strobe(O1_SCE);
#endif
	restore();
}

/* S8C, S8D, COMPUTEC "16/58/S" */

dword debug_NA;
word debug_R;

dword
vco_frequency_new(byte strobe, dword kHz)
{
	struct pll pll;
	dword NA, tmp;
	byte z = 0;

	NA = kHz;
	switch (ulmoddiv(&NA, 25)) {       /* XXX maybe /= 100 ? */
	case 24:
		z++;
	case 12:
		z++;
	case 0:
		pll.R = TCXO / 12500;
		NA = NA * 2 + z;
		z = 50;
		break;

#if 0 /* 6.25 kHz step omitted. Would create too large N in V->UHF bastard XXX */

	case 18:
		z += 2;
	case 6:
		z++;
		pll.R = TCXO / 6250;
		NA = NA * 4 + z;
		z = 25;
		break;
#endif

	default:
		z = 40;
		pll.R = TCXO / 10000;
		NA = kHz;
		if (ulmoddiv(&NA, 10) != 0) {

			/* no luck with 25, 12.5 or 10 kHz. (6.25 not checked XXX)
			 * search for best possible NA and R.
			 * error is usually under 200 Hz,
			 * but there are bad spots. */

			/* XXX got to be a better way XXX */

			word acc, err, r, rmax, x;

			/* N must fit into 10 bits:
			 * kHz / (TCXO/R) / prescaler <= 0x3FF
			 * or R < 0x400 prescaler TCXO / kHz */

			/* XXX off-by-one ? -1 to be safe */

			rmax = divide_dwdwdw(TCXO / 1000 * 0x400L * cfg.prescaler, kHz);
			rmax--;
			if (rmax > 2048)  /* 6.25 kHz */
				rmax = 2048;

			pll.R = 1024;     /* default to 12.5 */
			err = 0xFFFF;     /* max */

			tmp = 128 * kHz;
			if (ulmoddiv(&tmp, 25) > 12)    /* x = 0x10000 * kHz / TCXO */
				tmp++;
			x = tmp;

			acc = x * 1024;              /* from 12.5 kHz */
			r = 1024;
			do {
				if (err > acc) {         /* XXX or .9999 */
					err = acc;
					pll.R = r;
					if (err < 200)       /* arbitrary threshold */
						goto found;
				}
				acc += x;
			} while (++r < rmax);        /* down to 6.25 */

			acc = x * 1024;              /* again from 12.5 kHz */
			r = 1024;
			while (--r >= 512) {         /* up to 25 kHz */
				acc -= x;
				if (err > acc) {         /* XXX or .9999 */
					err = acc;
					pll.R = r;
					if (err < 200)
						goto found;
				}
			}
		found:
			NA = divide_dwdww(kHz * pll.R, TCXO / 1000);

			z = 0;     /* harder to calculate back to kHz */
		}
	}

	tmp = NA;
	pll.A = ulmoddiv(&tmp, cfg.prescaler);
	pll.N = tmp;
	pll.strobe = strobe;

	load_MB87006(&pll);

	debug_NA = NA;
	debug_R = pll.R;

	if (pll.N <= pll.A || tmp > 0x3FF) /* XXX */
		return (0);

	if (z)
		return (NA * z / 4);     /* resulting kHz */

	return divide_dwdww(TCXO / 1000 * NA, pll.R);
}

/* default S8M, 12.5 kHz */

dword
vco_frequency_S8M_12(byte strobe, dword kHz)
{
	struct pll pll;
	dword NA, tmp;
	byte x, z;

	NA = kHz;
	x = ulmoddiv(&NA, 25);
	z = 0;
	if (x > 6)
		z++;
	if (x > 18)
		z++;
	NA = NA * 2 + z;

	tmp = NA;
	pll.A = ulmoddiv(&tmp, cfg.prescaler);
	pll.N = tmp;
	pll.R = 0;
	pll.strobe = strobe;

	load_MC145156(&pll);

	debug_NA = NA;
	debug_R = 1024;

	if (pll.N <= pll.A || tmp > 0x3FF) /* XXX */
		return (0);

	return (NA * 25 / 2);    /* resulting kHz */
}

/* modified S8M, 10 kHz */

dword
vco_frequency_S8M_10(byte strobe, dword kHz)
{
	struct pll pll;
	dword NA, tmp;

	NA = kHz;
	if (ulmoddiv(&NA, 10) >= 5)
		NA++;

	tmp = NA;
	pll.A = ulmoddiv(&tmp, cfg.prescaler);
	pll.N = tmp;
	pll.R = 0;
	pll.strobe = strobe;

	load_MC145156(&pll);

	debug_NA = NA;
	debug_R = 1280;

	if (pll.N <= pll.A || tmp > 0x3FF) /* XXX */
		return (0);

	return (NA * 10);    /* resulting kHz */
}

dword
vco_frequency(byte strobe, dword kHz)
{
	switch (cfg.synthesizer) {

	case SYNTH_S8M_12:
		return (vco_frequency_S8M_12(strobe, kHz));

	case SYNTH_S8M_10:
		return (vco_frequency_S8M_10(strobe, kHz));

	/* SYNTH_S8C:
	   SYNTH_S8D:
	   SYNTH_COMPUTEC: */

	default:
		return (vco_frequency_new(strobe, kHz));
	}
}

void
receive(dword kHz)
{
	dword old = rx_freq;    /* to estimate recovery time */

	rx_freq = kHz + doppler;    /* XXX S8M: frequency dependent SW1, SW2 (?) */

	RFC_update();          /* before longwinded calculations */

	/* adjust rx_freq to selected channel */

	rx_freq = vco_frequency(O1_SRE, rx_freq + cfg.if_offset)
	        - cfg.if_offset;

	load_synth_ctrl();     /* rx a/b vco-band for S8D etc */

	/* XXX how long to wait ? */

	if (labs(old - rx_freq) > 10000
	 || (old < cfg.rx_band_center) != (rx_freq < cfg.rx_band_center)) {

		nop(); /* XXX */
	}
	ad_rssi_peak_timo = 1; /* forget peak */
}

const dword tx_bands[] = {
	  28000,   30000,      /* start and end kHz in ascending order */
	  50000,   54000,
	  70000,   70500,
	 144000,  148000,
	 222000,  225000,      /* XXX 219 ... 220 ? */
	 420000,  450000,
	 902000,  928000,
	1240000, 1300000,
	     -1,              /* end */
};

byte legal_tx(dword kHz); /* cfg.tx_band_start/_end and tx_bands[] */

void
transmit_off()
{
	word mark;

	encoder_turnoff();      /* XXX just testing */

	iow(DA_TPC, 0);
	vbuff_off();

	mark = timef_mark(8 MSEC);
	tot = 0;

	out_0_set(O0_MICM);
	dtmf(0);
	af_tx_stop();              /* XXX */
	encoder_end();

	LPF(cfg.lpf_rx);           /* when receiving, help ax input (or such) */

	transmit_state = 0;
	tx_led_off();
	update_leds();
	display();

	while (! timef_passed(mark))
		yield();

	out_1_change(O1_TXON | O1_TXOFF, O1_TXOFF);

	tx_vco_on = 0;
	load_synth_ctrl();

	audio_flags &= ~0x80;  /* sql_open msec follows */

	if (audio_flags & 3)   /* monitor pressed or toggled on */
		open_audio();

	RFC_update();      /* was used by encoders ? double duty with TPC ? */

	transmit_tail = 2;
	transmitting = 0;
	alert &= ~1;
}

static void
select_tx_freq()
{
	if (cfg.satellite)
		tx_freq = cfg.tx_offset - doppler / 3; /* XXX */
	else
		tx_freq = cfg.rx_freq + cfg.tx_offset;
}

void
transmit_on()
{
	word mark;

	transmitting = 1;
	transmit_state = 1;

	iow(DA_TPC, 0);

	select_tx_freq();

	if (!cfg.fdx) {
		audio_flags |= 0x80; /* half-duplex tx disables squelch */
		close_audio();
		close_squelch();
	}

	out_1_change(O1_TXON | O1_TXOFF, O1_TXOFF);

	tx_vco_on = 1;
	load_synth_ctrl(); /* tx a/b vco-band, supplies, deviation */

	tx_freq = vco_frequency(O1_STE, tx_freq);

	/* overlap the delay with useful stuff */

	mark = timef_mark(25 MSEC);

	LPF(cfg.lpf_tx);      /* tx audio lowpass */

	/* power only if amateur band */

	tot = time_mark(cfg.max_over);

	if (cfg.max_over && legal_tx(tx_freq)) {
		transmit_state = 0x80;
		tx_led_on();
		display();
		update_leds();

		if (cfg.ctcss_gain && cfg.ctcss)
#if 0
			dcs(cfg.ctcss);
#else
			ctcss(cfg.ctcss);
#endif

		out_1_change(O1_TXON | O1_TXOFF, O1_TXON);

		while (! timef_passed(mark))
			yield();

		iow(DA_TPC, cfg.tpc);
	} else {
		alert |= 1;
	}
}

static void
qsy()
{
	disable();
	close_squelch();
	restore();

	select_tx_freq();

	receive(cfg.rx_freq);

	/* XXX snap_squelch for quick decision ? */
}

static void
frequency(dword kHz)
{
	cfg.rx_freq   = kHz;
	cfg.tx_offset = 0;        /* XXX duplex band(s) */
	cfg.ctcss     = 0;

	doppler = 0;
	afc = 0;
	qsy();
}

static void
memory(word n)
{
	struct memory *p = &memories[n];

	if (p->rx_freq) {
		cfg.rx_freq   = p->rx_freq & 0xFFFFFF; /* XXX SDCC-3.0 has fixed a bug (?) */
		cfg.tx_offset = p->tx_offset;
		cfg.ctcss     = p->ctcss;

		last_memory = n;

		doppler = 0;
		afc = 0;
		qsy();
	}
}

byte keying_dtmf; /* latches break and timeout actions */

void
key_make(byte c)
{
	if ((keying_dtmf = ptt_pressed()) != 0) {
		dtmf(c);
		return;
	}
	if (setup_active && c != 'B') {
		key_setup(c);
		return;
	}
	if (isdigit(c)) {
		if (keycnt < display_2_len) {
			buffer[keycnt++] = c;
			bzero(buffer + keycnt, sizeof(buffer) - keycnt); /* XXX */
		}
		return;
	}
	switch (c) {
	case 'C':
		if (keycnt)
			buffer[--keycnt] = 0;
		break;

	case 'E':
		keycnt = 0;
		setup_active = 1;
		break;

	case '#':
		switch (keycnt) {
		case 0:
			break;

		case 1:
		case 2:
			memory(atow(buffer));
			break;

		case 3:
			frequency(atodw_seeded(buffer, cfg.implied));
			break;
		case 4:
			frequency(atodw_seeded(buffer, cfg.implied / 10));
			break;
		default:
			frequency(atodw(buffer));
		}
		keycnt = 0;
		break;

	case 'S':
		timer_mt(0);
		out_0_set(O0_MTC);
		silly();
		out_0_clr(O0_MTC);
		timer_mt(0);
		break;

	case 'R':
		if (afc) {
			afc = 0;
		} else {
			afc = 1;
			afc_zero = ad_in7;
		}
		break;

	case '*':
		keycnt = 0;
		break;

	case '+':
		volume(+1);
		break;

	case '-':
		volume(-1);
		break;

	case 'B':
		audio_flags |= 2; /* monitor while B pressed */
		open_audio();

		if (cfg.tx_offset && !cfg.satellite)
			receive(cfg.rx_freq + cfg.tx_offset);
		break;
	}
}

void
key_break(byte c)
{
	if (keying_dtmf) {
		dtmf(0);
		return;
	}
	switch (c) {
	case 'B':
		if (cfg.tx_offset && !cfg.satellite)
			receive(cfg.rx_freq);

		disable();
		audio_flags &= ~2;
		if (key_timo)
			audio_flags ^= 1; /* quick B, toggle */

		if (!audio_flags && !squelch_open)
			close_audio();

		restore();

		icon.star = (audio_flags & 1);
		display_soon();
		break;
	}

	setup_active &= ~2;  /* out of quick adjust */
}

/* if the key had been released before timeout,
 * 'key' is 0 but last_key still something */

void
key_timeout()
{
	byte d = 0;

	key_timo = 0;

	if (keying_dtmf) { /* XXX re-enable mic ? */
		return;
	}
	if (alphamode) {
		alphamode = 0;
		d = 1;
	}

	/* long keypresses */

	switch (key) {
	case 'C':
		if (keycnt) {
			keycnt = 0;
			d = 1;
		}
		break;

	case '-':
		if (!setup_active && cfg.volume == 1) /* no autorepeat to mute */
			break;
	case '+':
		key_make(key);
		d = 1;
		key_timo = timef_mark(200 MSEC); /* from now, break or no */
		break;

	/* XXX */

	case '1':
	case '4':
		keycnt = 0;
		setup_active |= 2;
		quick_setup(key);
		d = 1;
		key_timo = timef_mark(200 MSEC); /* from now, break or no */
		break;

	case '2':
	case '5':
		{
			byte n = last_memory;

			do {
				if (key == '2') {
					if (++n >= N_MEMORIES)
						n = 0;
				} else {
					if (--n >= N_MEMORIES)
						n = N_MEMORIES - 1;
				}
				if (memories[n].rx_freq) {
					memory(n);
					break;
				}
			} while (n != last_memory);
		}
		keycnt = 0;
		d = 1;
		key_timo = timef_mark(750 MSEC); /* from now, break or no */
		break;

	/* XXX up/down one channel XXX should have a bandplan... */

	case '3':
		if (cfg.satellite) {
			if (doppler < 10)
				doppler++;
			qsy();
			goto dopp_changed;
		}
		rx_freq = cfg.rx_freq;
		ulmoddiv(&rx_freq, cfg.up_down);  /* truncate to channel */
		++rx_freq;                        /* one channel up */
		goto updn;

	case '6':
		if (cfg.satellite) {
			if (doppler > -10)
				doppler--;
			qsy();
			goto dopp_changed;
		}
		rx_freq = cfg.rx_freq - 1;        /* one kHz down */
		ulmoddiv(&rx_freq, cfg.up_down);  /* truncate to channel */
	updn:
		rx_freq *= cfg.up_down;           /* back to kHz */

		frequency(rx_freq);
	dopp_changed:
		keycnt = 0;
		d = 1;
		key_timo = timef_mark(750 MSEC); /* from now, break or no */
		break;
	}

	if (d)
		display_soon();
}

static void
do_afc()
{
	signed char off = ad_in7 - afc_zero;

	if (off < -10) {
		if (doppler > -10) {
			doppler--;
			qsy();
			display_soon();
		}
	}
	if (off > +10) {
		if (doppler < 10) {
			doppler++;
			qsy();
			display_soon();
		}
	}
}

#ifdef P8x

/* CLK2 originally 1968.75 Hz,
 * might have been rewired to 4.032 MHz (TMR_CLK)
 *
 * OUT2 must idle at '1' or FX429/OUT2/HOOK interrupts get blocked.
 *
 * currently unused. */

static void
timer_2()
{
#if 0
	if (h z) {
		/* old ctcss tx was 64...254 (or 640...2540 if 4017) */

		word count = (TMR_CLK + hz / 2) / hz;

		disable();
		iow(TMR_CTRL, TMR_SEL_2 | TMR_BOTH | TMR_SQWAVE);
		iow(TMR_2, count);
		iow(TMR_2, count >> 8);
		restore();
	} else
#endif
	{
		/* mode 4, but count is not is written.
		 * output goes and stays HIGH */

		iow(TMR_CTRL, TMR_SEL_2 | TMR_LSB | TMR_SWTRIGG);
	}
}
#endif

#ifdef L8x

static void
timer_brg()
{
	/* XXX L8x could provide other baudrates than 2400,4800,9600 */

	dword hz = 16 * 9600L;
	word count = (TMR_CLK + hz / 2) / hz;

	disable();
	iow(TMR_CTRL, TMR_SEL_BRG | TMR_BOTH | TMR_SQWAVE);
	iow(TMR_BRG, count);
	iow(TMR_BRG, count >> 8);
	restore();
}
#endif

/* matches table in setup.c */

static const byte
	sio_formats[] = {
		WR4_X64_CLK, /* 2400 */
		WR4_X32_CLK, /* 4800 */
		WR4_X16_CLK, /* 9600 */

		WR4_X64_CLK | WR4_PARITY_ENB | WR4_PARITY_EVEN,
		WR4_X32_CLK | WR4_PARITY_ENB | WR4_PARITY_EVEN,
		WR4_X16_CLK | WR4_PARITY_ENB | WR4_PARITY_EVEN,

		WR4_X64_CLK | WR4_PARITY_ENB | WR4_PARITY_ODD,
		WR4_X32_CLK | WR4_PARITY_ENB | WR4_PARITY_ODD,
		WR4_X16_CLK | WR4_PARITY_ENB | WR4_PARITY_ODD,
	};

static byte
sio_format(byte i)
{
	if (i >= sizeof(sio_formats))
		i = 0;

	return (sio_formats[i]);
}

static void
sio_a_setup()
{
#ifdef P8x
	iow(SIO_A_CSR, 4);
	iow(SIO_A_CSR, WR4_1STOPBIT | sio_format(cfg.sio_a_fmt));
	iow(SIO_A_CSR, 3);
	iow(SIO_A_CSR, WR3_8BIT_RX | WR3_RX_ENB);
	iow(SIO_A_CSR, 5);
	iow(SIO_A_CSR, sio_a_wr5);
	iow(SIO_A_CSR, 1);
	iow(SIO_A_CSR, WR1_STATUS_VECTOR
	             | WR1_ESCINT_ENB | WR1_TXINT_ENB | WR1_INT_ALLC);
#endif

#ifdef L8x
	/*
	 *  8-bit sync D7 instead of 16-bit C4D7:
	 *  re-hunt has more time to catch back to back packets, and
	 *  tx start/stop does not blurt out real sync words.
	 */
	iow(SIO_A_CSR, 4);
	iow(SIO_A_CSR, WR4_8BITSYNC); /* XXX should be 16BITSYNC, but ... */
	iow(SIO_A_CSR, 6);
	iow(SIO_A_CSR, revbyte[0xAA]); /* tx sync is sent manually */
	iow(SIO_A_CSR, 7);
	iow(SIO_A_CSR, revbyte[0xD7]); /* rx sync XXX second byte of SYNC 0xC4D7 */
	iow(SIO_A_CSR, 1);
	iow(SIO_A_CSR, WR1_STATUS_VECTOR
	             | WR1_ESCINT_ENB | WR1_TXINT_ENB | WR1_INT_ALLC);
	iow(SIO_A_CSR, 5);
	iow(SIO_A_CSR, sio_a_wr5);
	iow(SIO_A_CSR, 3);
	iow(SIO_A_CSR, WR3_8BIT_RX | WR3_RX_ENB | WR3_ENTER_HUNT);
#endif
}

/* in place of the pibolar brom, chipselect /CSPROM.
 * A3 might work as /WR. or might not work. */

static void
init_MT8888()
{
#ifdef L8M
	/* no free chipselects available */
#else
	ior(MT8888_RS);
	iow(MT8888_WC, 0x0);
	iow(MT8888_WC, 0x0);

	iow(MT8888_WC, 0x8); /* CRA: RSEL 1  IRQ  0  CP/DTMF# 0  TOUT   0 */
	iow(MT8888_WC, 0x1); /* CRB: C/R# 0  S/D# 0  TEST     0  BURST# 1 */

	ior(MT8888_RS);
#endif
}

/* interrupts disabled, will be enabled after this returns */

void
initialise()
{
	iow(WDOG, 0);
	iow(DA_TPC, 0);

	/* sometimes battery a/d sticks at FF. input > vcc latchup ?
	 * maybe this unlatches, mux kept at IN7 == GND for a longish time.
	 * no, it doesn't help... XXX unsolvable ? */

	iow(AD_IN7, 0);

	/* select 48 kB contiguous ROM0
	 * and 16 kB work ram.
	 * O2_SMEM is really ROMA16 on P8E, 27C512 not affected.
	 * bootstrap has done something like this already. */
#ifdef P8x
	iow(CSMEM, 1);     /* P8E SMEM */
	iow(CSKEY, 0);     /* P8E bipolar prom (not fitted) +5v on */
#endif

	iow(OUT_0, out_0);
	iow(OUT_1, out_1);
	iow(OUT_2, out_2);    /* ROM0 and P8N SMEM */

	/* lml */

	timer_mt(444);
	out_0_change(O0_MTC | O0_VOLUME, O0_MTC | 1);

#ifdef P8x
	timer_2();         /* used to be optional ctcss tx */

	iow(FX429_CSR, 4); /* modem reset */
	delay_usec(3000);  /* delay_msec would require ticking */
	iow(FX429_CSR, 0);
#endif

#ifdef L8x
	timer_brg();
#endif

	/* PIO-A */

	iow(PIO_A_DATA, pio_a);        /* all inputs (except PA_SLOCKBE)    */
	iow(PIO_A_CTRL, 0);            /* in case it was waiting something  */

	iow(PIO_A_CTRL, (byte) &iv_pio_a);
	iow(PIO_A_CTRL, 0xCF);         /* operating mode 3                  */
	iow(PIO_A_CTRL, pio_a_ddr);    /* 1 input, 0 output pin             */
	iow(PIO_A_CTRL, 0x97);         /* interrupt control: IE OR LOW MASK */
	iow(PIO_A_CTRL, 0xFE);         /* monitor mask, just PA.0           */

	/* PIO-B */

	iow(PIO_B_DATA, pio_b);
	iow(PIO_B_CTRL, 0);

	iow(PIO_B_CTRL, (byte) &iv_pio_b);
	iow(PIO_B_CTRL, 0xCF);
	iow(PIO_B_CTRL, pio_b_ddr);
	iow(PIO_B_CTRL, 0x97);
#ifdef PB_AVAIL
	iow(PIO_B_CTRL, ~PB_AVAIL);    /* monitor mask, just PB_AVAIL */
#else
	iow(PIO_B_CTRL, ~0);    /* monitor mask, nothing */
#endif

	/* cfg
	 * after PIO init. PIOB.0 == RAMA12 must be valid with P8N */

#ifdef P8x
	display_2[7] = "ENMT"[logic];
#endif
	progress('0');

	load_bbram();        /* only once, here, while I is still 0 */
	new_defaults();
	synth_setup();

	progress('1');

	ior(TMR_CTRL); /* P8x: reset HOOK edge detective (and TMR_2_OUT latch) */

	/* SIO-A */

	ior(SIO_A_CSR);
	iow(SIO_A_CSR, WR0_CHANNEL_RESET);

	sio_a_setup();

	iow(SIO_A_CSR, WR0_RESET_ESCINT);
	sio_a_rr0 = ior(SIO_A_CSR);
	iow(SIO_A_CSR, WR0_RESET_ESCINT);

	/* SIO-B */

	ior(SIO_B_CSR);
	iow(SIO_B_CSR, WR0_CHANNEL_RESET);
	nop();

	iow(SIO_B_CSR, 4);
	iow(SIO_B_CSR, WR4_1STOPBIT | sio_format(cfg.sio_b_fmt));
	iow(SIO_B_CSR, 3);
	iow(SIO_B_CSR, WR3_8BIT_RX | WR3_RX_ENB);
	iow(SIO_B_CSR, 5);
	iow(SIO_B_CSR, sio_b_wr5);
	iow(SIO_B_CSR, 1);
	iow(SIO_B_CSR, WR1_STATUS_VECTOR
	             | WR1_ESCINT_ENB | WR1_TXINT_ENB | WR1_INT_ALLC);

	iow(SIO_B_CSR, 2);
	iow(SIO_B_CSR, (byte) iv_sio);

	iow(SIO_B_CSR, WR0_RESET_ESCINT);
	sio_b_rr0 = ior(SIO_B_CSR);
	iow(SIO_B_CSR, WR0_RESET_ESCINT);

	progress('2');

	/* handset, if any */

	cu_probe();

	progress('3');

	debug_puts("\r\n" "\033[2H");
	debug_puts(version + 4); /* XXX */
	debug_puts("\r\n");

	/* enable receive */

	LPF(cfg.lpf_rx);           /* when receiving, help ax input (or such) */

	doppler = 0;
	afc = 0;
	qsy();

	/* all short buffers to free list */

	{
		struct buf *p = buf_space;

		do {
			buf_relse(p);
		} while (++p < &buf_space[__SIZE(buf_space)]);
	}

#ifdef P8x
	fx429_cmd = FX429_RXENB;
	iow(FX429_CSR, fx429_cmd);
#endif

#ifdef L8x
	ior(SIO_A_CSR);
	iow(SIO_A_CSR, 3);
	iow(SIO_A_CSR, WR3_8BIT_RX | WR3_RX_ENB | WR3_ENTER_HUNT);
#endif

	init_MT8888();

	/* stop lml */

	timer_mt_off();
	out_0_clr(O0_MTC);

	volume(+0);

	progress('-');

	lights_on();
	update_leds();
	display_soon();
}

static void
ax_test()
{
	byte *p, buf[100];

	p = axaddr(axaddr(buf, "QST"), cfg.mycall);

	p[-1] |= 0x01;       /* Last */
	*p++ = 0x03;         /* UI Text */
	*p++ = 0xF0;

	p = strcpyx(p, "hoplofobinen hopliitti");

	ax_send(buf, p - buf);
}

static void
fx_test()
{
	fx_send("\xC1\x61\x60\x12\x34\x5F", 6);
	fx_send("\xC1\x61\x60\x22\x33\x55", 6);
}

#define HOME "\033[5H\033[L"

void
handle_incoming(struct buf *p)
{
	int i, n = p->len;

	switch (p->type) {

	case FSK:
		n -= 2;

		if (fx_check(p->data, n))         /* nz if error */
			break;

		debug_puts(HOME "FSK  ");
		for (i = 0; i < n; ++i) {
			debug_putc(hexdig[p->data[i] >> 4]);
			debug_putc(hexdig[p->data[i] & 15]);
		}
		debug_puts("\r\n");
		break;

	case CCIR:
		debug_puts(HOME "CCIR ");
		goto ascii;

	case DTMF:
		debug_puts(HOME "DTMF ");
ascii:
		for (i = 0; i < n; ++i) {
			debug_putc(p->data[i]);
		}
		debug_puts("\r\n");
		break;
	}
	buf_relse(p);
}

/* TASK_0 */

/* everytime */

void
hare()
{
#if 0 /* "I'm here" */

#ifdef P8x
	out_2_toggle(O2_ROMA15);
#endif

#ifdef L8TM
	iow(BANK, bank ^= BNK_ROM1A15);
#endif

#endif

	if (ptt_pressed()) {
		if (! transmitting && ptt_works) {
			transmit_on();

			if (keycnt == 1)       /* XXX */
				ax_test();

			if (keycnt == 2)
				fx_test();

			if (keycnt == 3)
				morse_send(buffer);

			if (keycnt == 4)
				dtmf_burst(buffer);

			if (keycnt >= 5)
				ccir_send(buffer);

			out_0_clr(O0_MICM);
		}
	} else {
		if (transmitting && ! fx_transmitting()) {
			transmit_off();
		}
		ptt_works = 1;
	}
	if (tot && time_passed(tot)) {
		transmit_off();
		ptt_works = 0;
	}
}

/* every second or so */

#define HOME "\033[3H"

void
slug()
{
	static byte zapata;

	if ((++zapata & 15) == 0) {
		int i;

		debug_puts(HOME " ");
		for (i = 0; i < 8; ++i) {
			byte c = display_1[i];
			debug_putc(c ? c : ' ');
		}
		for (i = 0; i < 10; ++i) {
			byte c = display_2[i];
			debug_putc(c ? c : ' ');
		}
		debug_puts("\r\n");
	}
	if (transmit_tail)
		transmit_tail--;

	if (afc && !transmitting && !transmit_tail)
		do_afc();
}

/* TASK_2, must not return */

void
task_2()
{
	for (;;) {
		sleep();

		icon.car_d = off_hook();
		display_soon();
	}
}

/* 256 Hz zerolevel interrupt */

void
tick()
{
}

/*
 *  NMI causes
 *  - RESET pulse from watchdog (doesn't create /RESET)
 *    or loss of supply (cpu /RESET in 1 msec or so)
 *  - on/off button changes to off-position (bounces sometimes)
 *
 *  return is safe, if not much is done.
 *
 *  if it is/was 'RESET', output ports are/were undefined XXX arg-g-h...
 */

void
power_loss()
{
	iow(DA_TPC, 0);

#ifdef L8M
	/* no time to store into EEPROM */
#else
	if (timex() > 3)        /* pumping condition ? */
		save_bbram();
#endif
}

void
power_off()
{
	disable();

	iow(DA_TPC, 0);

	save_bbram();

	iow(PIO_B_DATA, pio_b | PB_PWROFF);

	delay_usec(10000);  /* 10 msec */
	boot();             /* d/c to Abiko ? :) */
}

void
incoming(struct buf *p, byte type)
{
	p->type = type;

	enq_incoming(p);
}

/* XXX decoders could collect directly to a buf
 * XXX decoders should rewind on qsy XXX */

void
incoming_b(byte *buf, byte len, byte type)
{
	struct buf *p = buf_alloc();

	if (p) {
		bcopy(buf, p->data, p->len = len);
		incoming(p, type);
	}
}

/* dtmf_change/_timeout called from tick()
 *
 * MT8870, MT8888, MT8889 databits are all equal.
 * bit 0x10 is 8870-like 'StD', emulated if not real 8870.
 *
 * XXX MT8889 timing is incompatible with Z80,
 * XXX A1 must be used as R/#W.
 * XXX gaaah, seems not work. and 8888 timing looks unsuitable too */

static const byte dtmf_8870_tab[] = "D1234567890*#ABC";

static byte dtmf_n, dtmf_buf[60];

word dtmf_timo;

void
dtmf_timeout()
{
	/* ignore if truncated */

	if (dtmf_n && dtmf_n < sizeof(dtmf_buf))
		incoming_b(dtmf_buf, dtmf_n, DTMF);

	dtmf_n = 0;
	dtmf_timo = 0;
}

/* XXX adjust timeout according to input rate */

void
dtmf_change(byte c)
{
	if (c & 0x10) {
		if (dtmf_n < sizeof(dtmf_buf))
			dtmf_buf[dtmf_n++] = dtmf_8870_tab[c & 0xF];

		dtmf_timo = 0;

	} else if (dtmf_n)
		dtmf_timo = timef_mark(5000 MSEC); /* 5 seconds after StD drops */
}

struct buf *in_pkt;
byte in_pkt_ix, in_pkt_len;

#ifdef P8x

/* sometimes a false alarm, caused by HOOK or TMR_2 */

static void
fx429_interrupt()
{
	byte c, stat;

	/* XXX extra dancing around edge sensitivity */

	stat = ior(FX429_CSR);      /* clears /INT unless RXRDY */

	if (stat & FX429_RXRDY) {
		byte nstat;             /* stat |= ior() would confuse cc */

		c = ior(FX429_DATA);    /* clears /INT from RXRDY */

		nstat = ior(FX429_CSR); /* clears new /INT from other sources */
		stat |= nstat;
	}
	if (stat & FX429_TXRDY) {
		static byte out_pkt_ix;

		struct buf *p = fx_tx_head;

		if (p) {
			if (stat & FX429_TXIDL) /* oopsidaisies, try again */
				out_pkt_ix = 0;

			iow(FX429_DATA, p->data[out_pkt_ix++]);

			if (out_pkt_ix == p->len) {
				out_pkt_ix = 0;
				if ((fx_tx_head = p->next) == NULL)
					fx_tx_tail = &fx_tx_head;
				buf_relse(p);
			}
		} else {
			if (stat & FX429_TXIDL) {
				fx429_cmd &= ~FX429_TXENB;
				iow(FX429_CSR, fx429_cmd);
			}
		}
	}
	if (stat & FX429_SYNC) {
		in_pkt_ix = 0;
		if (in_pkt == NULL)
			in_pkt = buf_alloc();
	}
	if (stat & FX429_RXRDY) {
		if (in_pkt)
			in_pkt->data[in_pkt_ix] = c;

		if (++in_pkt_ix == 1) {
			if ((c & 0xF0) == 0xC0)
				in_pkt_len = 8;
			else
				in_pkt_len = 15;

			fx429_cmd |= FX429_RXFMT;         /* crc by software */
			iow(FX429_CSR, fx429_cmd);

		} else if (in_pkt_ix == in_pkt_len) {
			if (in_pkt) {
				in_pkt->len = in_pkt_len;
				incoming(in_pkt, FSK);
				in_pkt = NULL;
			}
			in_pkt_ix = 0;

			fx429_cmd &= ~FX429_RXFMT;    /* wait sync again */
			iow(FX429_CSR, fx429_cmd);
		}
	}
}

#endif /* P8x fx429_interrupt */

#ifdef L8x

void
sio_a_rca()
{
	byte c = revbyte[ior(SIO_A_DATA)];

	if (in_pkt == NULL)
		in_pkt = buf_alloc();

	if (in_pkt)
		in_pkt->data[in_pkt_ix] = c;

	if (++in_pkt_ix == 1) {
		if ((c & 0xF0) == 0xC0)
			in_pkt_len = 8;
		else
			in_pkt_len = 15;

	} else if (in_pkt_ix == in_pkt_len) {
		ior(SIO_A_CSR);
		iow(SIO_A_CSR, 3);
		iow(SIO_A_CSR, WR3_8BIT_RX | WR3_RX_ENB | WR3_ENTER_HUNT);

		if (in_pkt) {
			in_pkt->len = in_pkt_len;
			incoming(in_pkt, FSK);
			in_pkt = NULL;
		}
		in_pkt_ix = 0;
	}
}

void
sio_a_tbe()
{
	static byte out_pkt_ix, out_pkt_pad;

	struct buf *p = fx_tx_head;

	if (p) {
		iow(SIO_A_DATA, revbyte[p->data[out_pkt_ix]]);

		if (++out_pkt_ix == p->len) {

			if ((fx_tx_head = p->next) == NULL)
				fx_tx_tail = &fx_tx_head;

			buf_relse(p);
			out_pkt_ix = 0;
			out_pkt_pad = 2;
		}
		return;
	}
	if (out_pkt_pad) {
		iow(SIO_A_DATA, revbyte[0xAA]);
		out_pkt_pad--;
		return;
	}
	iow(SIO_A_CSR, WR0_RESET_TXINT);

	sio_a_change(SA_TXE, 0);
}

#endif /* L8x SIO A sync rca/tbe */

/*
 *  sio_a_src,
 *  sio_a_rca, sio_a_tbe (async),
 *  sio_a_getc, sio_a_putc,
 *  in boot.s
 */

/*
 *  dcd: HOOK (or FX429 or TMR_2 on P8x)
 *  cts: CU53 DA or CU58AF /INT (handled already)
 *  Underrun/EOM on L8x
 *  SYNC_A (unusable on L8x, pin is output in sync modes)
 *  BREAK character detected or terminated
 */

void
sio_a_esc(byte changes)      /* new state in sio_a_rr0 */
{
#ifdef P8x

	/* falling edge of FX429 /INT or TMR_2_OUT
	 * or either edge of HOOK
	 * toggles 74HC107 */

	if (changes & SA_KKINT) {
		static byte
			pa;
		byte diff;

		fx429_interrupt();

		diff = pa;
		pa = ior(PIO_A_DATA);
		diff ^= pa;
		if (diff & PA_OFFHOOK) {
			lights_on();
			wake(TASK_2);
		}
	}
#endif

#ifdef L8x

	if (changes & SA_ONHOOK) {
		lights_on();
		wake(TASK_2);
	}

	/* XXX if underrun, rewind fx tx */

#endif
}

/*
 *  sio_b_xxx
 *  good candidates to boot.s
 *  or just ++(byte)p
 */

#define sio_b_tx_bump(p) \
	do { if (++p == sio_b_tx_bufend) p = sio_b_tx_buf; } while (0)

#define sio_b_rx_bump(p) \
	do { if (++p == sio_b_rx_bufend) p = sio_b_rx_buf; } while (0)

byte all_mine;

void
sio_b_putc(word c)
{
	disable();

	if (sio_b_tx_active == 0
	 && ((ior(SIO_B_CSR) & SB_DMIDLE) || all_mine)) {
		iow(SIO_B_DATA, c);
		sio_b_tx_active = all_mine = 1;
	} else {
		*sio_b_tx_ins = c;
		sio_b_tx_bump(sio_b_tx_ins);
	}
	restore();
}

void
sio_b_puts(byte *s)
{
	while (*s)
		sio_b_putc(*s++);
}

void
sio_b_src()
{
	/* parity, framing or overrun. just discard XXX */

	ior(SIO_B_DATA);
	iow(SIO_B_CSR, WR0_ERROR_RESET);

	/* XXX if sending, rewind, then start or defer */
}

void
sio_b_rca()
{
	*sio_b_rx_ins = ior(SIO_B_DATA);
	sio_b_rx_bump(sio_b_rx_ins);
}

void
sio_b_tbe()
{
	if (sio_b_tx_rem != sio_b_tx_ins) {
		iow(SIO_B_DATA, *sio_b_tx_rem);
		sio_b_tx_bump(sio_b_tx_rem);
		sio_b_tx_active = all_mine = 1;
	} else {
		iow(SIO_B_CSR, WR0_RESET_TXINT);
		sio_b_tx_active = 0;
	}
}

/*
 *  cts: /PTT
 *  dcd: DMIDLE (aka NETFREE)
 *  sync: /LOCAL
 *  BREAK character detected or terminated
 *
 *  lowest priority
 */

void
sio_b_esc(byte changes)        /* new state in sio_b_rr0 */
{
	if (changes & SB_PTT)
		wake(TASK_0);

	/* some randomness too from 1 Mohm, 47nF and 74HC04. maybe not much */

	if (ior(SIO_B_CSR) & SB_DMIDLE) {

		/* last transmit-buffer-empty interrupt occurred a long time ago */

		if (sio_b_tx_active == 0
		 && sio_b_tx_rem != sio_b_tx_ins) {
			iow(SIO_B_DATA, *sio_b_tx_rem);
			sio_b_tx_bump(sio_b_tx_rem);
			sio_b_tx_active = all_mine = 1;
		} else
			all_mine = 0;
	}
}

#if 0
/*
 *  monitormask is currently empty
 *
 *  could be used to catch EXIN1/2
 *  or another periodic (from TMR_2, rewiring)
 *  or sw uart input to DCU
 *
 *  2nd highest priority
 */

void
pio_b_int()
{
}

void
pio_b_enable()
{
	iow(PIO_B_CTRL, 0x83);
}

void
pio_b_disable()
{
	disable();                 /* needed, according to folklore */
	iow(PIO_B_CTRL, 0x03);
	restore();
}

#endif

