
/* sigh. */

#include "head.h"
#include "cfg.h"
#include "cu.h"
#include "iomap.h"

/* typical defaults for band and synth parameters  */

#ifdef P8x
#define _433    433       /* RD58 ? */
#define _432000 432000
#define _433500 433500
#define _438000 438000
#define _440000 440000
#define _128    128
#define _86512  86512L
#define _SYNTH  SYNTH_S8D
#endif

#ifdef L8TM
#define _433    145       /* RB660 ? */
#define _432000 144000
#define _433500 145500
#define _438000 146000
#define _440000 150000
#define _128    64
#define _86512  21400L
#define _SYNTH  SYNTH_COMPUTEC
#endif

#ifdef L8M
#define _433    50        /* RB58VY ? */
#define _432000 50000
#define _433500 50000
#define _438000 52000
#define _440000 60000
#define _128    40
#define _86512  45000L
#define _SYNTH  SYNTH_S8M_12
#endif

extern dword debug_NA;
extern word debug_R;

typedef static const byte * const stables;
stables

speakers_tab[] = {
	"EAr",
	"IntErnAL",
	"ExtErnAL",
	NULL },

sio_fmt_tab[] = {
	"2400 8n", "4800 8n", "9600 8n",
	"2400 8E", "4800 8E", "9600 8E",
	"2400 8o", "4800 8o", "9600 8o",
	NULL },

sql_method_tab[] = {
	"nOISE",
	"rSSI",
	"in7",
	NULL },

logic_tab[] = {
	"P8E",
	"P8N",
	"L8M",
	"L8TM",
	NULL },

synthesizer_tab[] = {
	"S8d",
	"S8c",
	"COMPUtEC", /* whatsitsname S/16/58/S ? */
	"S8M_12",
	"S8M_10",
	NULL },

dtmf_dec_tab[] = {
	"OFF",
	"8870",
	"8888",
	NULL },

off_on_tab[] = {
	"OFF",
	"On",
	NULL };

enum {
	SUBMENU,
	MEMORY,
	SPECIAL,

	TABLE,
	BYTE,

	WORD,
	DWORD,
	LONG,

	STRING,
};

enum {
	SPEC_full_reset,
	SPEC_logic,        /* read only */
	SPEC_rfc,
	SPEC_squelch,
	SPEC_squelch_bi,
	SPEC_batt,         /* read only */
	SPEC_decoders,     /* read only */
	SPEC_time,         /* read only */
};

struct setup {
	byte type;
	void *addr;
	word def;         /* or submenu length */
	byte **tab;       /* or 16 msbits */
	byte pad;         /* align to 16 bytes */
	byte name[8];
};

static struct setup const

#define TAB__(n,a,d,t) { TABLE,   a, d, t,  0,n },
#define BYTE_(n,a,d)   { BYTE,    a, d, 0,  0,n },
#define WORD_(n,a,d)   { WORD,    a, d, 0,  0,n },
#define STR__(n,a)     { STRING,  a, 0, 0,  0,n },
#define SPEC_(n,x)     { SPECIAL, x, 0, 0,  0,n },
#define MENU_(n,m,z)   { SUBMENU, m, z, 0,  0,n },

#define LONG_(n,a,d)  { LONG,    a, (d) & 65535, (byte **) ((d) >> 16), 0,n },
#define DWORD_(n,a,d) { DWORD,   a, (d) & 65535, (byte **) ((d) >> 16), 0,n },

#define BOOL_(n,a,d)  TAB__(n,a,d, off_on_tab)
#define OFFS_(n,a,d)  LONG_(n,a,d)
#define FREQ_(n,a,d)  DWORD_(n,a,d)

setup_memories[] = {
	{ MEMORY, 0,0,0,0, "        " },
},
setup_sundry[] = {
	WORD_("LIGHtS",   &cfg.lights,             60)
	BYTE_("VOLUME",   &cfg.volume,              1)
	TAB__("AF to",    &cfg.af_dest,             0, speakers_tab)
	WORD_("LPFt",     &cfg.lpf_tx,           3000)
	WORD_("LPFr",     &cfg.lpf_rx,              0)
	SPEC_("rFc",      SPEC_rfc)
	BYTE_("dBZEro",   &cfg.dB_zero,             0) /* ADC value of zero dB */
	BYTE_("dBSCALE",  &cfg.dB_scale,          125) /* 5V ~ 1000cB */
	BYTE_("GAIn",     &cfg.deviation,          15)
	BYTE_("dtGAIn",   &cfg.dtmf_gain,          50)
	BYTE_("ccGAIn",   &cfg.ccir_gain,         200)
	BYTE_("PrGAIn",   &cfg.afsk_gain,         100)
	BYTE_("ctGAIn",   &cfg.ctcss_gain,         20)
	BYTE_("CW SPd",   &cfg.morse_cpm,         100)
	WORD_("CWtonE",   &cfg.morse_Hz,         1000)
	WORD_("dtbSPd",   &cfg.dtmf_speed,        100)
	WORD_("3diGit",   &cfg.implied,          _433)
	WORD_("UPdn",     &cfg.up_down,            25)
#if 0
	FREQ_("CALL F",   &cfg.rx_freq,       _433500)
	FREQ_("QSY F",    &cfg.rx_freq,       _433500)
#endif
	FREQ_("r cEntEr", &cfg.rx_band_center,_440000)
	FREQ_("t cEntEr", &cfg.tx_band_center,_440000)
	FREQ_("t bEGin",  &cfg.tx_band_start, _432000)
	FREQ_("t End",    &cfg.tx_band_end,   _438000)
	WORD_("tr to",    &cfg.max_over,          600)
	TAB__("Synth",    &cfg.synthesizer,    _SYNTH, synthesizer_tab)
	WORD_("PrEScALE", &cfg.prescaler,        _128)
	OFFS_("IF OFFSt", &cfg.if_offset,      _86512)
	SPEC_("LOGIC",    SPEC_logic)
	SPEC_("CLOCK",    SPEC_time)
	SPEC_("FuLrSt",   SPEC_full_reset)
	BOOL_("no OFF",   &cfg.stay_on,             0)
	TAB__("dtMF dEc", &cfg.dtmf_decoder,        0, dtmf_dec_tab)
	BOOL_("FULduPL",  &cfg.fdx,                 0)
	BOOL_("SAtELLiT", &cfg.satellite,           0)
	TAB__("A bAud",   &cfg.sio_a_fmt,           1, sio_fmt_tab) /* 4800 bd */
	TAB__("b bAud",   &cfg.sio_b_fmt,           2, sio_fmt_tab) /* 9600 bd */
	WORD_("LobAtt",   &cfg.batt_cutoff,  1670 * 6)
	STR__("r58bis",   version + 11)

	WORD_("PLL r",    &debug_R,             0)
	DWORD_("PLL nA",   &debug_NA,            0L)
},
setup_status[] = {
	BYTE_("AdrSSI",   &ad_rssi,      0)
	BYTE_("AdSQL",    &ad_sql,       0)
	SPEC_("AdbAtt",   SPEC_batt)
	BYTE_("AdtPc",    &ad_tpc,       0)
	BYTE_("AdFPM",    &ad_fpm,       0)
	BYTE_("AdrPM",    &ad_rpm,       0)
#ifdef P8x
	BYTE_("AdtP4",    &ad_tp4,       0)
#else
	BYTE_("AdrFc",    &ad_tp4,       0) /* XXX L8M ? */
#endif
	BYTE_("AdIn7",    &ad_in7,       0)
	SPEC_("dEcodErS", SPEC_decoders)
},
setup_squelch[] = {
	SPEC_("SQL",      SPEC_squelch)
	WORD_("dELAY",    &cfg.sql_open,          250)
	WORD_("tAIL",     &cfg.sql_close,         333)
	BYTE_("HYSt",     &cfg.sql_hyst,            2)
	SPEC_("SQL bi",   SPEC_squelch_bi)
	TAB__("tYPE",     &cfg.sql_method,          0, sql_method_tab)
	BOOL_("Lit UP",   &cfg.sql_light,           0)
},
setup_packet[] = {
	STR__("MYCALL",   &cfg.mycall)
	STR__("id",       &cfg.id)
	WORD_("tdELAY",   &cfg.ax_txdelay,        200)
#if 0
	STR__("P0",       0xD000 + 16 + 256 * 0)
	STR__("P1",       0xD000 + 16 + 256 * 1)
	STR__("P2",       0xD000 + 16 + 256 * 2)
	STR__("P3",       0xD000 + 16 + 256 * 3)
	STR__("P4",       0xD000 + 16 + 256 * 4)
	STR__("P5",       0xD000 + 16 + 256 * 5)
	STR__("P6",       0xD000 + 16 + 256 * 6)
	STR__("P7",       0xD000 + 16 + 256 * 7)
#endif
},
setup_main[] = {
	BYTE_("tPc",      &cfg.tpc,                 0)
	OFFS_("t OFFSEt", &cfg.tx_offset,           0L)
	WORD_("CtCSS",    &cfg.ctcss,               0)
	MENU_("MEMOrY",   setup_memories,  N_MEMORIES)
	MENU_("SQUELCH",  setup_squelch,   __SIZE(setup_squelch))
	MENU_("StAtuS",   setup_status,    __SIZE(setup_status))
	MENU_("PAcKEt",   setup_packet,    __SIZE(setup_packet))
	MENU_("SundrY",   setup_sundry,    __SIZE(setup_sundry))
};

struct setup *setup = setup_main;

byte
	setup_active,
	setup_last,
	setup_ix,
	setup_len = __SIZE(setup_main),
	*alphakeys = "00112233445566778899";

static byte const * const
squelch_sources[] = {
	&ad_sql, &ad_rssi, &ad_in7
};

static struct setup *
setup_ptr()
{
	struct setup *p = setup;

	if (p->type != MEMORY)
		p += setup_ix;

	return (p);
}

static byte *old_cfg_end;

static void
defaults(struct setup *p, byte len)
{
	if (p->type == MEMORY)
		return;

	do {
		if (p->type == SUBMENU) {
			defaults(p->addr, p->def);
		} else {
			byte *addr = p->addr;

			/* only touched, if cfg.xxx */

			if (addr >= old_cfg_end
			 && addr < cfg_END) {
				switch (p->type) {

				case TABLE:
				case BYTE:
					*addr = p->def;
					break;

				case WORD:
					store_word(addr, p->def);
					break;

				case DWORD:
				case LONG:
					store_dword(addr, *(dword *) &p->def);
					break;
				}
			}
		}
		++p;
	} while (--len);
}

void
new_defaults()
{
	word keep = cfg.size;

	/* - just garbage
	 * - newer or same version
	 * - older version */

	if (cfg.magic != MAGIC
	 || keep > cfg_END - cfg_START
	 || keep < offsetof(struct cfg, size) + sizeof(cfg.size))
		keep = 0;
	else
	if (keep >= sizeof(cfg))
		return;

	old_cfg_end = cfg_START + keep;

	bzero(old_cfg_end, sizeof(cfg) - keep);

	defaults(setup_main, __SIZE(setup_main));

	cfg.magic = MAGIC;
	cfg.size  = sizeof(cfg);
}

void
full_reset()
{
	disable();
	bzero(bbram, bbend - bbram);

	new_defaults();

	save_bbram();
	boot();
}

static long
repeat()
{
	if (keycnt) {
		keycnt = 0;
		return (atodw(buffer));
	}
	return (1);
}

byte
table_len(struct setup *p)
{
	byte **v = p->tab, n = 0;

	while (*v++)
		++n;

	return (n);
}

static void
show_string(byte *s)
{
	byte blank = display_2_len, n = strlen(s);

	if (n > blank)
		n = blank;
	blank -= n;

	bcopy(s, display_2 + blank, n);
}

/* display_1[] and display_2[] are surrounded by
 * guard areas, no need to check for crazy longs */

static void
show_num(word offset, word n)
{
	format_word(display_2_end - offset, n);
}

static void
show_dword(dword n)
{
	format_dword(display_2_end, n);
}

static void
show_long(long n)
{
	format_long(display_2_end, n);
}

static void
show_table(byte **tab, byte x)
{
	while (x && *tab) {
		++tab;
		--x;
	}
	show_string(*tab ? *tab : "?");
}

static void
show_special(struct setup *p)
{
	switch ((byte) p->addr) {

	case SPEC_logic:
		show_table(logic_tab, logic);
		break;

	case SPEC_rfc:               /* 654 210 */
		show_num(4, ad_rssi);
		show_num(0, RFC_get());
		dpy_XXX = 256 / 3;
		break;

	case SPEC_squelch:
		show_num(4, *squelch_sources[cfg.sql_method]);
		show_num(0, cfg.sql_level);
		break;

	case SPEC_squelch_bi:
		show_num(4, ad_rssi);
		show_num(0, cfg.sql_bi_rssi);
		break;

	case SPEC_batt:
		show_num(0, ad_batt_millivolt());
		break;

	case SPEC_decoders:
		format_hex(display_2_end - 0, read_ccir());   /* 65 32 0 */
		format_hex(display_2_end - 2, read_8870());
#ifndef L8M
		format_hex(display_2_end - 5, read_8888());
#endif
		dpy_XXX = 256 / 3;
		break;

	case SPEC_time:
		show_dword(timex());
		break;
	}
}

static void
assign_special(struct setup *p, byte *s)
{
	switch ((byte) p->addr) {

	case SPEC_squelch:
		cfg.sql_level = atob(s);
		snap_squelch();
		break;

	case SPEC_squelch_bi:
		cfg.sql_bi_rssi = atob(s);
		break;

	case SPEC_rfc:
		RFC_set(atob(s));
		break;

	case SPEC_full_reset:
		if (atow(s) == 666)
			full_reset();
		break;
	}
}

static void
bump_special(struct setup *p, int add)
{
	switch ((byte) p->addr) {

	case SPEC_squelch:
		cfg.sql_level += add;
		snap_squelch();
		break;

	case SPEC_squelch_bi:
		cfg.sql_bi_rssi += add;
		break;

	case SPEC_rfc:
		RFC_set(RFC_get() + add);
		break;
	}
}

/*      -NNN
 * NN  NNNNNN */

static void
show_memory(struct memory *m)
{
	icon.key = (m->ctcss != 0);

	show_dword(m->rx_freq);

	format_long(display_1_end, m->tx_offset);
	format_word(display_2 + 2, setup_ix);
}

static void
assign_memory(struct memory *m, byte *s)
{
	switch (strlen(s)) {
	case 1:
	case 2:
		setup_ix = atow(s);
		if (setup_ix >= N_MEMORIES)
			setup_ix = N_MEMORIES - 1;
	case 0:
		break;

	case 3:
		m->flags = atob(s);
		break;
	case 4:
		m->ctcss = atow(s);
		break;
	default:
		if ((m->rx_freq = atodw(s)) == 0)
			bzero(m, sizeof(*m));
		break;
	}
}

void
display_setup()
{
	struct setup *p = setup_ptr();
	byte *addr = p->addr;

	bcopy(p->name, display_1, 8);

	if (keycnt) {
		buffer[keycnt] = alphamode ? '~' : '_';

		if (p->type == MEMORY) {
			switch (keycnt) {
			case 1:
			case 2:
				strcpyx(display_1, "Goto_");
				break;
			case 3:
				strcpyx(display_1, "FLAGS_");
				break;
			case 4:
				strcpyx(display_1, "tonE_");
				break;
			default:
				strcpyx(display_1, "FrEq_");
				break;
			}
		}
		return;
	}
	bzero(display_2, 10);

	switch (p->type) {

	case MEMORY:
		show_memory(&memories[setup_ix]);
		break;

	case TABLE:
		show_table(p->tab, *addr);
		break;

	case STRING:
		show_string(addr);
		break;

	case SPECIAL:
		show_special(p);
		break;

	case BYTE:
		show_num(0, *addr);
		break;

	case WORD:
		show_num(0, *(word*) addr);
		break;

	case DWORD:
		show_dword(*(dword*) addr);
		break;

	case LONG:
		show_long(*(long*) addr);
		break;
	}
}

/* XXX not here */

void
display()
{
	dpy_XXX = 0;

	/* nowhere ? */

	if (cu_update == nop)
		return;

	icon.key = 0;
	icon.v_u = 0;
	icon.v_d = 0;

	icon.bar_l = ad_rssi_peak & 0x80;
	icon.bar_r = ad_rssi_peak & 0x40;

	if (setup_active) {
		display_setup();

	} else {
		/* XXX */

		bzero(display_1, 8);

		display_1[0] = cfg.volume + '0';

		if (transmit_state & 0x80)
			format_word(display_1_end, cfg.tpc);
		else
			format_short(display_1_end, centibel(ad_rssi_peak));

		if (keycnt == 0) {
			byte sign;

			bzero(display_2, 10);

			sign = ' ';
			if (tx_freq != rx_freq && !cfg.satellite) {
				if (tx_freq < rx_freq) {
					sign = '-';
					icon.v_d = 1;
				} else {
					sign = '+';
					icon.v_u = 1;
				}
			}
			if (transmit_state)
				format_dword(display_2_end, tx_freq);
			else
				format_dword(display_2_end, rx_freq)[-1] = alpha ? sign : ' ';

			if (doppler) {
				display_2[0] = " -"[doppler < 0];
				display_2[1] = hexdig[abs(doppler)];
			}
			if (afc)
				display_2[2] = 'A';
		}
		icon.key = (cfg.ctcss != 0);
	}

#if 0
{
	extern byte sio_a_rr0;

	format_hex(display_2 + 2, sio_a_rr0);
	format_hex(display_2 + 4, ior(SIO_A_CSR));
}
#endif

	(*cu_update)();
}

static void
assign(byte *s)
{
	struct setup *p = setup_ptr();
	byte *addr = p->addr;

	switch (p->type) {

	case MEMORY:
		assign_memory(&memories[setup_ix], s);
		break;

	case SPECIAL:
		assign_special(p, s);
		break;

	case STRING:
		strncpyx(addr, s, sizeof(cfg_string));
		break;

	case TABLE: {
			byte last = table_len(p) - 1;
			byte n = atob(s);

			if (n > last)
				n = last;
			*addr = n;
		}
		break;

	case BYTE:
		*addr = atob(s);
		break;

	case WORD:
		store_word(addr, atow(s));
		break;

	case DWORD:
		store_dword(addr, atodw(s));
		break;

	case LONG:
		if (*(long*) addr > 0)             /* keep sign */
			store_dword(addr, atodw(s));
		else
			store_dword(addr, -atodw(s));
		break;
	}
}

static void
bump(long add)
{
	struct setup *p = setup_ptr();
	byte *addr = p->addr;

	switch (p->type) {

	case MEMORY:
		memories[setup_ix].tx_offset = add; /* XXX plain +/- behave badly */
		break;

	case SPECIAL:
		bump_special(p, add);
		break;

	/* re-edit. copy string into buffer */

	case STRING:
		bzero(buffer, sizeof(buffer));
		strncpyx(buffer, addr, sizeof(buffer));
		keycnt = strlen(buffer);
		break;

	case TABLE: {
			byte last = table_len(p) - 1;
			byte n = *addr + add;

			if (n == 0xFF)
				n = last;
			if (n > last)
				n = 0;
			*addr = n;
		}
		break;

	case BYTE:
		*addr += add;
		break;

	case WORD:
		store_word(addr, *(word*) addr + add);
		break;

	case DWORD:
		store_dword(addr, *(dword*) addr + add);
		break;

	case LONG:
		store_dword(addr, *(long*) addr + add);
		break;
	}
}

void
key_setup(byte c)
{
	struct setup *p = setup_ptr();

	if (isdigit(c)) {
		if (p->type == SUBMENU) {
			setup_last = setup_ix;
			setup_len = p->def;
			setup = p->addr;
			setup_ix = 0;
			return;
		}
		if (alphamode && c == last_key && keycnt)
			c = index(alphakeys, buffer[--keycnt])[1];

		if (keycnt == 0)
			bzero(buffer, sizeof(buffer));

		if (keycnt < display_2_len)
			buffer[keycnt++] = c;

		alphamode = (p->type == STRING);
		return;
	}
	alphamode = 0;

	switch (c) {
	case 'C':
		if (keycnt) {
			buffer[keycnt] = 0;       /* cursor */
			buffer[--keycnt] = 0;     /* character */
			break;
		}
		if (setup != setup_main) {
			setup = setup_main;
			setup_ix = setup_last;
			setup_len = __SIZE(setup_main);
			break;
		}
	case 'E':
		setup_active = 0;
		save_bbram();
		break;

	case '*':
		keycnt = 0;
		if (setup_ix == 0)
			setup_ix = setup_len;
		setup_ix--;
		break;

	case '#':
		if (keycnt) {
			buffer[keycnt] = 0;
			keycnt = 0;
			assign(buffer);
			break;
		}
		setup_ix++;
		if (setup_ix >= setup_len)
			setup_ix = 0;
		break;

	case '+':
		bump(+repeat());
		break;
	case '-':
		bump(-repeat());
		break;

	case 'S':
	case 'R':
		break;
	}
}

/* XXX a stack of positions or something needed */

void
quick_setup(byte key)
{
	switch (key) {

	case '1':
	case '4':
		setup = setup_squelch;
		setup_ix = 0;
		setup_len = __SIZE(setup_squelch);

		key_setup((key <= '3') ? '+' : '-');
		break;
	}
}

